/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "constants.h"
#include "state.h"
#include "packet.h"
#include "defs.h"
#include "md5.h"
#include "sha1.h"
#include "crypto.h" /* requires sha1.h and md5.h */
#include "log.h"
#include "ipsec_doi.h"
#include "demux.h"
#include "des.h"
#include "timer.h"

/* This file does basic header checking and demux of
 * incoming packets.
 */

/* Receive a packet. If we pass buffer to a routine that does not return
 * failure indication (e.g. the packet handling routines), it's up to them
 * to free it; otherwise this routine does.
 */
void
comm_handle(int kernelfd, int sock)
{
    struct sockaddr_in sin;
    struct isakmp_hdr hdr;
    struct state *st;
    bool encrypted = FALSE;
    pb_stream packet_pbs, message_pbs;

    /* ??? The whole logic of detecting a retransmission
     * and resending our old response is totally fishy:
     *
     * - The retransmitted block won't decrypt because
     *   the IV differs (unless the message wasn't decrypted
     *   or both messages were supposed to use the end-of-main
     *   IV (i.e. the retransmitted message was an informational
     *   exchange).  Because of this, we won't detect it as
     *   a legitimate message so we won't check if it was like
     *   the previous one.
     *
     * - We detect retransmission by examining the first payload
     *   type (not the payload).  What if the retransmitted message
     *   were different in content?  We just assume that it is
     *   identical and resend our previous response.
     *
     * The right solution is to simply byte-by-byte compare the
     * incoming message with the last one we got (do we have
     * to skip anything in the header?).  If it matches, THEN
     * we just send out our last response.  Since informational
     * exchanges are not part of a dialogue, they need not be
     * considered.
     */

    /* Now really read the message.
     * Since we don't know its size, we read it into
     * an overly large buffer and then copy it to a
     * new, properly sized buffer.
     */
    {
	int sin_len;
	int packet_len;
	/* ??? this buffer seems *way* too big */
	u_int8_t bigbuffer[UDP_SIZE];

	mksin(sin, 0, 0);
	sin_len = sizeof(sin);

	if ((packet_len = recvfrom(sock, bigbuffer, sizeof(bigbuffer), 0,
			  (struct sockaddr *)&sin, &sin_len)) == -1)
	{
	    log_errno("recvfrom() failed in comm_handle");
	    return;
	}

	/* Allocate exactly that much space
	 * and set up packet_pbs to describe it.
	 */
	init_pbs(&packet_pbs,
	    clone_bytes(bigbuffer, packet_len, "message buffer in comm_handle()"),
	    packet_len, "packet");
    }

    DBG(DBG_RAW,
	DBG_log("read %d bytes from %s", (int) pbs_room(&packet_pbs),
	    show_sa_in(&sin));
	DBG_dump("", packet_pbs.start, pbs_room(&packet_pbs)));

    if ((pbs_room(&packet_pbs) & 0x3) != 0)
    {
	/* We don't care.  Should we?
	 * ??? We don't even take care to make our output packets conform.
	 */
	log("message from %s has %d bytes, not a multiple of 4",
	    show_sa_in(&sin), (int) pbs_room(&packet_pbs));
    }

    if (!in_struct(&hdr, &isakmp_hdr_desc, &packet_pbs, &message_pbs))
    {
	/* XXX specific failures (special notification?):
	 * - bad ISAKMP major/minor version numbers
	 * - size of packet vs size of message
	 */
	free(packet_pbs.start);
	return;
    }

    if (packet_pbs.roof != message_pbs.roof)
    {
	log("packet size (%u) differs from size specified in ISAKMP HDR (%u)",
	    (unsigned) pbs_room(&packet_pbs), hdr.isa_length);
	return;
    }

    if (hdr.isa_xchg == ISAKMP_XCHG_IDPROT
    && hdr.isa_msgid != 0)
    {
	free(packet_pbs.start);
	log("message id should be zero (was 0x%08x) in base mode",
	    hdr.isa_msgid);
	/* XXX Could send notification back */
	return;
    }

    st = find_full_state(hdr.isa_icookie, hdr.isa_rcookie,
	    (struct sockaddr *)&sin, hdr.isa_msgid);

    if (hdr.isa_flags & ISAKMP_FLAG_ENCRYPTION)
    {
	if (st == (struct state *) NULL)
	{
	    struct state *nst;

	    /* Find the Phase 1 object */
	    st = find_full_state(hdr.isa_icookie, hdr.isa_rcookie,
		(struct sockaddr *)&sin, 0);
	    if (st == (struct state *) NULL)
	    {
		log("attempt to use non-existant (expired?) ISAKMP SA with %s",
		    show_sa_in(&sin));
		free(packet_pbs.start);
		/* XXX Could send notification back */
		return;
	    }

	    if (st->st_state != OAKLEY_MAIN_I_4
	    && st->st_state != OAKLEY_MAIN_R_3)
	    {
		log("attempt to use non-fully established ISAKMP SA with %s",
		    show_sa_in(&sin));
		free(packet_pbs.start);
		/* XXX Could send notification back */
		return;
	    }

	    nst = duplicate_state(st);
	    nst->st_msgid = hdr.isa_msgid;
	    nst->st_state = OAKLEY_QUICK_R_0;
	    insert_state(nst);
	    update_messageid((struct sockaddr *)&sin, nst->st_msgid);

	    /* Quick Mode Initial IV */
	    init_phase2_iv(nst);

	    st = nst;
	}

	DBG(DBG_CRYPT, DBG_log("received encrypted packet from %s",
	    show_sa_in(&sin)));

	if (st->st_skeyid_e == (u_char *) NULL
	|| st->st_iv == (u_char *) NULL)
	{
	    log("unexpected encrypted packet received from %s",
		show_sa_in(&sin));
	    free(packet_pbs.start);
	    /* XXX Could send notification back */
	    return;
	}

	/* Mark as encrypted */
	encrypted = TRUE;

	DBG(DBG_CRYPT, DBG_log("decrypting %u bytes using algorithm %s",
	    (unsigned) pbs_left(&message_pbs),
	    enum_show(&oakley_enc_desc, st->st_oakley.encrypt)));

	/* do the specified decryption
	 *
	 * IV (in and out) is from st->st_iv (st->st_iv_len)
	 *
	 * See draft-ietf-ipsec-isakmp-oakley-07.txt Appendix B
	 *
	 * XXX The IV should only be updated really if the packet
	 * is successfully processed.
	 * We should keep this value, check for a success return
	 * value from the parsing routines and then replace.
	 *
	 * Each post phase 1 exchange generates IVs from
	 * the last phase 1 block, not the last block sent.
	 */
	{
	    const struct encrypt_desc *e = st->st_oakley.encrypter;

	    if (pbs_left(&message_pbs) % e->blocksize != 0)
	    {
		    log("message from %s not a multiple of ecryption blocksize",
			show_sa_in(&sin));
		    /* XXX Could send notification back */
		    free(packet_pbs.start);
		    return;
	    }

	    /* XXX Detect weak keys */

	    /* Decrypt everything after header */
	    e->crypt(FALSE, message_pbs.cur, pbs_left(&message_pbs), st);
	}

	DBG_cond_dump(DBG_CRYPT, "next IV:", st->st_iv, st->st_iv_len);

	DBG_cond_dump(DBG_CRYPT, "decrypted:\n", message_pbs.cur,
	    message_pbs.roof - message_pbs.cur);
    }

    /* Preliminary pass to de-pad message.
     * Padding must be removed to make hashing work.
     * Padding comes from encryption (so this code must be after decryption)
     * and from the requirement that messages must be "aligned" (padded
     * to a multiple of 4 octets).
     * ??? We don't actually pad for alignment.  All our messages
     * seem to be naturally aligned.
     */
    {
	pb_stream tpbs = message_pbs;	/* copy so as not to disturb */
	struct isakmp_generic payload;
	int np;

	for (np = hdr.isa_np; np != ISAKMP_NEXT_NONE; np = payload.isag_np)
	{
	    if (!in_struct(&payload, &isakmp_generic_desc, &tpbs, NULL))
	    {
		log("malformed payload in packet from %s", show_sa_in(&sin));
		free(packet_pbs.start);
		return;
	    }
	}

	DBG(DBG_PARSING,
	    if (pbs_left(&tpbs) != 0)
		DBG_log("removing %d bytes of padding", (int) pbs_left(&tpbs)));

	message_pbs.roof = tpbs.cur;
    }

    if (st == (struct state *) NULL)
	st = find_half_state(hdr.isa_icookie, (struct sockaddr *)&sin);

    if (st == (struct state *) NULL)
    {
	/* We have no record of an state, so this must be
	 * the begining of an exchange
	 */
	switch (hdr.isa_xchg)
	{
	    case ISAKMP_XCHG_IDPROT:
	    {
		if (hdr.isa_np != ISAKMP_NEXT_SA)
		{
		    log("invalid payload %s from %s (expected SA in main I1)",
			enum_show(&payload_desc, hdr.isa_np),
			show_sa_in(&sin));
		    /* XXX Could send notification back */
		    free(packet_pbs.start);
		    return;
		}
		main_inI1_outR1(sock, (struct sockaddr *)&sin, &message_pbs, &hdr);
		break;
	    }

	    default:
		log("out of order packet from %s; expected ISAKMP_XCHG_IDPROT but got %s",
		    show_sa_in(&sin), enum_show(&exchange_desc, hdr.isa_xchg));
		/* XXX Could send notification back */
		free(packet_pbs.start);
		return;
	}
	return;
    }

    /* We have found a state */

    if (hdr.isa_xchg == ISAKMP_XCHG_INFO)
    {
	/* Handle Informational Exchanges -- Delete/Notifications */
	/* XXX Handle deletions */
	/* XXX Handle error messages */
	/* XXX handle badly formed informational exchanges */

	int pt = hdr.isa_np;

	while (pt != ISAKMP_NEXT_NONE) {
	    switch (pt)
	    {
	    case ISAKMP_NEXT_N:	/* Notification */
	    {
		struct isakmp_notification note;
		pb_stream note_pbs;

		if (!in_struct(&note, &isakmp_notification_desc, &message_pbs, &note_pbs))
		{
		    free(packet_pbs.start);
		    return;
		}
		/* XXX don't just do sit there, do something! */
		log("ignoring informational message from %s, type %s",
		    show_sa_in(&sin),
		    enum_show(&notification_desc, note.isan_type));
		pt = note.isan_np;
		break;
	    }
	    case ISAKMP_NEXT_D:	/* Delete */
	    {
		struct isakmp_delete del;
		pb_stream del_body;

		if (!in_struct(&del, &isakmp_delete_desc, &message_pbs, &del_body))
		{
		    free(packet_pbs.start);
		    return;
		}
		/* XXX actually do deletion */
		log("ignoring delete SA message from %s",
		    show_sa_in(&sin));
		pt = del.isad_np;
		break;
	    }
	    default:
	    {
		struct isakmp_generic gen;
		pb_stream gen_pbs;

		if (!in_struct(&gen, &isakmp_generic_desc, &message_pbs, &gen_pbs))
		{
		    free(packet_pbs.start);
		    return;
		}
		log("ignoring payload type %s from %s",
		    enum_show(&payload_desc, pt), show_sa_in(&sin));
		pt = gen.isag_np;
		break;
	    }
	    }
	}
	free(packet_pbs.start);
	return;
    }

    /* XXX Handle Commit Bit set and unset it */

    /* XXX The handling routines should return success/fail status, not void */

    assert(OAKLEY_MAIN_I_1 <= st->st_state && st->st_state <= OAKLEY_QUICK_R_2);

    {
	struct state_check {
	    u_int8_t needed_first_payload;
	    u_int8_t retry_payload; /* may retransmit if this found instead */
	    u_int8_t timeout_event;
	    stf_status (*processor)(
		int sock,	/* socket for packet input/output */
		const struct sockaddr *sap,	/* network address */
		pb_stream *message,	/* message to be processed */
		const struct isakmp_hdr *hdr,	/* message's header */
		pb_stream *reply,	/* room for reply */
		struct state *st	/* current state object */
		);
	};
	static struct state_check sc[OAKLEY_QUICK_R_2] = {
	    /* No state for main_out_I1: --> HDR;SA */

	    /* No state for main_inI1_outR1: HDR;SA --> HDR;SA */

	    /* OAKLEY_MAIN_I_1: HDR;SA --> HDR;KE;Ni */
	    { ISAKMP_NEXT_SA, ISAKMP_NEXT_SA, EVENT_RETRANSMIT, main_inR1_outI2 },

	    /* OAKLEY_MAIN_R_1: HDR;KE;Ni --> HDR;KE;Nr */
	    /* ??? why EVENT_CLEANUP, not EVENT_RETRANSMIT? */
	    { ISAKMP_NEXT_KE, ISAKMP_NEXT_SA, EVENT_CLEANUP, main_inI2_outR2 },

	    /* for states from here on, output message must be encrypted */

	    /* OAKLEY_MAIN_I_2: HDR;KE;Nr --> HDR*;IDii;HASH_I */
	    { ISAKMP_NEXT_KE, ISAKMP_NEXT_SA, EVENT_RETRANSMIT, main_inR2_outI3 },

	    /* for states from here on, input message must be encrypted */

	    /* OAKLEY_MAIN_R_2: HDR*;IDii;HASH_I --> HDR*;IDir;HASH_R */
	    { ISAKMP_NEXT_ID, ISAKMP_NEXT_KE, EVENT_SA_EXPIRE, main_inI3_outR3 },

	    /* OAKLEY_MAIN_I_3: HDR*;IDir;HASH_R --> done
	     * May initiate quick mode be calling quick_outI1
	     */
	    { ISAKMP_NEXT_ID, ISAKMP_NEXT_ID, EVENT_SA_EXPIRE, main_inR3 },

	    /* OAKLEY_MAIN_R_3: can only get here due to packet loss */
	    { ISAKMP_NEXT_NONE, ISAKMP_NEXT_ID,  EVENT_NULL, NULL},

	    /* OAKLEY_MAIN_I_4: can only get here due to packet loss */
	    { ISAKMP_NEXT_NONE, ISAKMP_NEXT_ID, EVENT_NULL, NULL },

	    /* No state for quick_outI1: --> HDR*, HASH(1), SA, Nr [, KE ] [, IDci, IDcr ] */

	    /* OAKLEY_QUICK_R_0:
	     * HDR*, HASH(1), SA, Ni [, KE ] [, IDci, IDcr ] -->
	     * HDR*, HASH(2), SA, Nr [, KE ] [, IDci, IDcr ]
	     */
	    /* ??? why EVENT_CLEANUP, not EVENT_RETRANSMIT? */
	    { ISAKMP_NEXT_HASH, ISAKMP_NEXT_HASH, EVENT_CLEANUP, quick_inI1_outR1 },

	    /* OAKLEY_QUICK_I_1:
	     * HDR*, HASH(2), SA, Nr [, KE ] [, IDci, IDcr ] -->
	     * HDR*, HASH(3)
	     */
	    { ISAKMP_NEXT_HASH, ISAKMP_NEXT_HASH, EVENT_SA_EXPIRE, quick_inR1_outI2 },

	    /* OAKLEY_QUICK_R_1: HDR*, HASH(3) --> done */
	    { ISAKMP_NEXT_HASH, ISAKMP_NEXT_HASH, EVENT_SA_EXPIRE, quick_inI2 },

	    /* OAKLEY_QUICK_I_2: can only happen due to lost packet */
	    { ISAKMP_NEXT_NONE, ISAKMP_NEXT_HASH, EVENT_NULL, NULL },

	    /* OAKLEY_QUICK_R_2: can only happen due to lost packet */
	    { ISAKMP_NEXT_NONE, ISAKMP_NEXT_HASH, EVENT_NULL, NULL },
	};
	const int from_state = st->st_state;
	struct state_check *scp = &sc[from_state - OAKLEY_MAIN_I_1];
	stf_status result;

	/* ??? this buffer seems *way* too big */
	u_int8_t reply_buffer[UDP_SIZE];
	pb_stream reply;

	/* encryption of input is needed from OAKLEY_MAIN_R_2 on */
	if (!encrypted && from_state >= OAKLEY_MAIN_R_2)
	{
	    if (from_state <= OAKLEY_MAIN_I_3
	    && hdr.isa_np == scp->retry_payload)
	    {
		/* Even though the message should have been
		 * encrypted, and it was not, it looks like
		 * it might be a retransmission of the previous
		 * message from him (it starts with the right payload)
		 * and that need not have been encrypted.
		 * It looks like we should just retransmit our
		 * previous packet.  To do this, we just fall
		 * through to the payload checking code.  We
		 * are sure that it will decide to retransmit.
		 */
		assert(hdr.isa_np != scp->needed_first_payload);
	    }
	    else
	    {
		log("packet from %s should have been encrypted",
		    show_sa_in(&sin));
		free(packet_pbs.start);
		/* XXX Could send notification back */
		return;
	    }
	}

	/* Check that first payload type is right.
	 * ISAKMP_NEXT_NONE is never right; if the table entry specifies it,
	 * that means no message is acceptable (the phase is over!).
	 */
	if (hdr.isa_np != scp->needed_first_payload
	|| hdr.isa_np == ISAKMP_NEXT_NONE)
	{
	    /* Simple-minded retransmit heuristic:
	     * If the payload isn't right for us, and it is
	     * right for the previous state, retransmit
	     * the packet we sent last time!
	     * This doesn't work well in Quick Mode where
	     * everything starts with a hash payload.
	     */
	    if (hdr.isa_np == scp->retry_payload)
	    {
		/* ??? We ought to have a counter to stop us doing this forever */
		/* ??? surely there is other retransmission code: make a subroutine! */
		if (sendto(sock, st->st_packet, st->st_packet_len, 0,
			   &st->st_peer, sizeof(st->st_peer))
		!= st->st_packet_len)
		{
		    log_errno("sendto() for retransmission failed in comm_handle() for %s",
			show_sa(&st->st_peer));
		}
		else
		{
		    log("retransmitted %d bytes to %s",
			st->st_packet_len, show_sa(&st->st_peer));
		}
	    }
	    else
	    {
		log("dropping out of sequence packet from %s (expecting %s)",
		    show_sa_in(&sin),
		    enum_show(&payload_desc, hdr.isa_np));
	    }
	    free(packet_pbs.start);
	    return;
	}

	if (from_state == OAKLEY_MAIN_I_3)
	{
	    /* Last block of Phase 1 (R3), kept for Phase 2 IV generation */
	    st->st_lastblock = clone_bytes(st->st_iv, st->st_iv_len, "lastblock in comm_handle()");
	    st->st_lastblock_len = st->st_iv_len;

	    DBG_cond_dump(DBG_CRYPT, "last encrypted Phase 1 block:", st->st_lastblock, st->st_lastblock_len);
	}

	/* set up reply pb_stream */
	init_pbs(&reply, reply_buffer, sizeof(reply_buffer), "reply packet");

	/* do the real work! */
	result = scp->processor(sock, (struct sockaddr *)&sin,
	    &message_pbs, &hdr, &reply, st);
	switch (result)
	{
	    case STF_FAIL:
		DBG(DBG_CONTROL,
		    DBG_log("state transition function for %s failed",
			enum_show(&state_desc, from_state)));
		free(packet_pbs.start);
		break;

	    case STF_NO_REPLY:
	    case STF_REPLY:
		/* Delete previous retransmission event */
		delete_event(st);

		/* free previous reply packet with the new one */
		free(st->st_packet);
		st->st_packet = NULL;

		/* if requested, send the new reply packet */
		if (result == STF_REPLY)
		{
		    close_output_pbs(&reply);   /* good form, but actually a no-op */

		    st->st_packet = clone_bytes(reply.start, pbs_offset(&reply),
			"reply packet");
		    st->st_packet_len = pbs_offset(&reply);

		    DBG_cond_dump(DBG_RAW, "sending:\n", st->st_packet, st->st_packet_len);

		    /* actually send the packet */
		    if (sendto(sock, st->st_packet, st->st_packet_len, 0
		    , (struct sockaddr *)&sin, sizeof(sin))
		    != st->st_packet_len)
		    {
			log_errno("sendto() failed");
		    }
		}

		/* Schedule for whatever timeout is specified */
		{
		    u_int32_t delay;

		    switch (scp->timeout_event)
		    {
		    case EVENT_RETRANSMIT:	/* Retransmit packet */
			delay = EVENT_RETRANSMIT_DELAY;
			break;
		    case EVENT_CLEANUP:	/* Cleanup responder state */
			delay = EVENT_CLEANUP_DELAY;
			break;
		    case EVENT_SA_EXPIRE:	/* SA expiration event */
			delay = st->st_oakley.expire == 0? DEFAULT_SA_EXPIRE : st->st_oakley.expire;
			break;

		    case EVENT_NULL:	/* non-event */
		    case EVENT_REINIT_SECRET:	/* Refresh cookie secret */
		    default:
			assert(FALSE);
		    }
		    event_schedule(scp->timeout_event, delay, st, 0);
		}

		break;
	}
    }
}
