/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

/*
 * This file has all the IPsec DOI and Oakley resolution routines.
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "des.h"

#include "constants.h"
#include "state.h"
#include "preshared.h"
#include "packet.h"
#include "defs.h"
#include "kernel.h"
#include "log.h"
#include "cookie.h"
#include "spdb.h"
#include "timer.h"
#include "rnd.h"
#include "ipsec_doi.h"

#include "sha1.h"
#include "md5.h"
#include "crypto.h" /* requires sha1.h and md5.h */

static void quick_outI1(int sock, struct state *st,
	struct sockaddr_in our_client_net, struct sockaddr_in our_client_mask,
	struct sockaddr_in peer_client_net, struct sockaddr_in peer_client_mask);

/* pad_up(n, m) is the amount to add to n to make it a multiple of m */
#define pad_up(n, m) (((m) - 1) - (((n) + (m) - 1) % (m)))

/* Convert MP_INT to network form (even number of binary bytes, big-endian).
 * We do the malloc; caller must eventually do free.
 */
static void
mpz_to_n(u_char **rptr, size_t *rlen, const MP_INT *mp)
{
    MP_INT temp1, temp2;
    size_t len = (mpz_sizeinbase(mp, 16) + 1) / 2;
    u_char *p = alloc_bytes(len, "host representation of large integer");
    int i;

    mpz_init(&temp1);
    mpz_init(&temp2);

    mpz_set(&temp1, mp);

    for (i = len-1; i >= 0; i--)
    {
	p[i] = mpz_mdivmod_ui(&temp2, NULL, &temp1, 1 << BITS_PER_BYTE);
	mpz_set(&temp1, &temp2);
    }

    mpz_clear(&temp1);
    mpz_clear(&temp2);

    *rptr = p;
    *rlen = len;
}

/* Convert network form (binary bytes, big-endian) to MP_INT
 * We do the mpz_init; caller must eventually do mpz_clear.
 */
static void
n_to_mpz(MP_INT *mp, const u_char *nbytes, size_t nlen)
{
    size_t i;

    mpz_init_set_ui(mp, 0);

    for (i = 0; i != nlen; i++)
    {
	mpz_mul_ui(mp, mp, 1 << BITS_PER_BYTE);
	mpz_add_ui(mp, mp, nbytes[i]);
    }
}

/* Dump number in network order hex */

#ifdef DEBUG
static void
DBG_dump_mpz(const char *label, const MP_INT *mp)
{
    u_char *p;
    size_t l;

    mpz_to_n(&p, &l, mp);

    DBG_dump(label, p, l);
    free(p);
}
#endif

/* Figure out which interface / IP address we are using for this exchange.
 * It is a peculiar fact that we don't know until we get a message
 * from our peer (in main_inI1_outR1 or main_inR1_outI2)!
 */
static void
glean_myidentity(int sock, struct state *st)
{
    struct sockaddr_in *sin = (struct sockaddr_in *)&st->st_me;
    int k = sizeof(st->st_me);

    if (getsockname(sock, &st->st_me, &k) < 0 || k != sizeof(st->st_me))
	exit_log_errno("getsockname() failed");

    if (sin->sin_family != AF_INET)
	exit_log("getsockname() returned unacceptable family (%d)", sin->sin_family);

    st->st_myidentity = clone_bytes(&sin->sin_addr, sizeof(sin->sin_addr), "st_myidentity in glean_myidentity1()");
    st->st_myidentity_len = sizeof(sin->sin_addr);
    st->st_myidentity_type = ID_IPV4_ADDR;
    DBG(DBG_CONTROL, DBG_log("my identity is %s", inet_ntoa(sin->sin_addr)));
}

/*
 * Send a notification to the peer. We could make a decision on
 * whether to send the notification, based on the type and the
 * destination, if we care to.
 * XXX It doesn't handle DELETE notifications (which are also
 * XXX informational exchanges).
 */
#if 0 /* not currently used */
//static void
//send_notification(int sock,
//    u_int16_t type,
//    u_char *spi,
//    u_char spilen,
//    u_char protoid,
//    u_char *icookie,
//    u_char *rcookie,
//    u_int32_t /*network order*/ msgid,
//    struct sockaddr sa)
//{
//    u_char buffer[sizeof(struct isakmp_hdr) +
//		 sizeof(struct isakmp_notification)];
//    struct isakmp_hdr *isa = (struct isakmp_hdr *) buffer;
//    struct isakmp_notification *isan = (struct isakmp_notification *)
//				       (buffer + sizeof(struct isakmp_hdr));
//
//    memset(buffer, '\0', sizeof(struct isakmp_hdr) +
//	  sizeof(struct isakmp_notification));
//
//    if (icookie != (u_char *) NULL)
//	memcpy(isa->isa_icookie, icookie, COOKIE_SIZE);
//
//    if (rcookie != (u_char *) NULL)
//	memcpy(isa->isa_rcookie, rcookie, COOKIE_SIZE);
//
//    /* Standard header */
//    isa->isa_np = ISAKMP_NEXT_N;
//    isa->isa_version = ISAKMP_MAJOR_VERSION << ISA_MAJ_SHIFT | ISAKMP_MINOR_VERSION;
//    isa->isa_xchg = ISAKMP_XCHG_INFO;
//    isa->isa_msgid = msgid;
//    isa->isa_length = htonl(sizeof(struct isakmp_hdr) +
//			    sizeof(struct isakmp_notification) +
//			    spilen);
//
//    /* Notification header */
//    isan->isan_type = htons(type);
//    isan->isan_doi = htonl(ISAKMP_DOI_IPSEC);
//    isan->isan_length = htons(sizeof(struct isakmp_notification) + spilen);
//    isan->isan_spisize = spilen;
//    memcpy((u_char *)isan + sizeof(struct isakmp_notification), spi, spilen);
//    isan->isan_protoid = protoid;
//
//    DBG(DBG_CONTROL, DBG_log("sending INFO type %s to %s",
//	enum_show(&notification_desc, type),
//	show_sa(&sa)));
//
//    if (sendto(sock, buffer, ntohl(isa->isa_length), 0, &sa,
//	       sizeof(sa)) != ntohl(isa->isa_length))
//	log_errno("sendto() failed in send_notification() to %s",
//	    show_sa(&sa));
//    else
//    {
//	DBG(DBG_CONTROL, DBG_log("transmitted %d bytes", ntohl(isa->isa_length)));
//    }
//}
#endif /* not currently used */

/*
 * Get a msgid for Phase 2 messages (network order!).
 */
static u_int32_t
oakley_get_msgid(struct sockaddr sa)
{
    u_int32_t /*network order*/ msgid;
    struct messageid *mid;

    msgid = find_messageid(sa);
    if (msgid == 0)
    {
	mid = insert_messageid(sa);
	msgid = mid->md_msgid;
    }

    DBG(DBG_CONTROL, DBG_log("MSG-ID is 0x%08x", msgid));

    return msgid;
}


/* Initiate an Oakley exchange.
 * --> HDR;SA
 */
static void
main_out_I1(
	int sock,
	struct sockaddr_in peer_sin,
	struct sockaddr_in our_client_net, struct sockaddr_in our_client_mask,
	struct sockaddr_in peer_client_net, struct sockaddr_in peer_client_mask,
	int goal)
{
    u_char space[8192];	/* NOTE: we assume 8192 is big enough to build the packet */
    pb_stream reply_pbs;	/* not actually a reply, but you know what I mean */
    pb_stream r_hdr_pbs;

    struct isakmp_hdr hdr;
    struct state *st;

    /* set up new state */
    st = new_state();

    st->st_myuser = our_client_net;
    st->st_myusernetmask = our_client_mask;
    st->st_peeruser = peer_client_net;
    st->st_peerusernetmask = peer_client_mask;

    memcpy(&st->st_peer, &peer_sin, sizeof(peer_sin));	/* sockaddr from sockaddr_in */

    /* Initiator cookie */
    get_cookie(ISAKMP_INITIATOR, st->st_icookie, COOKIE_SIZE, (struct sockaddr *)&peer_sin);

    /* set up reply_pbs */
    init_pbs(&reply_pbs, space, sizeof(space), "reply packet");

    /* HDR out */

    memset(&hdr, '\0', sizeof(hdr));	/* default to 0 */

    hdr.isa_version = ISAKMP_MAJOR_VERSION << ISA_MAJ_SHIFT | ISAKMP_MINOR_VERSION;
    hdr.isa_np = ISAKMP_NEXT_SA;
    hdr.isa_xchg = ISAKMP_XCHG_IDPROT;
    memcpy(hdr.isa_icookie, st->st_icookie, COOKIE_SIZE);
    /* R-cookie, flags and MessageID are left zero */

    if (!out_struct(&hdr, &isakmp_hdr_desc, &reply_pbs, &r_hdr_pbs))
	assert(FALSE);

    /* SA out */
    {
	u_char *sa_start = r_hdr_pbs.cur;

	if (!out_sa(&r_hdr_pbs, &oakley_sadb, st, TRUE, ISAKMP_NEXT_NONE))
	    assert(FALSE);

	/* save initiator SA for later HASH */
	assert(st->st_p1isa == NULL);	/* no leak! */
	st->st_p1isa_len = r_hdr_pbs.cur - sa_start;
	st->st_p1isa = clone_bytes(sa_start, st->st_p1isa_len, "sa in main_out_I1");
    }

    close_output_pbs(&r_hdr_pbs);
    close_output_pbs(&reply_pbs);

    st->st_goal = goal;
    st->st_state = OAKLEY_MAIN_I_1;

    st->st_packet = clone_bytes(reply_pbs.start, pbs_offset(&reply_pbs), "reply packet for main_out_I1");
    st->st_packet_len = pbs_offset(&reply_pbs);

    insert_state(st);

    /* Transmit */

    DBG_cond_dump(DBG_RAW, "sending:\n", st->st_packet, st->st_packet_len);

    if (sendto(sock, st->st_packet, st->st_packet_len, 0, (struct sockaddr *)&peer_sin,
	       sizeof(peer_sin)) != st->st_packet_len)
	log_errno("sendto() failed in main_out_I1()");

    /* Set up a retransmission event, half a minute henceforth */
    event_schedule(EVENT_RETRANSMIT, EVENT_RETRANSMIT_DELAY, st, 0);
}

/* Demux for IPsec DOI initiator.
 * --> HDR;SA
 */
void
ipsecdoi_initiate(
	int sock,
	struct sockaddr_in peer_sin,
	struct sockaddr_in our_client_net, struct sockaddr_in our_client_mask,
	struct sockaddr_in peer_client_net, struct sockaddr_in peer_client_mask,
	int goal,
	int protocol)
{
    switch (protocol)
    {
	case KEY_IKE:
	{
	    /* If there's already an ISAKMP SA established, use that and
	     * go directly to Quick Mode.
	     * XXX If an ISAKMP SA is *being* established, we foolishly
	     * XXX try to establish another one, in parallel. We could
	     * XXX issue an event to wait for it to finish and then try
	     * XXX to establish it.
	     */
	    struct state *st = find_phase1_state((struct sockaddr *)&peer_sin);

	    if (st == NULL)
		main_out_I1(sock, peer_sin, our_client_net, our_client_mask, peer_client_net, peer_client_mask, goal);
	    else
	    {
		DBG(DBG_CONTROL, DBG_log("Doing Quick Mode with %s, goal %s",
		    show_sa_in(&peer_sin),
		    bitnamesof(sa_goal_bit_names, goal)));
		st->st_goal = goal;
		quick_outI1(sock, st, our_client_net, our_client_mask, peer_client_net, peer_client_mask);
	    }
	    break;
	}

	default:
	    log("unsupported protocol %s specified in ipsecdoi_initiate()",
		enum_show(&isakmp_transformid_desc, protocol));
	    break;
    }
}

/* Handle a Main Mode Oakley first packet (responder side).
 * HDR;SA --> HDR;SA
 */
void
main_inI1_outR1(
	int sock,
	struct sockaddr *sap,
	pb_stream *message,
	const struct isakmp_hdr *hdr)
{
    struct isakmp_sa sa;
    pb_stream sa_pbs;
    struct state *st;

    u_char space[8192];	/* NOTE: we assume 8192 is big enough to build the packet */
    pb_stream reply_pbs;	/* not really a reply */
    pb_stream r_hdr_pbs;
    pb_stream r_sa_pbs;

    assert(hdr->isa_np == ISAKMP_NEXT_SA);

    /* start processing of SA -- rest later */

    if (!in_struct(&sa, &isakmp_sa_desc, message, &sa_pbs))
    {
#if 0	/* This may once have worked.  It is an endangered species. */
	log("malformed packet from %s detected in main_inI1_outR1",
	    show_sa(sap));
	send_notification(sock, PAYLOAD_MALFORMED, NULL, 0,
			  PROTO_ISAKMP, hdr->isa_icookie, hdr->isa_rcookie, 0,
			  *sap);
#endif
	free(message->start);
	return;
    }

    if (sa.isasa_np != ISAKMP_NEXT_NONE)
    {
	log("unexpected second payload (%s) in main I1 message from %s",
	    enum_show(&payload_desc, sa.isasa_np),
	    show_sa(sap));
	/* XXX Could send notification back */
	free(message->start);
	return;
    }

    /* Set up state */
    st = new_state();

    memcpy(st->st_icookie, hdr->isa_icookie, COOKIE_SIZE);

    get_cookie(ISAKMP_RESPONDER, st->st_rcookie, COOKIE_SIZE, sap);

    glean_myidentity(sock, st);

    st->st_peer = *sap;
    st->st_doi = ISAKMP_DOI_IPSEC;
    st->st_situation = SIT_IDENTITY_ONLY; /* We only support this */
    st->st_state = OAKLEY_MAIN_R_1;

    /* parse_isakmp_sa also spits out a winning SA into our reply,
     * so we have to build our reply_pbs and emit HDR before calling it.
     */
    init_pbs(&reply_pbs, space, sizeof(space), "reply packet");

    /* HDR out */
    {
	struct isakmp_hdr r_hdr = *hdr;

	memcpy(r_hdr.isa_rcookie, st->st_rcookie, COOKIE_SIZE);
	r_hdr.isa_np = ISAKMP_NEXT_SA;	/* redundant, but clean */
	if (!out_struct(&r_hdr, &isakmp_hdr_desc, &reply_pbs, &r_hdr_pbs))
	    assert(FALSE);
    }

    /* start of SA out */
    {
	struct isakmp_sa r_sa = sa;

	r_sa.isasa_np = ISAKMP_NEXT_NONE;
	if (!out_struct(&r_sa, &isakmp_sa_desc, &r_hdr_pbs, &r_sa_pbs))
	    assert(FALSE);
    }

    /* SA body in and out */

    if (!parse_isakmp_sa_body(&sa_pbs, &sa, &r_sa_pbs, FALSE, st
    , st->st_rcookie, st->st_icookie, sap))
    {
	/* forget this happened */
	free(st);
	free(message->start);
	return;
    }

    close_output_pbs(&r_hdr_pbs);

    /* save packet, now that we know its size */
    st->st_packet = clone_bytes(reply_pbs.start, pbs_offset(&reply_pbs),
	"reply packet from main_inI1_outR1");
    st->st_packet_len = pbs_offset(&reply_pbs);

    /* save initiator SA for HASH */
    assert(st->st_p1isa == NULL);	/* no leak! */
    st->st_p1isa_len = pbs_room(&sa_pbs);
    st->st_p1isa = clone_bytes(sa_pbs.start, pbs_room(&sa_pbs), "sa in main_inI1_outR1()");

    /* send the packet */

    DBG_cond_dump(DBG_RAW, "sending:\n", st->st_packet, st->st_packet_len);

    if (sendto(sock, st->st_packet, st->st_packet_len, 0, &st->st_peer,
	       sizeof(st->st_peer)) != st->st_packet_len)
	log_errno("sendto() failed in quick_outI1()");

    insert_state(st);

    event_schedule(EVENT_RETRANSMIT, EVENT_RETRANSMIT_DELAY, st, 0);
}


/*
 * SKEYID for preshared keys.
 */
static void
skeyid_preshared(struct state *st, u_char *id, int length)
{
    struct hmac_ctx ctx;

    hmac_init(&ctx, st->st_oakley.hasher, id, length);

    st->st_skeyid = alloc_bytes(ctx.hmac_digest_len, "st_skeyid in skeyid_preshared()");
    st->st_skeyid_len = ctx.hmac_digest_len;

    hmac_update(&ctx, st->st_ni, st->st_ni_len);
    hmac_update(&ctx, st->st_nr, st->st_nr_len);
    hmac_final(st->st_skeyid, &ctx);
}


/*
 * Generate the SKEYID_*
 */
static void
generate_skeyids(struct state *st)
{
    switch (st->st_oakley.auth)
    {
	case OAKLEY_PRESHARED_KEY:
	    /* Generate the SKEYID */
	    {
		int len;
		u_char *id = get_preshared_key(st, &len);

		skeyid_preshared(st, id, len);
		free(id);
	    }
	    break;

	case OAKLEY_DSS_SIG:
	case OAKLEY_RSA_SIG:
	    /* XXX */

	case OAKLEY_RSA_ENC:
	case OAKLEY_RSA_ENC_REV:
	    /* XXX */

	default:
	    exit_log("generate_skeyids(): unsupported authentication method %s",
		enum_show(&oakley_auth_desc, st->st_oakley.auth));
    }

    {
	u_char *gxy;
	size_t gxylen;
	struct hmac_ctx ctx;

	/* Get the g^xy for HMAC'ing */
	mpz_to_n(&gxy, &gxylen, &st->st_shared);

	DBG(DBG_CRYPT, DBG_log("size of g^xy is %d", gxylen));

	hmac_init(&ctx, st->st_oakley.hasher, st->st_skeyid, st->st_skeyid_len);

	st->st_skeyid_d_len = ctx.hmac_digest_len;
	st->st_skeyid_d = alloc_bytes(ctx.hmac_digest_len, "st_skeyid_d in generate_skeyids()");

	st->st_skeyid_a_len = ctx.hmac_digest_len;
	st->st_skeyid_a = alloc_bytes(ctx.hmac_digest_len, "st_skeyid_a in generate_skeyids()");

	st->st_skeyid_e_len = ctx.hmac_digest_len;
	st->st_skeyid_e = alloc_bytes(ctx.hmac_digest_len, "st_skeyid_e in generate_skeyids()");

	/* SKEYID_D */
	hmac_update(&ctx, gxy, gxylen);
	hmac_update(&ctx, st->st_icookie, COOKIE_SIZE);
	hmac_update(&ctx, st->st_rcookie, COOKIE_SIZE);
	hmac_update(&ctx, "\0", 1);
	hmac_final(st->st_skeyid_d, &ctx);

	/* SKEYID_A */
	hmac_reinit(&ctx);
	hmac_update(&ctx, st->st_skeyid_d, st->st_skeyid_d_len);
	hmac_update(&ctx, gxy, gxylen);
	hmac_update(&ctx, st->st_icookie, COOKIE_SIZE);
	hmac_update(&ctx, st->st_rcookie, COOKIE_SIZE);
	hmac_update(&ctx, "\1", 1);
	hmac_final(st->st_skeyid_a, &ctx);

	/* SKEYID_E */
	hmac_reinit(&ctx);
	hmac_update(&ctx, st->st_skeyid_a, st->st_skeyid_a_len);
	hmac_update(&ctx, gxy, gxylen);
	hmac_update(&ctx, st->st_icookie, COOKIE_SIZE);
	hmac_update(&ctx, st->st_rcookie, COOKIE_SIZE);
	hmac_update(&ctx, "\2", 1);
	hmac_final(st->st_skeyid_e, &ctx);

	free(gxy);
    }

    /* IV */
    {
	u_char *gi, *gr;
	size_t gilen, grlen;
	union hash_ctx hash_ctx;
	const struct hash_desc *h = st->st_oakley.hasher;

	mpz_to_n(&gi, &gilen, &st->st_gi);
	mpz_to_n(&gr, &grlen, &st->st_gr);

	st->st_iv_len = h->hash_digest_len;
	st->st_iv = alloc_bytes(st->st_iv_len, "st_iv in generate_skeyids()");

	h->hash_init(&hash_ctx);
	h->hash_update(&hash_ctx, gi, gilen);
	h->hash_update(&hash_ctx, gr, grlen);
	h->hash_final(st->st_iv, &hash_ctx);

	free(gi);
	free(gr);
    }

    /* Oakley Keying Material
     * Derived from Skeyid_e: if it is not big enough, generate more
     * using the PRF.
     * See draft-ietf-ipsec-isakmp-oakley-07.txt Appendix B
     */
    {
	const size_t keysize = st->st_oakley.encrypter->keysize;

	st->st_enc_key_len = keysize;
	if (keysize <= st->st_skeyid_e_len)
	{
	    st->st_enc_key = clone_bytes(st->st_skeyid_e, keysize, "st_enc_key 1");
	}
	else
	{
	    u_char keytemp[MAX_OAKLEY_KEY_LEN + MAX_DIGEST_LEN];
	    struct hmac_ctx ctx;
	    int i = 0;

	    hmac_init(&ctx, st->st_oakley.hasher, st->st_skeyid_e, st->st_skeyid_e_len);
	    hmac_update(&ctx, "\0", 1);
	    for (;;)
	    {
		hmac_final(&keytemp[i], &ctx);
		i += ctx.hmac_digest_len;
		if (i >= keysize)
		    break;
		hmac_reinit(&ctx);
		hmac_update(&ctx, &keytemp[i - ctx.hmac_digest_len], ctx.hmac_digest_len);
	    }
	    st->st_enc_key = clone_bytes(keytemp, keysize, "st_enc_key 2");
	}
    }

    DBG(DBG_CRYPT,
	DBG_dump("Skeyid:  ", st->st_skeyid, st->st_skeyid_len);
	DBG_dump("Skeyid_d:", st->st_skeyid_d, st->st_skeyid_d_len);
	DBG_dump("Skeyid_a:", st->st_skeyid_a, st->st_skeyid_a_len);
	DBG_dump("Skeyid_e:", st->st_skeyid_e, st->st_skeyid_e_len);
	DBG_dump("enc key:", st->st_enc_key, st->st_enc_key_len);
	DBG_dump("IV:", st->st_iv, st->st_iv_len));
}

/* Handle HDR;SA from responder (verify proposal),
 * and send back HDR;KE;Ni.
 */
stf_status
main_inR1_outI2(
    int sock,	/* socket for packet input/output */
    const struct sockaddr *sap,	/* network address */
    pb_stream *message,	/* message to be processed */
    const struct isakmp_hdr *hdr,	/* message's header */
    pb_stream *reply,	/* room for reply */
    struct state *st)	/* current state object */
{
    struct isakmp_sa sa;
    pb_stream sa_pbs;
    pb_stream r_hdr_pbs;    /* reply message pbs */

    glean_myidentity(sock, st);

    if (!in_struct(&sa, &isakmp_sa_desc, message, &sa_pbs))
	return STF_FAIL;

    if (!parse_isakmp_sa_body(&sa_pbs, &sa, NULL, TRUE, st
    , st->st_icookie, hdr->isa_rcookie, sap))
	return STF_FAIL;

    /* st->st_sec := localsecret */
    {
	u_char tmp[LOCALSECRETSIZE];

	get_rnd_bytes(tmp, LOCALSECRETSIZE);
	st->st_sec_in_use = TRUE;
	n_to_mpz(&st->st_sec, tmp, LOCALSECRETSIZE);
    }

    /* st->st_gi := groupgenerator ** st->st_sec  %  st->st_oakley.modp_modulus */
    mpz_init(&st->st_gi);
    st->st_gi_in_use = TRUE;
    mpz_powm(&st->st_gi, &groupgenerator, &st->st_sec, st->st_oakley.modp_modulus);

    DBG(DBG_CRYPT,
	DBG_dump_mpz("Local secret:", &st->st_sec);
	DBG_dump_mpz("Public value sent:", &st->st_gi));

    /**************** build output packet HDR;KE;Ni ****************/

    /* HDR */
    {
	struct isakmp_hdr r_hdr = *hdr;	/* mostly same as incoming header */

	r_hdr.isa_np = ISAKMP_NEXT_KE;
	if (!out_struct(&r_hdr, &isakmp_hdr_desc, reply, &r_hdr_pbs))
	    return STF_FAIL;
    }

    /* KE */
    {
	pb_stream keyex_pbs;

	u_char *gi;
	size_t gilen;

	if (!out_generic(ISAKMP_NEXT_NONCE, &isakmp_keyex_desc, &r_hdr_pbs, &keyex_pbs))
	    return STF_FAIL;

	mpz_to_n(&gi, &gilen, &st->st_gi);

	if (!out_raw(gi, gilen, &keyex_pbs, "keyex value"))
	    return STF_FAIL;
	free(gi);
	close_output_pbs(&keyex_pbs);
    }

    /* Ni */
    {
	pb_stream nonce_pbs;
	u_int8_t nonce_bytes[DEFAULTNONCESIZE];

	if (!out_generic(ISAKMP_NEXT_NONE, &isakmp_nonce_desc, &r_hdr_pbs, &nonce_pbs))
	    return STF_FAIL;
	get_rnd_bytes(nonce_bytes, DEFAULTNONCESIZE);
	if (!out_raw(nonce_bytes, DEFAULTNONCESIZE, &nonce_pbs, "nonce value"))
	    return STF_FAIL;
	close_output_pbs(&nonce_pbs);

	/* Copy the Ni nonce into the state object */
	st->st_ni = clone_bytes(nonce_bytes, DEFAULTNONCESIZE,
	    "st_ni in main_inR1_outI2()");
	st->st_ni_len = DEFAULTNONCESIZE;
    }

    /* finish message */
    close_output_pbs(&r_hdr_pbs);

    /* Reinsert the state, using the responder cookie we just received */
    delete_state(st);
    memcpy(st->st_rcookie, hdr->isa_rcookie, COOKIE_SIZE);
    insert_state(st);

    st->st_state = OAKLEY_MAIN_I_2;

    return STF_REPLY;
}

/* Generate HASH_I or HASH_R for ISAKMP Phase I.
 * This will *not* generate other hash payloads (eg. Phase II of Quick Mode,
 * New Group Mode, or ISAKMP Informational Exchanges).
 * If the hashi argument is TRUE, generate HASH_I; if FALSE generate HASH_R.
 * If hashus argument is TRUE, we're generating a hash for our end.
 */
static void
main_mode_hash(struct state *st, u_char *hash_val, int *hash_len, bool hashi, bool hashus)
{
    struct hmac_ctx ctx;

    hmac_init(&ctx, st->st_oakley.hasher, st->st_skeyid, st->st_skeyid_len);

    *hash_len = ctx.hmac_digest_len;

    {
	u_char *gi, *gr;
	size_t gilen, grlen;

	mpz_to_n(&gi, &gilen, &st->st_gi);
	mpz_to_n(&gr, &grlen, &st->st_gr);
	if (hashi)
	{
	    hmac_update(&ctx, gi, gilen);
	    hmac_update(&ctx, gr, grlen);
	    hmac_update(&ctx, st->st_icookie, COOKIE_SIZE);
	    hmac_update(&ctx, st->st_rcookie, COOKIE_SIZE);
	}
	else
	{
	    hmac_update(&ctx, gr, grlen);
	    hmac_update(&ctx, gi, gilen);
	    hmac_update(&ctx, st->st_rcookie, COOKIE_SIZE);
	    hmac_update(&ctx, st->st_icookie, COOKIE_SIZE);
	}
	free(gi);
	free(gr);
    }

    DBG(DBG_CRYPT, DBG_log("hashing %d bytes of SA",
	    st->st_p1isa_len - sizeof(struct isakmp_generic)));

    /* SA_b */
    hmac_update(&ctx, st->st_p1isa + sizeof(struct isakmp_generic),
	 st->st_p1isa_len - sizeof(struct isakmp_generic));

    /* IDio_b (o stands for originator: i or r) */
    {
	/* hash identification payload, without generic payload header */
	struct isakmp_ipsec_id id;	/* body used, in network order! */

	memset(&id, '\0', sizeof(id));

	if (hashus) {
	    id.isaiid_idtype = st->st_myidentity_type;
	    /* leave protoid and port empty (zero) */
	} else {
	    id.isaiid_idtype = st->st_peeridentity_type;
	    id.isaiid_protoid = st->st_peeridentity_protocol;
	    id.isaiid_port = htons(st->st_peeridentity_port);
	}
	DBG(DBG_CRYPT,
	    DBG_log("Hashing %s ID: Type %s, Protocol %d, Port %d",
		hashus? "my" : "his",
		enum_show(&ident_desc, id.isaiid_idtype),
		id.isaiid_protoid, htons(id.isaiid_port)));

	/* NOTE: hash does NOT include the generic payload part of
	 * Identity Payload
	 */
	hmac_update(&ctx,
	    (u_char *)&id + sizeof(struct isakmp_generic),
	    sizeof(id) - sizeof(struct isakmp_generic));
    }

    if (hashus)
	hmac_update(&ctx, st->st_myidentity, st->st_myidentity_len);
    else
	hmac_update(&ctx, st->st_peeridentity, st->st_peeridentity_len);

    hmac_final(hash_val, &ctx);
}

/* encrypt message, sans fixed part of header */
static bool
encrypt_message(pb_stream *pbs, struct state *st)
{
    const struct encrypt_desc *e = st->st_oakley.encrypter;
    u_int8_t *enc_start = pbs->start + sizeof(struct isakmp_hdr);
    size_t enc_len = pbs_offset(pbs) - sizeof(struct isakmp_hdr);

    DBG_cond_dump(DBG_CRYPT | DBG_RAW, "encrypting:\n", enc_start, enc_len);

    /* pad up to multiple of encryption blocksize */
    {
	size_t padding = pad_up(enc_len, e->blocksize);

	if (!out_zero(padding, pbs, "encryption padding"))
	    return FALSE;
	enc_len += padding;
    }

    DBG(DBG_CRYPT, DBG_log("encrypting using %s", enum_show(&oakley_enc_desc, st->st_oakley.encrypt)));

    e->crypt(TRUE, enc_start, enc_len, st);

    DBG_cond_dump(DBG_CRYPT, "next IV:", st->st_iv, st->st_iv_len);
    return TRUE;
}

/* Handle HDR;KE;Nr from responder.  Send a HDR*;IDii;HASH_I back.
 */
stf_status
main_inR2_outI3(
    int sock,	/* socket for packet input/output */
    const struct sockaddr *sap,	/* network address */
    pb_stream *message,	/* message to be processed */
    const struct isakmp_hdr *hdr,	/* message's header */
    pb_stream *reply,	/* room for reply */
    struct state *st)	/* current state object */
{
    struct isakmp_generic keyex;
    pb_stream keyex_pbs;
    struct isakmp_generic nonce;
    pb_stream nonce_pbs;
    pb_stream r_hdr_pbs;    /* reply message pbs */

    if (hdr->isa_np != ISAKMP_NEXT_KE)
    {
	log("expected ISAKMP_NEXT_KE first in main R2 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, hdr->isa_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    if (!in_struct(&keyex, &isakmp_keyex_desc, message, &keyex_pbs))
	return STF_FAIL;

    if (pbs_left(&keyex_pbs) < MINIMUM_PUBLIC_VALUE_SIZE)
    {
	log("too short public size (%d) sent from %s",
	    pbs_left(&keyex_pbs), show_sa(sap));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    if (keyex.isag_np != ISAKMP_NEXT_NONCE)
    {
	log("expected ISAKMP_NEXT_NONCE second in main I2 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, keyex.isag_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    if (!in_struct(&nonce, &isakmp_nonce_desc, message, &nonce_pbs))
	return STF_FAIL;

    if (nonce.isag_np != ISAKMP_NEXT_NONE)
    {
	log("expected ISAKMP_NEXT_NONE third in main I2 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, nonce.isag_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    DBG_cond_dump(DBG_CRYPT, "Ni:", nonce_pbs.cur, pbs_left(&nonce_pbs));

    /* done parsing */

    /* decode Gr from network form */
    st->st_gr_in_use = TRUE;
    n_to_mpz(&st->st_gr, keyex_pbs.cur, pbs_left(&keyex_pbs));

    /* st->st_shared := st->st_gr ** st->st_sec  %  st->st_oakley.modp_modulus */
    mpz_init(&st->st_shared);
    st->st_shared_in_use = TRUE;
    mpz_powm(&st->st_shared, &st->st_gr, &st->st_sec, st->st_oakley.modp_modulus);

    DBG(DBG_CRYPT,
	DBG_dump_mpz("public value received:", &st->st_gr);
	DBG_dump_mpz("shared secret:", &st->st_shared));

    /* Save Nr */
    st->st_nr = clone_bytes(nonce_pbs.cur, pbs_left(&nonce_pbs),
	"st_nr in main_inR2_outI3()");
    st->st_nr_len = pbs_left(&nonce_pbs);

    /* Generate SKEYID, SKEYID_A, SKEYID_D, SKEYID_E */
    generate_skeyids(st);

    /**************** build output packet HDR*;IDii;HASH_I ****************/
    /* ??? NOTE: this is almost the same as main_inI3_outR3's code */

    /* HDR*  */
    {
	struct isakmp_hdr r_hdr = *hdr;	/* mostly same as incoming header */

	r_hdr.isa_flags |= ISAKMP_FLAG_ENCRYPTION;
	r_hdr.isa_np = ISAKMP_NEXT_ID;
	if (!out_struct(&r_hdr, &isakmp_hdr_desc, reply, &r_hdr_pbs))
	    return STF_FAIL;
    }

    /* IDii */
    {
	struct isakmp_ipsec_id id;
	pb_stream id_pbs;

	id.isaiid_np = ISAKMP_NEXT_HASH;
	id.isaiid_idtype = st->st_myidentity_type;
	id.isaiid_protoid = 0;	/* ??? is this right?  IPPROTO_UDP? */
	id.isaiid_port = 0;	/* ??? is this right?  our_port? */
	if (!out_struct(&id, &isakmp_ipsec_identification_desc, &r_hdr_pbs, &id_pbs)
	|| !out_raw(st->st_myidentity, st->st_myidentity_len, &id_pbs, "my identity"))
	    return STF_FAIL;
	close_output_pbs(&id_pbs);
    }

    /* HASH_I */
    {
	u_char hash_val[MAX_DIGEST_LEN];
	int hash_len;
	pb_stream hash_pbs;
	bool r;

	main_mode_hash(st, hash_val, &hash_len, TRUE, TRUE);
	DBG_cond_dump(DBG_CRYPT, "HASH_I sent:", hash_val, hash_len);

	if (!out_generic(ISAKMP_NEXT_NONE, &isakmp_hash_desc, &r_hdr_pbs, &hash_pbs))
	    return STF_FAIL;
	r = out_raw(hash_val, hash_len, &hash_pbs, "HASH_I");
	if (!r)
	    return STF_FAIL;
	close_output_pbs(&hash_pbs);
    }

    /* encrypt message, except for fixed part of header */

    if (!encrypt_message(&r_hdr_pbs, st))
	return STF_FAIL;	/* ??? we may be partly committed */

    /* finish message */
    close_output_pbs(&r_hdr_pbs);

    /* Advance state */
    st->st_state = OAKLEY_MAIN_I_3;

    return STF_REPLY;
}

/* Compute HASH(1), HASH(2) of Quick Mode.
 * HASH(1) is part of Quick I1 message.
 * HASH(2) is part of Quick R1 message.
 * Used by: quick_outI1, quick_inI1_outR1 (twice), quick_inR1_outI2
 * (see draft-ietf-ipsec-isakmp-oakley-07.txt 5.5)
 */
static void
quick_mode_hash12(
    u_char *dest, size_t *dest_len,
    const u_char *start, const u_char *roof,
    const struct state *st, bool hash2)
{
    struct hmac_ctx ctx;

    hmac_init(&ctx, st->st_oakley.hasher, st->st_skeyid_a, st->st_skeyid_a_len);
    hmac_update(&ctx, (u_char *) &st->st_msgid, sizeof(st->st_msgid));
    if (hash2)
    {
	/* include Ni_b in the hash */
	hmac_update(&ctx, st->st_ni, st->st_ni_len);
    }
    hmac_update(&ctx, start, roof-start);
    hmac_final(dest, &ctx);

    if (dest_len != NULL)
	*dest_len = ctx.hmac_digest_len;

    DBG(DBG_CRYPT,
	DBG_log("HASH(%d) computed:", hash2 + 1);
	DBG_dump("", dest, ctx.hmac_digest_len));
}

void
init_phase2_iv(struct state *st)
{
    const struct hash_desc *h = st->st_oakley.hasher;
    union hash_ctx ctx;

    st->st_iv_len = h->hash_digest_len;
    st->st_iv = alloc_bytes(st->st_iv_len, "initial phase2 IV");

    h->hash_init(&ctx);
    h->hash_update(&ctx, st->st_phase1->st_lastblock, st->st_phase1->st_lastblock_len);
    h->hash_update(&ctx, (u_char *) &st->st_msgid, sizeof(st->st_msgid));
    h->hash_final(st->st_iv, &ctx);

    DBG_cond_dump(DBG_CRYPT, "computed Phase 2 IV:", st->st_iv, st->st_iv_len);
}

/* Initiate quick mode.
 * --> HDR*, HASH(1), SA, Nr [, KE ] [, IDci, IDcr ]
 * (see draft-ietf-ipsec-isakmp-oakley-07.txt 5.5)
 */
static void
quick_outI1(int sock, struct state *st,
    struct sockaddr_in our_client_net, struct sockaddr_in our_client_mask,	/* our user net addr, mask */
    struct sockaddr_in peer_client_net, struct sockaddr_in peer_client_mask)	/* peer's user net addr, mask */
{
    struct state *nst = duplicate_state(st);
    u_char space[8192];	/* NOTE: we assume 8192 is big enough to build the packet */
    pb_stream reply_pbs;	/* not really a reply */
    pb_stream r_hdr_pbs;
    u_char
	*r_hashval,	/* where in reply to jam hash value */
	*r_hash_start;	/* start of what is to be hashed */
    bool self_client = our_client_net.sin_addr.s_addr == 0;

    /* set up reply_pbs */
    init_pbs(&reply_pbs, space, sizeof(space), "reply packet");

    if (self_client)
    {
	/* note: parallels a test in quick_inR1_outI2 */
	mksin(nst->st_myuser, ((struct in_addr *)nst->st_myidentity)->s_addr, 0);
	nst->st_myusernetmask = mask32;
	mksin(nst->st_peeruser, ((struct in_addr *)nst->st_peeridentity)->s_addr, 0);
	nst->st_peerusernetmask = mask32;
    }
    else
    {
	nst->st_myuser = our_client_net;
	nst->st_myusernetmask = our_client_mask;
	nst->st_peeruser = peer_client_net;
	nst->st_peerusernetmask = peer_client_mask;
    }

    nst->st_myuserprotoid = 0;
    nst->st_myuserport = 0;

    nst->st_peeruserprotoid = 0;
    nst->st_peeruserport = 0;

    nst->st_msgid = oakley_get_msgid(nst->st_peer);
    nst->st_state = OAKLEY_QUICK_I_1;

    /* Don't move this before the oakley_get_msgid() call */
    insert_state(nst);

    /* HDR* */
    {
	struct isakmp_hdr hdr;

	hdr.isa_version = ISAKMP_MAJOR_VERSION << ISA_MAJ_SHIFT | ISAKMP_MINOR_VERSION;
	hdr.isa_np = ISAKMP_NEXT_HASH;
	hdr.isa_xchg = ISAKMP_XCHG_QUICK;
	hdr.isa_msgid = nst->st_msgid;
	hdr.isa_flags = ISAKMP_FLAG_ENCRYPTION;
	memcpy(hdr.isa_icookie, nst->st_icookie, COOKIE_SIZE);
	memcpy(hdr.isa_rcookie, nst->st_rcookie, COOKIE_SIZE);
	if (!out_struct(&hdr, &isakmp_hdr_desc, &reply_pbs, &r_hdr_pbs))
	    assert(FALSE);
    }

    /* HASH(1) -- space to be filled later */
    {
	pb_stream hash_pbs;

	if (!out_generic(ISAKMP_NEXT_SA, &isakmp_hash_desc, &r_hdr_pbs, &hash_pbs))
	    assert(FALSE);
	r_hashval = hash_pbs.cur;	/* remember where to plant value */
	if (!out_zero(nst->st_oakley.hasher->hash_digest_len, &hash_pbs, "HASH(1)"))
	    assert(FALSE);
	close_output_pbs(&hash_pbs);
	r_hash_start = r_hdr_pbs.cur;	/* hash from after HASH(1) */
    }

    /* SA */
    if (!out_sa(&r_hdr_pbs
    , &ipsec_sadb[nst->st_goal & (GOAL_ENCRYPT | GOAL_AUTHENTICATE | GOAL_TUNNEL)]
    , nst, FALSE, ISAKMP_NEXT_NONCE))
	assert(FALSE);

    /* Ni */
    {
	pb_stream nonce_pbs;
	u_int8_t nonce_bytes[DEFAULTNONCESIZE];

	/* next payload depends on whether KE or ID payloads will be sent.
	 * ??? But we don't do KE at the moment.
	 */
	if (!out_generic(self_client? ISAKMP_NEXT_NONE : ISAKMP_NEXT_ID
	, &isakmp_nonce_desc, &r_hdr_pbs, &nonce_pbs))
	    assert(FALSE);
	get_rnd_bytes(nonce_bytes, DEFAULTNONCESIZE);
	if (!out_raw(nonce_bytes, DEFAULTNONCESIZE, &nonce_pbs, "nonce value"))
	    assert(FALSE);
	close_output_pbs(&nonce_pbs);

	/* Copy Ni into the state object */
	nst->st_ni = clone_bytes(nonce_bytes, DEFAULTNONCESIZE,
	    "st_ni in quick_outI1()");
	nst->st_ni_len = DEFAULTNONCESIZE;

	DBG_cond_dump(DBG_CRYPT, "Ni sent:", nonce_bytes, DEFAULTNONCESIZE);
    }

    /* [ KE ] if we knew how to do this (needed for PFS) */

    /* [ IDci, IDcr ] */
    if (!self_client)
    {
	struct isakmp_ipsec_id id;
	pb_stream id_pbs;

	/* IDci (we are initiator) */
	id.isaiid_np = ISAKMP_NEXT_ID;
	id.isaiid_idtype = ID_IPV4_ADDR_SUBNET;
	id.isaiid_protoid = nst->st_myuserprotoid;
	id.isaiid_port = nst->st_myuserport;

	if (!out_struct(&id, &isakmp_ipsec_identification_desc, &r_hdr_pbs, &id_pbs))
	    assert(FALSE);
	if (!out_raw(&nst->st_myuser.sin_addr.s_addr, sizeof(nst->st_myuser.sin_addr)
	, &id_pbs, "initiator's client network"))
	    assert(FALSE);
	if (!out_raw(&nst->st_myusernetmask.sin_addr.s_addr, sizeof(nst->st_myusernetmask.sin_addr)
	, &id_pbs, "initiator's client mask"))
	    assert(FALSE);
	close_output_pbs(&id_pbs);

	/* IDcr (peer is responder) */
	id.isaiid_np = ISAKMP_NEXT_NONE;
	id.isaiid_idtype = ID_IPV4_ADDR_SUBNET;
	id.isaiid_protoid = nst->st_peeruserprotoid;
	id.isaiid_port = nst->st_peeruserport;
	if (!out_struct(&id, &isakmp_ipsec_identification_desc, &r_hdr_pbs, &id_pbs))
	    assert(FALSE);
	if (!out_raw(&nst->st_peeruser.sin_addr.s_addr, sizeof(nst->st_peeruser.sin_addr)
	, &id_pbs, "peer's client network"))
	    assert(FALSE);
	if (!out_raw(&nst->st_peerusernetmask.sin_addr.s_addr, sizeof(nst->st_peerusernetmask.sin_addr)
	, &id_pbs, "peer's client mask"))
	    assert(FALSE);
	close_output_pbs(&id_pbs);
    }

    /* finish computing HASH(1) */
    quick_mode_hash12(r_hashval, NULL, r_hash_start, r_hdr_pbs.cur, nst, FALSE);

    init_phase2_iv(nst);

    /* encrypt message, except for fixed part of header */
    if (!encrypt_message(&r_hdr_pbs, nst))
	assert(FALSE);

    close_output_pbs(&r_hdr_pbs);

    /* save packet, now that we know its size */
    nst->st_packet = clone_bytes(reply_pbs.start, pbs_offset(&reply_pbs),
	"reply packet from quick_outI1");
    nst->st_packet_len = pbs_offset(&reply_pbs);

    /* send the packet */

    DBG_cond_dump(DBG_RAW, "sending:\n", nst->st_packet, nst->st_packet_len);

    if (sendto(sock, nst->st_packet, nst->st_packet_len, 0, &nst->st_peer,
	       sizeof(nst->st_peer)) != nst->st_packet_len)
	log_errno("sendto() failed in quick_outI1()");

    event_schedule(EVENT_RETRANSMIT, EVENT_RETRANSMIT_DELAY, nst, 0);
}

/* Decode the ID payload of Phase 1 (main_inI3_outR3 and main_inR3) */
static bool
decode_peer_id(
    struct isakmp_ipsec_id *id,
    pb_stream *message,
    struct state *st,
    const struct sockaddr *sap)
{
    pb_stream id_pbs;

    if (!in_struct(id, &isakmp_ipsec_identification_desc, message, &id_pbs))
	return FALSE;

    /* XXX Check for valid ID types? */
    switch (id->isaiid_idtype)
    {
    case ID_IPV4_ADDR:
	if (id->isaiid_port != 0 || id->isaiid_protoid != 0)
	{
	    /* I think that this is a bug, but it is common */
	    log("port or protocol id in ID not zero (%d/%d)",
		id->isaiid_port, id->isaiid_protoid);
	}
	if (pbs_left(&id_pbs) != sizeof(struct in_addr))
	{
	    log("size of ID_IPV4_ADDR identification should be %u but is %d in ID payload from %s",
		(unsigned) sizeof(struct in_addr),
		(unsigned) pbs_left(&id_pbs),
		show_sa(sap));
	    /* XXX Could send notification back */
	    return FALSE;
	}
	st->st_peeridentity_protocol = id->isaiid_protoid;
	st->st_peeridentity_port = id->isaiid_port;
	st->st_peeridentity_len = pbs_left(&id_pbs);
	st->st_peeridentity = clone_bytes(id_pbs.cur, st->st_peeridentity_len,
	    "st_peeridentity in decode_peer_id()");

	DBG(DBG_PARSING,
	    DBG_log("Peer's ID type is %s: %s",
		enum_show(&ident_desc, id->isaiid_idtype),
		inet_ntoa(*(struct in_addr *)st->st_peeridentity)));
	break;

    default:
	/* XXX Could send notification back */
	log("Unacceptable identity type (%s) in ID payload from %s",
	    enum_show(&ident_desc, id->isaiid_idtype),
	    show_sa(sap));
	return FALSE;
    }
    st->st_peeridentity_type = id->isaiid_idtype;
    return TRUE;
}

/*
 * Handle HDR*;IDir;HASH_R from responder.
 * It then calls quick_outI1() if there is a goal.
 */
stf_status
main_inR3(
    int sock,	/* socket for packet input/output */
    const struct sockaddr *sap,	/* network address */
    pb_stream *message,	/* message to be processed */
    const struct isakmp_hdr *hdr,	/* message's header */
    pb_stream *reply,	/* room for reply */
    struct state *st)	/* current state object */
{
    /* input code stolen from main_inI3_outR3 */
    struct isakmp_ipsec_id id;
    struct isakmp_generic hash;

    if (hdr->isa_np != ISAKMP_NEXT_ID)
    {
	log("expected ISAKMP_NEXT_ID first in main R3 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, hdr->isa_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    /* IDir */
    if (!decode_peer_id(&id, message, st, sap))
	return STF_FAIL;

    /* HASH_R */

    if (id.isaiid_np != ISAKMP_NEXT_HASH)
    {
	log("expected ISAKMP_NEXT_HASH second in main R3 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, id.isaiid_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    {
	pb_stream hash_pbs;
	u_char hash_val[MAX_DIGEST_LEN];
	int hash_len;

	if (!in_struct(&hash, &isakmp_hash_desc, message, &hash_pbs))
	    return STF_FAIL;

	main_mode_hash(st, hash_val, &hash_len, FALSE, FALSE);

	DBG(DBG_CRYPT,
	    DBG_dump("computed HASH_R:", hash_val, hash_len);
	    DBG_dump("received HASH_R:", hash_pbs.cur, pbs_left(&hash_pbs)));

	if (pbs_left(&hash_pbs) != hash_len
	|| memcmp(hash_pbs.cur, hash_val, hash_len) != 0)
	{
	    log("received HASH_R does not match computed value in main R3 from %s",
		show_sa(sap));
	    /* XXX Could send notification back */
	    return STF_FAIL;
	}
    }

    if (hash.isag_np != ISAKMP_NEXT_NONE)
    {
	log("expected ISAKMP_NEXT_NONE third in main R3 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, hash.isag_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    /**************** done input ****************/

    /* Advance state (may be overridden by quick_outI1()) */
    st->st_state = OAKLEY_MAIN_I_4;

    if (st->st_goal != 0)
    {
	DBG(DBG_CONTROL, DBG_log("Doing Quick Mode with %s, goal %s",
	    show_sa(sap),
	    bitnamesof(sa_goal_bit_names, st->st_goal)));
	quick_outI1(sock, st, st->st_myuser, st->st_myusernetmask,
		       st->st_peeruser, st->st_peerusernetmask);
    }
    return STF_NO_REPLY;
}

/* Handle HDR;KE;Ni from Initiator.  Send a HDR;KE;Nr back.
 */
stf_status
main_inI2_outR2(
    int sock,	/* socket for packet input/output */
    const struct sockaddr *sap,	/* network address */
    pb_stream *message,	/* message to be processed */
    const struct isakmp_hdr *hdr,	/* message's header */
    pb_stream *reply,	/* room for reply */
    struct state *st)	/* current state object */
{
    struct isakmp_generic keyex;
    pb_stream keyex_pbs;
    struct isakmp_generic nonce;
    pb_stream nonce_pbs;
    pb_stream r_hdr_pbs;    /* reply message pbs */

    /* KE */

    if (hdr->isa_np != ISAKMP_NEXT_KE)
    {
	log("expected ISAKMP_NEXT_KE first in main I2 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, hdr->isa_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    if (!in_struct(&keyex, &isakmp_keyex_desc, message, &keyex_pbs))
	return STF_FAIL;

    if (pbs_left(&keyex_pbs) < MINIMUM_PUBLIC_VALUE_SIZE)
    {
	log("too short public size (%d) sent from %s",
	    pbs_left(&keyex_pbs), show_sa(sap));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    /* Ni */

    if (keyex.isag_np != ISAKMP_NEXT_NONCE)
    {
	log("expected ISAKMP_NEXT_NONCE second in main I2 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, keyex.isag_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    if (!in_struct(&nonce, &isakmp_nonce_desc, message, &nonce_pbs))
	return STF_FAIL;

    if (nonce.isag_np != ISAKMP_NEXT_NONE)
    {
	log("expected ISAKMP_NEXT_NONE third in main I2 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, nonce.isag_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    /* done parsing */

    /* Decode Gi from network form */
    st->st_gi_in_use = TRUE;
    n_to_mpz(&st->st_gi, keyex_pbs.cur, pbs_left(&keyex_pbs));

    DBG(DBG_CRYPT,
	DBG_dump_mpz("public value received:", &st->st_gi));

    /* st->st_sec := localsecret */
    {
	u_char tmp[LOCALSECRETSIZE];

	get_rnd_bytes(tmp, LOCALSECRETSIZE);
	st->st_sec_in_use = TRUE;
	n_to_mpz(&st->st_sec, tmp, LOCALSECRETSIZE);
    }

    /* st->st_gr := groupgenerator ** st->st_sec  %  st->st_oakley.modp_modulus */
    mpz_init(&st->st_gr);
    st->st_gr_in_use = TRUE;
    mpz_powm(&st->st_gr, &groupgenerator, &st->st_sec, st->st_oakley.modp_modulus);

    /* st->st_shared := st->st_gi ** st->st_sec  %  st->st_oakley.modp_modulus */
    mpz_init(&st->st_shared);
    st->st_shared_in_use = TRUE;
    mpz_powm(&st->st_shared, &st->st_gi, &st->st_sec, st->st_oakley.modp_modulus);

    DBG(DBG_CRYPT,
	DBG_dump_mpz("our secret value:", &st->st_sec);
	DBG_dump_mpz("our public value:", &st->st_gr);
	DBG_dump_mpz("shared secret:", &st->st_shared));

    /* Save Ni */
    st->st_ni = clone_bytes(nonce_pbs.cur, pbs_left(&nonce_pbs),
	"st_ni in main_inI2_outR2()");
    st->st_ni_len = pbs_left(&nonce_pbs);

    /**************** build output packet HDR;KE;Nr ****************/

    /* HDR  */
    {
	struct isakmp_hdr r_hdr = *hdr;	/* mostly same as incoming header */

	r_hdr.isa_np = ISAKMP_NEXT_KE;
	if (!out_struct(&r_hdr, &isakmp_hdr_desc, reply, &r_hdr_pbs))
	    return STF_FAIL;
    }

    /* KE */
    {
	pb_stream keyex_pbs;
	u_char *gr;
	size_t grlen;
	bool r;

	if (!out_generic(ISAKMP_NEXT_NONCE, &isakmp_keyex_desc, &r_hdr_pbs, &keyex_pbs))
	    return STF_FAIL;

	mpz_to_n(&gr, &grlen, &st->st_gr);
	r = out_raw(gr, grlen, &keyex_pbs, "keyex value");
	free(gr);

	if (!r)
	    return STF_FAIL;
	close_output_pbs(&keyex_pbs);
    }

    /* Nr */
    {
	pb_stream nonce_pbs;
	u_int8_t nonce_bytes[DEFAULTNONCESIZE];

	if (!out_generic(ISAKMP_NEXT_NONE, &isakmp_nonce_desc, &r_hdr_pbs, &nonce_pbs))
	    return STF_FAIL;
	get_rnd_bytes(nonce_bytes, DEFAULTNONCESIZE);
	if (!out_raw(nonce_bytes, DEFAULTNONCESIZE, &nonce_pbs, "nonce value"))
	    return STF_FAIL;
	close_output_pbs(&nonce_pbs);

	/* Copy Nr into the state object */
	st->st_nr = clone_bytes(nonce_bytes, DEFAULTNONCESIZE,
	    "st_nr in main_inR1_outI2()");
	st->st_nr_len = DEFAULTNONCESIZE;
    }

    /* finish message */
    close_output_pbs(&r_hdr_pbs);

    generate_skeyids(st);

    /* Advance state */
    st->st_state = OAKLEY_MAIN_R_2;

    return STF_REPLY;
}

/*
 * Handle HDR*;IDii;HASH_I from initiator. Send a HDR*;IDir;HASH_R back.
 */
stf_status
main_inI3_outR3(
    int sock,	/* socket for packet input/output */
    const struct sockaddr *sap,	/* network address */
    pb_stream *message,	/* message to be processed */
    const struct isakmp_hdr *hdr,	/* message's header */
    pb_stream *reply,	/* room for reply */
    struct state *st)	/* current state object */
{
    struct isakmp_ipsec_id id;
    struct isakmp_generic hash;
    pb_stream r_hdr_pbs;

    if (hdr->isa_np != ISAKMP_NEXT_ID)
    {
	log("expected ISAKMP_NEXT_ID first in main I3 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, hdr->isa_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    /* IDii */
    if (!decode_peer_id(&id, message, st, sap))
	return STF_FAIL;

    /* HASH_I */

    if (id.isaiid_np != ISAKMP_NEXT_HASH)
    {
	log("expected ISAKMP_NEXT_HASH second in main I3 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, id.isaiid_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    {
	pb_stream hash_pbs;
	u_char hash_val[MAX_DIGEST_LEN];
	int hash_len;

	if (!in_struct(&hash, &isakmp_hash_desc, message, &hash_pbs))
	    return STF_FAIL;

	main_mode_hash(st, hash_val, &hash_len, TRUE, FALSE);

	DBG(DBG_CRYPT,
	    DBG_dump("computed HASH_I:", hash_val, hash_len);
	    DBG_dump("received HASH_I:", hash_pbs.cur, pbs_left(&hash_pbs)));

	if (pbs_left(&hash_pbs) != hash_len
	|| memcmp(hash_pbs.cur, hash_val, hash_len) != 0)
	{
	    log("received HASH_I does not match computed value in main I3 from %s",
		show_sa(sap));
	    /* XXX Could send notification back */
	    return STF_FAIL;
	}
    }

    if (hash.isag_np != ISAKMP_NEXT_NONE)
    {
	log("expected ISAKMP_NEXT_NONE third in main I3 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, hash.isag_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    /**************** build output packet HDR*;IDir;HASH_R ****************/
    /* ??? NOTE: this is almost the same as main_inR2_outI3's code */

    /* HDR*  */
    {
	struct isakmp_hdr r_hdr = *hdr;	/* mostly same as incoming header */

	r_hdr.isa_flags |= ISAKMP_FLAG_ENCRYPTION;
	r_hdr.isa_np = ISAKMP_NEXT_ID;
	if (!out_struct(&r_hdr, &isakmp_hdr_desc, reply, &r_hdr_pbs))
	    return STF_FAIL;
    }

    /* IDii */
    {
	struct isakmp_ipsec_id id;
	pb_stream id_pbs;

	id.isaiid_np = ISAKMP_NEXT_HASH;
	id.isaiid_idtype = st->st_myidentity_type;
	id.isaiid_protoid = 0;	/* ??? is this right? */
	id.isaiid_port = 0;	/* ??? is this right? */
	if (!out_struct(&id, &isakmp_ipsec_identification_desc, &r_hdr_pbs, &id_pbs)
	|| !out_raw(st->st_myidentity, st->st_myidentity_len, &id_pbs, "my identity"))
	    return STF_FAIL;
	close_output_pbs(&id_pbs);
    }

    /* HASH_I */
    {
	u_char hash_val[MAX_DIGEST_LEN];
	int hash_len;
	pb_stream hash_pbs;
	bool r;

	main_mode_hash(st, hash_val, &hash_len, FALSE, TRUE);
	DBG_cond_dump(DBG_CRYPT, "HASH_R sent:", hash_val, hash_len);

	if (!out_generic(ISAKMP_NEXT_NONE, &isakmp_hash_desc, &r_hdr_pbs, &hash_pbs))
	    return STF_FAIL;
	r = out_raw(hash_val, hash_len, &hash_pbs, "HASH_R");
	if (!r)
	    return STF_FAIL;
	close_output_pbs(&hash_pbs);
    }

    /* encrypt message, sans fixed part of header */
    if (!encrypt_message(&r_hdr_pbs, st))
	return STF_FAIL;	/* ??? we may be partly committed */

    /* Last block of Phase 1 (R3), kept for Phase 2 IV generation */
    /* ??? how does this differ from st_iv? */
    st->st_lastblock_len = st->st_iv_len;
    st->st_lastblock = clone_bytes(st->st_iv, st->st_iv_len,
	"st_lastblock in main_inI3_outR3()");
    DBG_cond_dump(DBG_CRYPT, "last encrypted block of Phase 1:", st->st_lastblock, st->st_lastblock_len);

    /* finish message */
    close_output_pbs(&r_hdr_pbs);

    /* Advance state */
    st->st_state = OAKLEY_MAIN_R_3;

    return STF_REPLY;
}

/* Decode the variable part of an ID packet in (during Quick Mode).
 * This is designed for packets that identify clients, not peers.
 * Currently this will only accept two forms.
 */
static bool
decode_net_id(
    struct isakmp_ipsec_id *id,
    pb_stream *id_pbs,
    struct sockaddr_in *net, struct sockaddr_in *mask,
    const char *which, const struct sockaddr *sap)
{
    struct sockaddr_in t;	/* temp in which to build proper sockaddr_in */

    mksin(t, 0, 0);
    switch (id->isaiid_idtype)
    {
	case ID_IPV4_ADDR:
	    if (pbs_left(id_pbs) != sizeof(net->sin_addr))
	    {
		log("%s ID payload ID_IPV4_ADDR wrong length in Quick I1 from %s",
		    which, show_sa(sap));
		/* XXX Could send notification back */
		return FALSE;
	    }
	    memcpy(&t.sin_addr, id_pbs->cur, sizeof(t.sin_addr));
	    *net = t;
	    *mask = mask32;
	    DBG(DBG_PARSING | DBG_CONTROL,
		DBG_log("%s user is IP address %s", which, get_address_in(*net)));
	    break;

	case ID_IPV4_ADDR_SUBNET:
	    if (pbs_left(id_pbs) != sizeof(net->sin_addr) + sizeof(mask->sin_addr))
	    {
		log("%s ID payload ID_IPV4_ADDR_SUBNET wrong length in Quick I1 from %s",
		    which, show_sa(sap));
		/* XXX Could send notification back */
		return FALSE;
	    }
	    memcpy(&t.sin_addr, id_pbs->cur, sizeof(t.sin_addr));
	    *net = t;
	    memcpy(&t.sin_addr, id_pbs->cur + sizeof(t.sin_addr), sizeof(t.sin_addr));
	    *mask = t;
	    DBG(DBG_PARSING | DBG_CONTROL,
		DBG_log("%s user is IP subnet with address %s...",
		    which, get_address_in(*net));
		DBG_log("...and netmask %s",
		    get_address_in(*mask)));
	    break;

	default:
	    /* XXX support more */
	    log("bad id type %s from %s",
		enum_show(&ident_desc, id->isaiid_idtype), show_sa(sap));
	    /* XXX Could send notification back */
	    return FALSE;
    }
    return TRUE;
}

/* like decode, but checks that what is received matches what was sent */
static bool
check_net_id(
    struct isakmp_ipsec_id *id,
    pb_stream *id_pbs,
    u_int8_t *protoid, u_int16_t *port,
    struct sockaddr_in *net, struct sockaddr_in *mask,
    const char *which, const struct sockaddr *sap)
{
    struct sockaddr_in net_temp;
    struct sockaddr_in mask_temp;

    if (!decode_net_id(id, id_pbs, &net_temp, &mask_temp, which, sap))
	return FALSE;

    if (net->sin_addr.s_addr != net_temp.sin_addr.s_addr
    || mask->sin_addr.s_addr != mask_temp.sin_addr.s_addr
    || *protoid != id->isaiid_protoid || *port != id->isaiid_port)
    {
	log("%s ID returned by %s doesn't match my proposal",
	    which, show_sa(sap));
	return FALSE;
    }
    return TRUE;
}

/* Handle first message of Phase 2 -- Quick Mode.
 * HDR*, HASH(1), SA, Ni [, KE ] [, IDci, IDcr ] -->
 * HDR*, HASH(2), SA, Nr [, KE ] [, IDci, IDcr ]
 * (see draft-ietf-ipsec-isakmp-oakley-07.txt 5.5)
 */
stf_status
quick_inI1_outR1(
    int sock,	/* socket for packet input/output */
    const struct sockaddr *sap,	/* network address */
    pb_stream *message,	/* message to be processed */
    const struct isakmp_hdr *hdr,	/* message's header */
    pb_stream *reply,	/* room for reply */
    struct state *st)	/* current state object */
{
    /* we build reply packet as we parse the message since
     * the parse_ipsec_sa_body emits the reply SA
     */

    struct isakmp_generic hash;
    struct isakmp_sa sa;

    struct isakmp_generic nonce;
    pb_stream nonce_pbs;

    u_int8_t np;

    bool ids_used = FALSE;
    struct isakmp_ipsec_id IDci, IDcr;
    pb_stream IDci_pbs, IDcr_pbs;

    pb_stream r_hdr_pbs;    /* reply header */

    u_char
	*r_hashval,	/* where in reply to jam hash value */
	*r_hash_start;	/* from where to start hashing */


    /* emit HDR* */
    {
	struct isakmp_hdr r_hdr = *hdr;	/* mostly same as incoming header */

	r_hdr.isa_flags |= ISAKMP_FLAG_ENCRYPTION;
	r_hdr.isa_np = ISAKMP_NEXT_HASH;
	if (!out_struct(&r_hdr, &isakmp_hdr_desc, reply, &r_hdr_pbs))
	    return STF_FAIL;
    }

    /* accept HASH(1) */
    {
	pb_stream hash_pbs;
	u_char hashval[MAX_DIGEST_LEN];
	size_t hashlen;

	/* hdr.isa_np already checked as ISAKMP_NEXT_HASH */

	if (!in_struct(&hash, &isakmp_hash_desc, message, &hash_pbs))
	    return STF_FAIL;
	DBG_cond_dump(DBG_CRYPT, "received HASH(1):", hash_pbs.cur, pbs_left(&hash_pbs));

	quick_mode_hash12(hashval, &hashlen, message->cur, message->roof, st, FALSE);

	if (hashlen != pbs_left(&hash_pbs)
	|| memcmp(hashval, hash_pbs.cur, hashlen) != 0)
	{
	    log("HASH(1) from %s does not verify", show_sa(sap));
	    /* XXX Could send notification back */
	    return STF_FAIL;
	}
	DBG(DBG_CRYPT, DBG_log("HASH(1) verified"));
    }

    /* emit HASH(2) -- first pass */
    {
	pb_stream hash_pbs;

	if (!out_generic(ISAKMP_NEXT_SA, &isakmp_hash_desc, &r_hdr_pbs, &hash_pbs))
	    return STF_FAIL;
	r_hashval = hash_pbs.cur;	/* remember where to plant value */
	if (!out_zero(st->st_oakley.hasher->hash_digest_len, &hash_pbs, "HASH(2)"))
	    return STF_FAIL;
	close_output_pbs(&hash_pbs);
	r_hash_start = r_hdr_pbs.cur;	/* hash from after HASH*/
    }

    /* check that SA Payload comes after Hash Payload */

    if (hash.isag_np != ISAKMP_NEXT_SA)
    {
	log("expected ISAKMP_NEXT_SA second in Quick I1 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, hash.isag_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    /* process SA (in and out) */
    {
	pb_stream sa_pbs;
	pb_stream r_sa_pbs;

	/* accept header */
	if (!in_struct(&sa, &isakmp_sa_desc, message, &sa_pbs))
	    return STF_FAIL;

	/* sa header is unchanged -- even np is the same */
	if (!out_struct(&sa, &isakmp_sa_desc, &r_hdr_pbs, &r_sa_pbs))
	    assert(FALSE);

	/* parse and accept body */
	if (!parse_ipsec_sa_body(&sa_pbs, &sa, &r_sa_pbs, FALSE, st, sap))
	    return STF_FAIL;

	/* ??? don't actually set st until whole packet accepted */

	/* Set goal to match what peer does. */
	/* ??? Is goal used? */
	/* ??? what exactly is the rule for tunneling? */

	if (st->st_ah.present)
	{
	    st->st_goal |= GOAL_AUTHENTICATE;
	    if (st->st_ah.attrs.encapsulation == ENCAPSULATION_MODE_TUNNEL)
		st->st_goal |= GOAL_TUNNEL;
	}

	if (st->st_esp.present)
	{
	    st->st_goal |= GOAL_ENCRYPT;
	    if (st->st_esp.attrs.encapsulation == ENCAPSULATION_MODE_TUNNEL)
		st->st_goal |= GOAL_TUNNEL;
	}
    }

    /* accept Ni payload */

    if (sa.isasa_np != ISAKMP_NEXT_NONCE)
    {
	log("expected ISAKMP_NEXT_NONCE third in Quick I1 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, sa.isasa_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    if (!in_struct(&nonce, &isakmp_nonce_desc, message, &nonce_pbs))
	return STF_FAIL;

    /* XXX accept optional KE payload -- needed for PFS */

    np = nonce.isag_np;

    if (np == ISAKMP_NEXT_KE)
    {
	log("we don't yet support KE in Quick Mode");
	return STF_FAIL;
    }

    /* accept optional IDci, IDcr */

    if (np == ISAKMP_NEXT_ID)
    {
	/* ??? we are assuming IPSEC_DOI */
	ids_used = TRUE;

	/* IDci (initiator is peer) */

	if (!in_struct(&IDci, &isakmp_ipsec_identification_desc, message, &IDci_pbs))
	    return STF_FAIL;

	if (IDci.isaiid_np != ISAKMP_NEXT_ID)
	{
	    log("IDci must be followed by IDcr in Quick I1.  Found %s from %s",
		enum_show(&payload_desc, IDci.isaiid_np), show_sa(sap));
	    /* XXX Could send notification back */
	    return STF_FAIL;
	}

	if (!decode_net_id(&IDci, &IDci_pbs
	, &st->st_peeruser, &st->st_peerusernetmask
	, "peer client", sap))
	    return STF_FAIL;

	st->st_peeruserprotoid = IDcr.isaiid_protoid;
	st->st_peeruserport = IDcr.isaiid_port;

	/* IDcr (we are responder) */

	if (!in_struct(&IDcr, &isakmp_ipsec_identification_desc, message, &IDcr_pbs))
	    return STF_FAIL;

	if (IDcr.isaiid_np != ISAKMP_NEXT_NONE)
	{
	    log("IDcr must be the last payload in Quick I1.  Found %s from %s",
		enum_show(&payload_desc, IDcr.isaiid_np), show_sa(sap));
	    /* XXX Could send notification back */
	    return STF_FAIL;
	}

	if (!decode_net_id(&IDcr, &IDcr_pbs
	, &st->st_myuser, &st->st_myusernetmask
	, "our client", sap))
	    return STF_FAIL;

	st->st_myuserprotoid = IDcr.isaiid_protoid;
	st->st_myuserport = IDcr.isaiid_port;
    }
    else
    {
	/* implicit IDci and IDcr: peer and self */
	mksin(st->st_myuser, ((struct in_addr *)st->st_myidentity)->s_addr, 0);
	st->st_myusernetmask = mask32;
	mksin(st->st_peeruser, ((struct in_addr *)st->st_peeridentity)->s_addr, 0);
	st->st_peerusernetmask = mask32;
    }

    /* copy Ni into the state */
    st->st_ni = clone_bytes(nonce_pbs.cur, pbs_left(&nonce_pbs),
	"st_ni in quick_inI1_outR1()");
    st->st_ni_len = pbs_left(&nonce_pbs);


    /**** finish reply packet: Nr [, KE ] [, IDci, IDcr ] ****/

    /* reply Nr */
    {
	pb_stream nonce_pbs;
	u_int8_t nonce_bytes[DEFAULTNONCESIZE];

	/* next payload depends on whether KE or ID payloads will be sent.
	 * ??? But we don't do KE at the moment.
	 */
	if (!out_generic(ids_used? ISAKMP_NEXT_ID : ISAKMP_NEXT_NONE
	, &isakmp_nonce_desc, &r_hdr_pbs, &nonce_pbs))
	    return STF_FAIL;
	get_rnd_bytes(nonce_bytes, DEFAULTNONCESIZE);
	if (!out_raw(nonce_bytes, DEFAULTNONCESIZE, &nonce_pbs, "nonce value"))
	    return STF_FAIL;
	close_output_pbs(&nonce_pbs);

	/* Copy Nr into the state object */
	st->st_nr = clone_bytes(nonce_bytes, DEFAULTNONCESIZE,
	    "st_nr in quick_inI1_outR1()");
	st->st_nr_len = DEFAULTNONCESIZE;

	DBG(DBG_CRYPT, DBG_dump("Nr sent:", nonce_bytes, DEFAULTNONCESIZE));
    }

    /* reply [, KE ] if we knew how to do this */

    /* reply [, IDci, IDcr ] */
    if  (ids_used)
    {
	struct isakmp_ipsec_id *p = (void *)r_hdr_pbs.cur;	/* UGH! */

	if (!out_raw(IDci_pbs.start, pbs_room(&IDci_pbs), &r_hdr_pbs, "IDci"))
	    return STF_FAIL;
	p->isaiid_np = ISAKMP_NEXT_ID;

	p = (void *)r_hdr_pbs.cur;	/* UGH! */

	if (!out_raw(IDcr_pbs.start, pbs_room(&IDcr_pbs), &r_hdr_pbs, "IDcr"))
	    return STF_FAIL;
	p->isaiid_np = ISAKMP_NEXT_NONE;
    }

    /* Compute reply HASH(2) */
    quick_mode_hash12(r_hashval, NULL, r_hash_start, r_hdr_pbs.cur, st, TRUE);

    /* encrypt message, except for fixed part of header */
    if (!encrypt_message(&r_hdr_pbs, st))
	return STF_FAIL;	/* ??? we may be partly committed */

    close_output_pbs(&r_hdr_pbs);

    /* Update state of exchange */
    st->st_state = OAKLEY_QUICK_R_1;

    return STF_REPLY;
}

/*
 * Produce the new key material of Quick Mode.
 * draft-ietf-ipsec-isakmp-oakley-06.txt section 5.5
 * specifies how this is to be done.
 */
static void
compute_proto_keymat(
    struct state *st,
    u_int8_t protoid,
    struct ipsec_proto_info *pi)
{
    size_t needed_len; /* bytes of keying material needed */

    /* Add up the requirements for keying material
     * (It probably doesn't matter if we produce too much!)
     */
    switch (protoid)
    {
    case PROTO_IPSEC_ESP:
	    switch (pi->attrs.transid)
	    {
	    case ESP_NULL:
		needed_len = 0;
		break;
	    case ESP_DES:
		needed_len = DES_CBC_BLOCK_SIZE;
		break;
	    case ESP_3DES:
		needed_len = DES_CBC_BLOCK_SIZE * 3;
		break;
	    default:
		exit_log("transform %s not implemented yet",
		    enum_show(&esp_transformid_desc, pi->attrs.transid));
	    }

	    switch (pi->attrs.auth)
	    {
	    case AUTH_ALGORITHM_NONE:
		break;
	    case AUTH_ALGORITHM_HMAC_MD5:
		needed_len += HMAC_MD5_KEY_LEN;
		break;
	    case AUTH_ALGORITHM_HMAC_SHA1:
		needed_len += HMAC_SHA1_KEY_LEN;
		break;
	    case AUTH_ALGORITHM_DES_MAC:
	    default:
		exit_log("AUTH algorithm %s not implemented yet",
		    enum_show(&auth_alg_desc, pi->attrs.auth));
	    }
	    break;

    case PROTO_IPSEC_AH:
	    switch (pi->attrs.transid)
	    {
	    case AH_MD5:
		needed_len = HMAC_MD5_KEY_LEN;
		break;
	    case AH_SHA:
		needed_len = HMAC_SHA1_KEY_LEN;
		break;
	    default:
		exit_log("transform %s not implemented yet",
		    enum_show(&ah_transformid_desc, pi->attrs.transid));
	    }
	    break;

    default:
	exit_log("protocol %s not implemented yet",
	    enum_show(&protocol_desc, protoid));
	break;
    }

    pi->keymat_len = needed_len;

    /* Allocate space for the keying material.
     * Although only needed_len bytes are desired, we
     * must round up to a multiple of ctx.hmac_digest_len
     * so that our buffer isn't overrun.
     */
    {
	struct hmac_ctx ctx_me, ctx_peer;
	size_t needed_space;	/* space needed for keying material (rounded up) */
	size_t i;

	/* XXX PFS requires different hasher */
	hmac_init(&ctx_me, st->st_oakley.hasher, st->st_skeyid_d, st->st_skeyid_d_len);
	ctx_peer = ctx_me;	/* duplicate initial conditions */

	needed_space = needed_len + pad_up(needed_len, ctx_me.hmac_digest_len);
	pi->our_keymat = alloc_bytes(needed_space, "keymat in compute_keymat()");
	pi->peer_keymat = alloc_bytes(needed_space, "peer_keymat in quick_inI1_outR1()");

	for (i = 0;; )
	{
	    hmac_update(&ctx_me, &protoid, sizeof(protoid));
	    hmac_update(&ctx_peer, &protoid, sizeof(protoid));

	    hmac_update(&ctx_me, pi->our_spi, IPSEC_DOI_SPI_SIZE);
	    hmac_update(&ctx_peer, pi->attrs.spi, IPSEC_DOI_SPI_SIZE);

	    hmac_update(&ctx_me, st->st_ni, st->st_ni_len);
	    hmac_update(&ctx_peer, st->st_ni, st->st_ni_len);

	    hmac_update(&ctx_me, st->st_nr, st->st_nr_len);
	    hmac_update(&ctx_peer, st->st_nr, st->st_nr_len);

	    hmac_final(pi->our_keymat + i, &ctx_me);
	    hmac_final(pi->peer_keymat + i, &ctx_peer);

	    i += ctx_me.hmac_digest_len;
	    if (i >= needed_space)
		break;

	    /* more keying material needed: prepare to go around again */

	    hmac_reinit(&ctx_me);
	    hmac_reinit(&ctx_peer);

	    hmac_update(&ctx_me, pi->our_keymat + i - ctx_me.hmac_digest_len,
		ctx_me.hmac_digest_len);
	    hmac_update(&ctx_peer, pi->peer_keymat + i - ctx_peer.hmac_digest_len,
		ctx_peer.hmac_digest_len);
	}
    }

    DBG(DBG_CRYPT,
	DBG_dump("KEYMAT computed:\n", pi->our_keymat, pi->keymat_len);
	DBG_dump("Peer KEYMAT computed:\n", pi->peer_keymat, pi->keymat_len));
}

static void
compute_keymats(struct state *st)
{
    if (st->st_ah.present)
	compute_proto_keymat(st, PROTO_IPSEC_AH, &st->st_ah);
    if (st->st_esp.present)
	compute_proto_keymat(st, PROTO_IPSEC_ESP, &st->st_esp);
}

/* Compute HASH(3) in Quick Mode (part of quick I2 message).
 * Used by: quick_inR1_outI2, quick_inI2
 * (see draft-ietf-ipsec-isakmp-oakley-07.txt 5.5)
 */
static void
quick_mode_hash3(u_char *dest, size_t *destlen, struct state *st)
{
    struct hmac_ctx ctx;

    hmac_init(&ctx, st->st_oakley.hasher, st->st_skeyid_a, st->st_skeyid_a_len);
    hmac_update(&ctx, "\0", 1);
    hmac_update(&ctx, (u_char *) &st->st_msgid, sizeof(st->st_msgid));
    hmac_update(&ctx, st->st_ni, st->st_ni_len);
    hmac_update(&ctx, st->st_nr, st->st_nr_len);
    hmac_final(dest, &ctx);

    if (destlen != NULL)
	*destlen = ctx.hmac_digest_len;

    DBG_cond_dump(DBG_CRYPT, "HASH(3) computed:", dest, ctx.hmac_digest_len);
}

/* Handle (the single) message from Responder in Quick Mode.
 * HDR*, HASH(2), SA, Nr [, KE ] [, IDci, IDcr ] -->
 * HDR*, HASH(3)
 * (see draft-ietf-ipsec-isakmp-oakley-07.txt 5.5)
 */
stf_status
quick_inR1_outI2(
    int sock,	/* socket for packet input/output */
    const struct sockaddr *sap,	/* network address */
    pb_stream *message,	/* message to be processed */
    const struct isakmp_hdr *hdr,	/* message's header */
    pb_stream *reply,	/* room for reply */
    struct state *st)	/* current state object */
{
    struct isakmp_generic hash;

    struct isakmp_sa sa;
    pb_stream sa_pbs;

    struct isakmp_generic nonce;
    pb_stream nonce_pbs;

    u_int8_t np;

    struct isakmp_ipsec_id IDci, IDcr;
    pb_stream IDci_pbs, IDcr_pbs;

    pb_stream r_hdr_pbs;    /* reply header */

    /* HASH(2) */
    {
	pb_stream hash_pbs;
	u_char hashval[MAX_DIGEST_LEN];
	size_t hashlen;

	/* hdr.isa_np already checked as ISAKMP_NEXT_HASH */

	if (!in_struct(&hash, &isakmp_hash_desc, message, &hash_pbs))
	    return STF_FAIL;
	DBG_cond_dump(DBG_CRYPT, "received HASH(2):", hash_pbs.cur, pbs_left(&hash_pbs));

	quick_mode_hash12(hashval, &hashlen, message->cur, message->roof, st, TRUE);

	if (hashlen != pbs_left(&hash_pbs)
	|| memcmp(hashval, hash_pbs.cur, hashlen) != 0)
	{
	    log("HASH(2) from %s does not verify", show_sa(sap));
	    /* XXX Could send notification back */
	    return STF_FAIL;
	}
	DBG(DBG_CRYPT, DBG_log("HASH(2) verified"));
    }

    /* check that SA Payload comes after Hash Payload */
    /* ??? should not blindly accept -- should check if we proposed this! */

    if (hash.isag_np != ISAKMP_NEXT_SA)
    {
	log("expected ISAKMP_NEXT_SA second in Quick R1 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, hash.isag_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    {
	if (!in_struct(&sa, &isakmp_sa_desc, message, &sa_pbs))
	    return STF_FAIL;

	if (!parse_ipsec_sa_body(&sa_pbs, &sa, NULL, TRUE, st, sap))
	    return STF_FAIL;
    }

    /* Ni payload */

    if (sa.isasa_np != ISAKMP_NEXT_NONCE)
    {
	log("expected ISAKMP_NEXT_NONCE third in Quick R1 from %s, but found %s",
	    show_sa(sap), enum_show(&payload_desc, sa.isasa_np));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    if (!in_struct(&nonce, &isakmp_nonce_desc, message, &nonce_pbs))
	return STF_FAIL;

    /* XXX Handle optional KE payload -- needed for PFS */

    np = nonce.isag_np;

    if (np == ISAKMP_NEXT_KE)
    {
	log("we don't yet support KE in Quick Mode");
	return STF_FAIL;
    }

    /* handle optional IDci, IDcr; these must match what we sent */

    if (np == ISAKMP_NEXT_ID)
    {
	/* ??? we are assuming IPSEC_DOI */

	/* IDci (we are initiator) */

	if (!in_struct(&IDci, &isakmp_ipsec_identification_desc, message, &IDci_pbs))
	    return STF_FAIL;

	if (IDci.isaiid_np != ISAKMP_NEXT_ID)
	{
	    log("IDci must be followed by IDcr in Quick R1.  Found %s from %s",
		enum_show(&payload_desc, IDci.isaiid_np), show_sa(sap));
	    /* XXX Could send notification back */
	    return STF_FAIL;
	}

	if (!check_net_id(&IDci, &IDci_pbs
	, &st->st_myuserprotoid, &st->st_myuserport
	, &st->st_myuser, &st->st_myusernetmask
	, "our client", sap))
	    return STF_FAIL;

	/* IDcr (responder is peer) */

	if (!in_struct(&IDcr, &isakmp_ipsec_identification_desc, message, &IDcr_pbs))
	    return STF_FAIL;

	if (IDcr.isaiid_np != ISAKMP_NEXT_NONE)
	{
	    log("IDcr must be the last payload in Quick R1.  Found %s from %s",
		enum_show(&payload_desc, IDcr.isaiid_np), show_sa(sap));
	    /* XXX Could send notification back */
	    return STF_FAIL;
	}

	if (!check_net_id(&IDcr, &IDcr_pbs
	, &st->st_peeruserprotoid, &st->st_peeruserport
	, &st->st_peeruser, &st->st_peerusernetmask
	, "peer client", sap))
	    return STF_FAIL;
    }
    else
    {
	/* No IDci, IDcr: we must check that the defaults match our proposal.
	 * Parallels a sequence of assignments in quick_outI1.
	 */
	if (st->st_myuser.sin_addr.s_addr != ((struct in_addr *)st->st_myidentity)->s_addr
	|| st->st_myusernetmask.sin_addr.s_addr != mask32.sin_addr.s_addr
	|| st->st_peeruser.sin_addr.s_addr !=  ((struct in_addr *)st->st_peeridentity)->s_addr
	|| st->st_peerusernetmask.sin_addr.s_addr != mask32.sin_addr.s_addr)
	{
	    log("IDci, IDcr payloads missing in message from %s but default does not match proposal",
		show_sa(sap));
	    return STF_FAIL;
	}
    }

    /* copy Nr into the state */
    st->st_nr = clone_bytes(nonce_pbs.cur, pbs_left(&nonce_pbs),
	"st_nr in quick_inR1_outI2()");
    st->st_nr_len = pbs_left(&nonce_pbs);

    DBG_cond_dump(DBG_CRYPT, "Nr received:", nonce_pbs.cur, pbs_left(&nonce_pbs));

    /* Copy accepted proposal into the state.  The whole payload! */
    st->st_proposal = clone_bytes(sa_pbs.start, pbs_room(&sa_pbs),
	"st_proposal in quick_inR1_outI2()");

    /**************** build reply packet HDR*, HASH(3) ****************/

    /* HDR* */

    {
	struct isakmp_hdr r_hdr = *hdr;	/* mostly same as incoming header */

	r_hdr.isa_flags |= ISAKMP_FLAG_ENCRYPTION;
	r_hdr.isa_np = ISAKMP_NEXT_HASH;
	if (!out_struct(&r_hdr, &isakmp_hdr_desc, reply, &r_hdr_pbs))
	    return STF_FAIL;
    }

    /* HASH(3) -- since this is the only contents, no passes needed */

    {
	pb_stream hash_pbs;
	u_char *r_hashval;	/* where in reply to jam hash value */

	if (!out_generic(ISAKMP_NEXT_NONE, &isakmp_hash_desc, &r_hdr_pbs, &hash_pbs))
	    return STF_FAIL;
	r_hashval = hash_pbs.cur;	/* remember where to plant value */
	if (!out_zero(st->st_oakley.hasher->hash_digest_len, &hash_pbs, "HASH(3)"))
	    return STF_FAIL;
	quick_mode_hash3(r_hashval, NULL, st);
	close_output_pbs(&hash_pbs);
    }

    /* encrypt message, except for fixed part of header */
    if (!encrypt_message(&r_hdr_pbs, st))
	return STF_FAIL;	/* ??? we may be partly committed */

    close_output_pbs(&r_hdr_pbs);

    /* Update state of exchange */
    st->st_state = OAKLEY_QUICK_I_2;

    /* Derive new keying material */
    compute_keymats(st);

    /* XXX Unless the commit bit is set, we need to send a message
     * XXX to the kernel to establish the new SA.
     */
    if (!install_ipsec_sa(st, TRUE))
	return STF_FAIL;

    return STF_REPLY;
}

/* Handle last message of Quick Mode.
 * HDR*, HASH(3) -> done
 * (see draft-ietf-ipsec-isakmp-oakley-07.txt 5.5)
 */
stf_status
quick_inI2(
    int sock,	/* socket for packet input/output */
    const struct sockaddr *sap,	/* network address */
    pb_stream *message,	/* message to be processed */
    const struct isakmp_hdr *hdr,	/* message's header */
    pb_stream *reply,	/* room for reply */
    struct state *st)	/* current state object */
{
    struct isakmp_generic hash;

    /* HASH(3) */
    {
	pb_stream hash_pbs;
	u_char hashval[MAX_DIGEST_LEN];
	size_t hashlen;

	/* hdr.isa_np already checked as ISAKMP_NEXT_HASH */

	if (!in_struct(&hash, &isakmp_hash_desc, message, &hash_pbs))
	    return STF_FAIL;
	DBG_cond_dump(DBG_CRYPT, "received HASH(3):", hash_pbs.cur, pbs_left(&hash_pbs));

	quick_mode_hash3(hashval, &hashlen, st);

	if (hashlen != pbs_left(&hash_pbs)
	|| memcmp(hashval, hash_pbs.cur, hashlen) != 0)
	{
	    log("HASH(3) from %s does not verify", show_sa(sap));
	    /* XXX Could send notification back */
	    return STF_FAIL;
	}
	DBG(DBG_CRYPT, DBG_log("HASH(3) authentication succeeded"));
    }

    if (hash.isag_np != ISAKMP_NEXT_NONE)
    {
	log("HASH(3) must be the only payload in Quick I2.  Found %s from %s",
	    enum_show(&payload_desc, hash.isag_np), show_sa(sap));
	/* XXX Could send notification back */
	return STF_FAIL;
    }

    /* Advance state */
    st->st_state = OAKLEY_QUICK_R_2;

    /* Derive new keying material */
    compute_keymats(st);

    /* XXX Unless the commit bit is set, we need to send a message
     * XXX to the kernel to establish the new SA.
     */

    
    if (!install_ipsec_sa(st, FALSE))
	return STF_FAIL;

    return STF_NO_REPLY;
}
