/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

/* This file contains the routines that interface with the kernel's
 * IPsec mechanism.
 */

#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#ifdef KLIPS
/* A little experimentation indicates that the order of the next six
 * includes is critical.  I don't know why, or what the rules are.  Ugh!
 */
# include <sys/types.h>
# include <linux/types.h>
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>
# include <netdb.h>

# include <unistd.h>
# include <fcntl.h>

# include <radij.h>
# include <ipsec_encap.h>
# include <ipsec_netlink.h>
# include <ipsec_xform.h>
# include <ipsec_ipe4.h>
# include <ipsec_ah.h>
# include <ipsec_esp.h>

#else

/* this collection is included in the other case, but in an order
 * that makes it hard to make the #includes common.
 */
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>

#endif

#include "constants.h"
#include "rnd.h"
#include "state.h"
#include "kernel.h"
#include "log.h"


bool no_klips = FALSE;	/* don't actually use KLIPS */

/* Generate Unique SPI numbers.
 *
 * The specs say that the number must not be less than 0x100.
 * XXX This should be replaced by a call to the kernel when
 * XXX we get an API.
 * The returned SPI is in network byte order.
 * We use a random number as the initial SPI so that there is
 * a good chance that different Pluto instances will choose
 * different SPIs.  This is good for two reasons.
 * - the keying material for the initiator and responder only
 *   differs if the SPIs differ.
 * - if Pluto is restarted, it would otherwise recycle the SPI
 *   numbers and confuse everything.  When the kernel generates
 *   SPIs, this will no longer matter.
 */
void
get_ipsec_spi(u_char spi_out[IPSEC_DOI_SPI_SIZE])
{
    static u_int32_t spi = 0;
    u_int32_t nspi;

    while (spi < 0x100)
	get_rnd_bytes((u_char *)&spi, sizeof(spi));

    nspi = htonl(spi);
    spi++;
    memcpy(spi_out, &nspi, IPSEC_DOI_SPI_SIZE);
    DBG(DBG_CONTROL, DBG_dump("generate SPI:", spi_out, IPSEC_DOI_SPI_SIZE));
}

/* do the "route" commands required:
 *  route del <peeruser> netmask <peerusermask>
 *  route add <peeruser> netmask <peerusermask> dev ipsec0 gw <peer>
 * Note that this is not general:
 * - "ipsec0" is wired in as the device
 * - the peer's client must be a network
 * - the peer must be the gateway
 * - the peer's IP address must not be in the network
 */

#ifndef ROUTECMD
# define ROUTECMD "/sbin/route"
#endif

#ifndef IPSECDEV
# define IPSECDEV "ipsec0"
#endif

static bool
do_route(struct state *st, bool add)
{
    char cmd_buf[100];
    char
	peer_str[SOCKADDR_STRING_SIZE],
	peeruser_str[SOCKADDR_STRING_SIZE],
	peerusernetmask_str[SOCKADDR_STRING_SIZE];
    int r;

    strcpy(peer_str, get_address(st->st_peer));
    strcpy(peeruser_str, get_address_in(st->st_peeruser));
    strcpy(peerusernetmask_str, get_address_in(st->st_peerusernetmask));

    /* Check to see if peer is within its client's network.
     * If so, what we were about to do won't work.
     * We require a non-zero (peer xor peerclient) & peerclientnetmask
     */
    if (0 == ((((struct sockaddr_in *)&st->st_peer)->sin_addr.s_addr
    ^ st->st_peeruser.sin_addr.s_addr) & st->st_peerusernetmask.sin_addr.s_addr))
    {
	log("cannot perform routing: peer %s is within peer's user network %s/%s",
	    peer_str, peeruser_str, peerusernetmask_str);
	return FALSE;
    }

    r = snprintf(cmd_buf, sizeof(cmd_buf),
	    "%s del -net %s netmask %s",
	    ROUTECMD, peeruser_str,  peerusernetmask_str);
    assert(r > 0);
    DBG(DBG_CONTROL, DBG_log("executing command: %s", cmd_buf));
#ifdef KLIPS
    if (!no_klips)
    {
	r = system(cmd_buf);
	if (r != 0)
	{
	    log("command failure %d ignored: %s", r, cmd_buf);
# if 0	/* ??? only r == 1792 should be ignored (no route to delete)? */
	    return FALSE;
# endif
	}
    }
#endif /* KLIPS */

    if (add)
    {
	r = snprintf(cmd_buf, sizeof(cmd_buf),
		"%s add -net %s netmask %s dev %s gw %s",
		ROUTECMD, peeruser_str,  peerusernetmask_str, IPSECDEV, peer_str);
	assert(r > 0);
	DBG(DBG_CONTROL, DBG_log("executing command: %s", cmd_buf));
#ifdef KLIPS
	if (!no_klips)
	{
	    r = system(cmd_buf);
	    if (r != 0)
	    {
		log("command failed %d: %s", r, cmd_buf);
		return FALSE;
	    }
	}
#endif
    }

    return TRUE;
}

#ifdef KLIPS				/* Support for ipsec-0.5 */
/*
 * Setup an IPsec route entry. Code taken from addrt.c.
 * XXX Doesn't handle proxy (firewall mode) yet; it does
 * XXX encapsulation on our packets only (useful to protect
 * XXX the IP options).
 */
static bool
do_eroute(int fd, struct state *st, __u8 op)
{
    struct encap_msghdr emh;
    struct sockaddr_in sin;

    memset(&emh, '\0', sizeof(emh));	/* clear unused parts */

    memcpy(&sin, &st->st_peer, sizeof(sin));	/* sockaddr => sockaddr_in */

    emh.em_magic = EM_MAGIC;
    emh.em_msglen = sizeof(emh);
    emh.em_version = 0;
    emh.em_type = op;
    emh.em_eaddr.sen_len = emh.em_emask.sen_len
			 = sizeof(struct sockaddr_encap);
    emh.em_eaddr.sen_family = emh.em_emask.sen_family = 26;
    emh.em_eaddr.sen_type = SENT_IP4;
    emh.em_emask.sen_type = 255;

    emh.em_eaddr.sen_ip_src.s_addr = st->st_myuser.sin_addr.s_addr;
    emh.em_emask.sen_ip_src.s_addr = st->st_myusernetmask.sin_addr.s_addr;

    emh.em_eaddr.sen_ip_dst.s_addr = st->st_peeruser.sin_addr.s_addr;
    emh.em_emask.sen_ip_dst.s_addr = st->st_peerusernetmask.sin_addr.s_addr;

    emh.em_erdst.s_addr = sin.sin_addr.s_addr;

    /* multiple SPIs will be grouped; we must refer to the
     * last in the sequence: ah esp encap.
     */
    if (st->st_ah.present)
	memcpy(&emh.em_erspi, st->st_ah.attrs.spi, sizeof(emh.em_erspi));

    if (st->st_esp.present)
	memcpy(&emh.em_erspi, st->st_esp.attrs.spi, sizeof(emh.em_erspi));

    if (st->st_ah.attrs.encapsulation == ENCAPSULATION_MODE_TUNNEL
    || st->st_esp.attrs.encapsulation == ENCAPSULATION_MODE_TUNNEL)
    {
	/* ??? the other sending SPIs have been assigned by our peer.
	 * get_ipsec_spi is designed to select a receiving SPI.
	 * Is this therefore broken?
	 */
	get_ipsec_spi(st->st_tunnel_spi);
	memcpy(&emh.em_erspi, st->st_tunnel_spi, sizeof(emh.em_erspi));
    }

    DBG_cond_dump(DBG_KLIPS, "eroute => KLIPS:\n", &emh, sizeof(emh));

    if (!no_klips)
    {
	if (write(fd, &emh, sizeof(emh)) != sizeof(emh))
	{
	    log_errno("write() failed in do_eroute()");
	    return FALSE;
	}
    }

    DBG(DBG_CONTROL, DBG_log("eroute to %s %s",
	get_address_in(sin), op==EMT_SETEROUTE? "set up" : "torn down"));
    return TRUE;
}

static bool
del_spi(int fd, const u_char spi[IPSEC_DOI_SPI_SIZE], struct in_addr dest)
{
    struct encap_msghdr delmsg;

    memset(&delmsg, '\0', sizeof(delmsg));
    delmsg.em_magic = EM_MAGIC;
    delmsg.em_version = 0;
    delmsg.em_msglen = EMT_SETSPI_FLEN;
    delmsg.em_type = EMT_DELSPI;
    delmsg.em_alg = XF_DEL;
    delmsg.em_dst = dest;
    memcpy(&delmsg.em_spi, spi, IPSEC_DOI_SPI_SIZE);
    DBG_cond_dump(DBG_KLIPS, "delete SPI => KLIPS:\n",
	(void *) &delmsg, EMT_SETSPI_FLEN);
    if (!no_klips)
    {
	if (write(fd, (caddr_t)&delmsg, EMT_SETSPI_FLEN) != EMT_SETSPI_FLEN)
	{
	    log_errno("write() of delete SPI failed");
	    return FALSE;
	}
    }
    return TRUE;
}

/* Setup a pair of SAs. Code taken from setsa.c and spigrp.c, in
 * ipsec-0.5.
 */
static bool
setup_ipsec_sa(int fd, struct state *st, bool initiator)
{
    /* We need to build both a Sending SA and a Receiving SA.
     * Since the logic for each is similar, we build both
     * netlink packets at once.  The parallel variables are
     * prefixed by s_ and r_.
     * The Receiver uses st->st_keymat
     * The Sender uses st->st_peerkeymat
     */

    /* encap_msghdr is from "ipsec_netlink.h"
     * It preceded each netlink message.
     * We create "generic" s_ and r_ headers to save work.
     * User must fill in:
     *	?.h.em_alg
     *	?.h.em_msglen
     *	?.h.em_type
     *	?.h.em_spi
     */
    struct encap_msghdr r_h, s_h;

    struct sockaddr_in sin;

    /* sender SPIs, saved for spigrouping, if necessary*/
    u_int32_t
	s_spi[EM_MAXRELSPIS],
	*s_spi_next = s_spi,
	r_spi[EM_MAXRELSPIS],
	*r_spi_next = r_spi;

    memcpy(&sin, &st->st_peer, sizeof(sin));

    /* build generic message headers */

    memset(&r_h, '\0', sizeof(r_h));
    r_h.em_magic = EM_MAGIC;
    r_h.em_version = 0;
    r_h.em_if = 0;	/* ??? what is this? */

    s_h = r_h;	/* replicate hdr */

    r_h.em_dst = ((struct sockaddr_in *)&st->st_me)->sin_addr;
    s_h.em_dst = sin.sin_addr;

    /* Set our receiving SA */

    if (st->st_ah.present)
    {
	/* PROTO_IPSEC_AH */
	/* ??? mysterious code stolen from setsa.c:f_hmac()
	 * ??? why is struct ahhmacmd5_edata OK for SHA too?
	 */

	/* ahhmacmd5_edata is from "ipsec_ah.h"
	 * ahhmacsha1_edata has identical fields except that the ame_key
	 * field has a different length.
	 */
	union {
	    struct encap_msghdr h;
	    struct {
		char dummy[offsetof(struct encap_msghdr, em_dat)];
		struct ahhmacmd5_edata d;	/* works for both */
	    } u;
	} r_ah, s_ah;
	int alg;
	size_t ah_alen;

	memset(&r_ah, '\0', sizeof(r_ah));
	memset(&s_ah, '\0', sizeof(s_ah));

	r_ah.h = r_h;
	s_ah.h = s_h;

	r_ah.h.em_type = s_ah.h.em_type = EMT_SETSPI;

	memcpy(&r_ah.h.em_spi, st->st_ah.our_spi, IPSEC_DOI_SPI_SIZE);
	memcpy(&s_ah.h.em_spi, st->st_ah.attrs.spi, IPSEC_DOI_SPI_SIZE);

	switch (st->st_ah.attrs.transid)
	{
	case AH_MD5:
	    alg = XF_AHHMACMD5;
	    ah_alen = MD5_DIGEST_SIZE;
	    break;

	case AH_SHA:
	    alg = XF_AHHMACSHA1;
	    ah_alen = SHA1_DIGEST_SIZE;
	    break;

	case AH_DES:
	default:
	    log("AH transform %s not implemented yet",
		enum_show(&ah_transformid_desc, st->st_ah.attrs.transid));
	    goto fail;
	}

	/* ??? this length calculation is repulsive */
	r_ah.h.em_msglen = s_ah.h.em_msglen =
	    EMT_SETSPI_FLEN + 3 * sizeof(u_short) + 2 * sizeof(u_char) + st->st_ah.keymat_len;
	r_ah.h.em_alg = s_ah.h.em_alg = alg;
	r_ah.u.d.ame_alen = s_ah.u.d.ame_alen = ah_alen;
	r_ah.u.d.ame_klen = s_ah.u.d.ame_klen = st->st_ah.keymat_len;
	r_ah.u.d.ame_replayp = s_ah.u.d.ame_replayp = TRUE;
	r_ah.u.d.ame_ooowin = s_ah.u.d.ame_ooowin = 32;    /* ??? pure guess */
	memcpy(r_ah.u.d.ame_key, st->st_ah.our_keymat, st->st_ah.keymat_len);
	memcpy(s_ah.u.d.ame_key, st->st_ah.peer_keymat, st->st_ah.keymat_len);

	DBG_cond_dump(DBG_KLIPS, "setup inbound AH transform => KLIPS:\n", &r_ah, r_ah.h.em_msglen);

	if (!no_klips)
	{
	    if (write(fd, &r_ah, r_ah.h.em_msglen) != r_ah.h.em_msglen)
	    {
		log_errno("write() of inbound AH SPI failed in setup_ipsec_sa()");
		goto fail;
	    }
	}
	*r_spi_next++ = r_ah.h.em_spi;

	DBG_cond_dump(DBG_KLIPS, "setup outbound AH transform => KLIPS:\n", &s_ah, s_ah.h.em_msglen);

	if (!no_klips)
	{
	    if (write(fd, &s_ah, s_ah.h.em_msglen) != s_ah.h.em_msglen)
	    {
		log_errno("write() of outbound AH SPI failed in setup_ipsec_sa()");
		goto fail;
	    }
	}
	*s_spi_next++ = s_ah.h.em_spi;
    }

    if (st->st_esp.present)
    {
	/* PROTO_IPSEC_ESP */

	/* espblkrply_edata is from "ipsec_esp.h"
	 */
	union {
	    struct encap_msghdr h;
	    struct {
		char dummy[offsetof(struct encap_msghdr, em_dat)];
		struct espblkrply_edata d;	/* works for all */
	    } u;
	} r_esp, s_esp;

	struct esp_info {
	    u_int8_t transid;	/* negotiated ESP transform */
	    u_int16_t auth;	/* negotiated AUTH */
	    int alg;	/* KLIPS encoding for ESP x AUTH */
	    size_t ivlen;	/* length of IV */
	    size_t enckeylen;	/* keylength for ESP transform */
	    size_t authkeylen;	/* keylength for AUTH */
	};

	static const struct esp_info esp_info[] = {
	    { ESP_NULL, AUTH_ALGORITHM_HMAC_MD5,
		XF_ESPNULLMD596, 0, 0, AHMD596_KLEN },
	    { ESP_NULL, AUTH_ALGORITHM_HMAC_SHA1,
		XF_ESPNULLSHA196, 0, 0, AHSHA196_KLEN },

	    { ESP_DES, AUTH_ALGORITHM_NONE,
		XF_ESPDES, EMT_ESPDES_IV_SZ, EMT_ESPDES_KEY_SZ, 0 },
	    { ESP_DES, AUTH_ALGORITHM_HMAC_MD5,
		XF_ESPDESMD596, EMT_ESPDES_IV_SZ, EMT_ESPDES_KEY_SZ, AHMD596_KLEN },
	    { ESP_DES, AUTH_ALGORITHM_HMAC_SHA1,
		XF_ESPDESSHA196, EMT_ESPDES_IV_SZ, EMT_ESPDES_KEY_SZ, AHSHA196_KLEN },

	    { ESP_3DES, AUTH_ALGORITHM_NONE,
		XF_ESP3DES, EMT_ESPDES_IV_SZ, EMT_ESP3DES_KEY_SZ, 0 },
	    { ESP_3DES, AUTH_ALGORITHM_HMAC_MD5,
		XF_ESP3DESMD596, EMT_ESPDES_IV_SZ, EMT_ESP3DES_KEY_SZ, AHMD596_KLEN },
	    { ESP_3DES, AUTH_ALGORITHM_HMAC_SHA1,
		XF_ESP3DESSHA196, EMT_ESPDES_IV_SZ, EMT_ESP3DES_KEY_SZ, AHSHA196_KLEN },
	};

	const struct esp_info *ei;

	memset(&r_esp, '\0', sizeof(r_esp));
	memset(&s_esp, '\0', sizeof(s_esp));

	r_esp.h = r_h;
	s_esp.h = s_h;

	r_esp.h.em_type = s_esp.h.em_type = EMT_SETSPI;

	memcpy(&r_esp.h.em_spi, st->st_esp.our_spi, IPSEC_DOI_SPI_SIZE);
	memcpy(&s_esp.h.em_spi, st->st_esp.attrs.spi, IPSEC_DOI_SPI_SIZE);

	for (ei = esp_info; ; ei++)
	{
	    if (ei == &esp_info[elemsof(esp_info)])
	    {
		/* note: enum_show may use a static buffer, so two
		 * calls in one printf would be a mistake.
		 * enum_name does the same job, without a static buffer,
		 * assuming the name will be found.
		 */
		log("ESP transform %s / auth %s not implemented yet",
		    enum_name(&esp_transformid_desc, st->st_esp.attrs.transid),
		    enum_name(&auth_alg_desc, st->st_esp.attrs.auth));
		goto fail;
	    }

	    if (st->st_esp.attrs.transid == ei->transid
	    && st->st_esp.attrs.auth == ei->auth)
		break;
	}

	r_esp.h.em_alg = s_esp.h.em_alg = ei->alg;

	if (initiator)
	{
	    r_esp.u.d.eme_flags |= EME_INITIATOR;
	    s_esp.u.d.eme_flags |= EME_INITIATOR;
	}

	/* IV */
	if (ei->ivlen != 0)
	{
	    r_esp.u.d.eme_ivlen = s_esp.u.d.eme_ivlen = ei->ivlen;
	    get_rnd_bytes(r_esp.u.d.eme_iv, ei->ivlen);
	    get_rnd_bytes(s_esp.u.d.eme_iv, ei->ivlen);
	}

	/* divide up keying material */
	assert(st->st_esp.keymat_len == ei->enckeylen + ei->authkeylen);
	assert(ei->enckeylen <= sizeof(r_esp.u.d.eme_key));
	assert(ei->authkeylen <= sizeof(r_esp.u.d.ame_key));

	memcpy(r_esp.u.d.eme_key, st->st_esp.our_keymat, ei->enckeylen);
	memcpy(s_esp.u.d.eme_key, st->st_esp.peer_keymat, ei->enckeylen);
	r_esp.u.d.eme_klen = s_esp.u.d.eme_klen = ei->enckeylen;

	memcpy(r_esp.u.d.ame_key, st->st_esp.our_keymat+ei->enckeylen, ei->authkeylen);
	memcpy(s_esp.u.d.ame_key, st->st_esp.peer_keymat+ei->enckeylen, ei->authkeylen);
	r_esp.u.d.ame_klen = s_esp.u.d.ame_klen = ei->authkeylen;

	/* window: always full size */
	r_esp.u.d.eme_ooowin = s_esp.u.d.eme_ooowin = 32;

	r_esp.h.em_msglen = s_esp.h.em_msglen =
	    EMT_SETSPI_FLEN + sizeof(struct espblkrply_edata);

	DBG_cond_dump(DBG_KLIPS, "setup inbound ESP transform => KLIPS:\n", &r_esp, r_esp.h.em_msglen);

	if (!no_klips)
	{
	    if (write(fd, &r_esp, r_esp.h.em_msglen) != r_esp.h.em_msglen)
	    {
		log_errno("write() of inbound ESP SPI failed in setup_ipsec_sa()");
		goto fail;
	    }
	}
	*r_spi_next++ = r_esp.h.em_spi;

	DBG_cond_dump(DBG_KLIPS, "setup outbound ESP transform => KLIPS:\n", &s_esp, s_esp.h.em_msglen);

	if (!no_klips)
	{
	    if (write(fd, &s_esp, s_esp.h.em_msglen) != s_esp.h.em_msglen)
	    {
		log_errno("write() of outbound ESP SPI failed in setup_ipsec_sa()");
		goto fail;
	    }
	}
	*s_spi_next++ = s_esp.h.em_spi;
    }

    if (st->st_ah.attrs.encapsulation == ENCAPSULATION_MODE_TUNNEL
    || st->st_esp.attrs.encapsulation == ENCAPSULATION_MODE_TUNNEL)
    {
	/*
	 * XXX hack alert -- we SHOULD NOT HAVE TO HAVE A DIFFERENT SPI
	 * XXX FOR IP-in-IP ENCAPSULATION!
	 */

	/* ipe4_xdata is from "ipsec_ipe4.h" */
	union {
	    struct encap_msghdr h;
	    struct {
		char dummy[offsetof(struct encap_msghdr, em_dat)];
		struct ipe4_xdata d;
	    } u;
	} m;	/* encapsulating message */

	memset(&m, '\0', sizeof(m));
	m.h = s_h;

	m.h.em_type = EMT_SETSPI;
	memcpy(&m.h.em_spi, st->st_tunnel_spi, IPSEC_DOI_SPI_SIZE);

	m.h.em_msglen = EMT_SETSPI_FLEN + EMT_IPE4_ULEN;
	m.h.em_alg = XF_IP4;

	m.u.d.i4_src = ((struct sockaddr_in *)&st->st_me)->sin_addr;
	m.u.d.i4_dst = sin.sin_addr;

	DBG_cond_dump(DBG_KLIPS, "setup encapsulation => KLIPS:\n", &m, m.h.em_msglen);

	if (!no_klips)
	{
	    if (write(fd, &m, m.h.em_msglen) != m.h.em_msglen)
	    {
		log_errno("write() of encapsulation failed in setup_ipsec_sa()");
		goto fail;
	    }
	}
	*s_spi_next++ = m.h.em_spi;
    }

    if (s_spi_next > &s_spi[1])
    {
	/* Group the multiple sending SPIs
	 * All the info lives in the header!
	 * EM_MAXRELSPIS is max number of groupings.
	 */
	struct encap_msghdr m;
	int i;

	memset(&m, '\0', sizeof(m));

	m.em_magic = EM_MAGIC;
	m.em_version = 0;
	m.em_type = EMT_GRPSPIS;
	m.em_msglen = EM_MINLEN;

	/* reverse order of creation in above code:
	 * encap, esp, ah.
	 */
	for (i = 0; i != s_spi_next - s_spi; i++)
	{
	    m.em_rel[i].emr_spi = s_spi_next[ - 1 - i];
	    m.em_rel[i].emr_dst.s_addr = sin.sin_addr.s_addr;
	    m.em_msglen += sizeof(m.em_rel[0]);

	    DBG(DBG_KLIPS,
		DBG_log("grouping %s/%08lx",
		    inet_ntoa(sin.sin_addr),
		    (unsigned long)ntohl(m.em_rel[i].emr_spi)));
	}

	DBG_cond_dump(DBG_KLIPS, "setup grouping => KLIPS:\n", &m, m.em_msglen);

	if (!no_klips)
	{
	    if (write(fd, &m, m.em_msglen) != m.em_msglen)
	    {
		log_errno("write() of grouping failed in setup_ipsec_sa()");
		goto fail;
	    }
	}
	/* could update s_spi, but it will not be used */
    }

    return TRUE;

fail:
    {
	/* undo the done SPIs */
	while (r_spi_next-- != r_spi)
	    (void) del_spi(fd, (u_char *)r_spi_next,
		((struct sockaddr_in *)&st->st_me)->sin_addr);

	while (s_spi_next-- != s_spi)
	    (void) del_spi(fd, (u_char *)s_spi_next, sin.sin_addr);
	return FALSE;
    }
}

/* teardown_ipsec_sa is a canibalized version of setup_ipsec_sa */

static bool
teardown_ipsec_sa(int fd, struct state *st, bool initiator)
{
    /* We need to tear down both a Sending SA and a Receiving SA.
     * On the receiving side, only AH and ESP are possible (no encapsulation).
     * On the sending side, first of AH and ESP is fine:
     * since only one of a group needs to be deleted, either there
     * was no group, or the AH or ESP will be in it.
     */
    u_char *an_out_spi = NULL;

    /* delete receiving SPIs */

    if (st->st_ah.present)
    {
	(void) del_spi(fd, st->st_ah.our_spi,
	    ((struct sockaddr_in *)&st->st_me)->sin_addr);
	an_out_spi = st->st_ah.attrs.spi;
    }

    if (st->st_esp.present)
    {
	(void) del_spi(fd, st->st_esp.our_spi,
	    ((struct sockaddr_in *)&st->st_me)->sin_addr);
	an_out_spi = st->st_esp.attrs.spi;
    }

    /* delete some outgoing SPI (grouping gets rest) */

    (void) del_spi(fd, an_out_spi,
	((struct sockaddr_in *)&st->st_peer)->sin_addr);
    return TRUE;
}

#endif /* KLIPS */

bool
install_ipsec_sa(struct state *st, bool initiator)
{
    /* do routing commands, if we can */
    if (!do_route(st, TRUE))
	return FALSE;

#ifdef KLIPS
    if (no_klips)
    {
	return do_eroute(-1, st, EMT_SETEROUTE)
	    && setup_ipsec_sa(-1, st, initiator);
    }
    else
    {
	int fd = open("/dev/ipsec", O_WRONLY);
	bool res;

	if (fd < 0)
	{
	    log_errno("open() of /dev/ipsec failed in install_ipsec_sa()");
	    return FALSE;
	}

	res = do_eroute(fd, st, EMT_SETEROUTE);

	if (res)
	{
	    res = setup_ipsec_sa(fd, st, initiator);
	    if (!res)
		(void)do_eroute(fd, st, EMT_DELEROUTE);
	}

	close(fd);	/* rain or shine */

	return res;
    }
#else
    DBG(DBG_CONTROL, DBG_log("if I knew how, I'd do_eroute() and setup_ipsec_sa()"));
    return TRUE;
#endif
}

bool
delete_ipsec_sa(struct state *st, bool initiator)
{
    /* do routing commands, if we can */
    if (!do_route(st, FALSE))
	return FALSE;

#ifdef KLIPS
    if (no_klips)
    {
	return do_eroute(-1, st, EMT_DELEROUTE)
	    && teardown_ipsec_sa(-1, st, initiator);
    }
    else
    {
	int fd = open("/dev/ipsec", O_WRONLY);
	bool res;

	if (fd < 0)
	{
	    log_errno("open() of /dev/ipsec failed in install_ipsec_sa()");
	    return FALSE;
	}

	res = do_eroute(fd, st, EMT_DELEROUTE)
	    && teardown_ipsec_sa(fd, st, initiator);

	close(fd);	/* rain or shine */

	return res;
    }
#else
    DBG(DBG_CONTROL, DBG_log("if I knew how, I'd do_eroute() and teardown_ipsec_sa()"));
    return TRUE;
#endif
}
