/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "constants.h"
#include "defs.h"
#include "whack.h"
#include "packet.h"	/* needed by ipsec_doi.h */
#include "ipsec_doi.h"
#include "kernel_comm.h"
#include "log.h"

/* This file implements all the kernel communicating routines.
 * XXX Eventually it should use PF_KEY or what kernel API is available
 * XXX but for now this will have to do.
 */

/* Handle a kernel request. Supposedly, there's a message in
 * the kernelsock socket.
 */
void
kernel_handle(int kernelfd, int sock)
{
    struct whack_message msg;
    int n;

    n = read(kernelfd, &msg, sizeof(msg));
    if (n == -1)
    {
	log_errno("read() failed in kernel_handle()");
	return;
    }
    if (n != sizeof(msg))
    {
	log("truncated message from whack: got %d bytes, expected %d.  Ignored.",
	    n, (int) sizeof(msg));
	return;
    }

    if (msg.magic != WHACK_MAGIC)
    {
	log("message from whack has bad magic -- probably wrong version");
	return;
    }

    if (msg.whack_options)
    {
#ifdef DEBUG
	debugging = msg.debugging;
	DBG(DBG_CONTROL,
	    DBG_log("debugging = %s", bitnamesof(debug_bit_names, debugging)));
#endif
    }

    if (msg.whack_initiate)
    {
	struct sockaddr_in
	    peer,
	    local_client_net,
	    local_client_mask,
	    peer_client_net,
	    peer_client_mask;


	/* check for exceeding a  current limitation */
	if (msg.explicit_local_client != msg.explicit_peer_client)
	{
	    log("must specify both client subnets (or neither)");
	    return;
	}

	msg.local_client_net.s_addr &= msg.local_client_mask.s_addr;
	msg.peer_client_net.s_addr &= msg.peer_client_mask.s_addr;

	/* Note: this code is broken up because inet_ntoa returns its result
	 * in a single static buffer.
	 * Note: macro parameters must not have commas outside of parentheses
	 * so the declarations have to be written without commas.
	 * Braces are supplied by the DBG macro.
	 */
	DBG(DBG_CONTROL,
	    char ocn[SOCKADDR_STRING_SIZE];
	    char ocnm[SOCKADDR_STRING_SIZE];
	    char pcn[SOCKADDR_STRING_SIZE];
	    char pcnm[SOCKADDR_STRING_SIZE];

	    strcpy(ocn, inet_ntoa(msg.local_client_net));
	    strcpy(ocnm, inet_ntoa(msg.local_client_mask));
	    strcpy(pcn, inet_ntoa(msg.peer_client_net));
	    strcpy(pcnm, inet_ntoa(msg.peer_client_mask));

	    DBG_log("initiating exchange with [%s], port %d, goal %s",
		inet_ntoa(msg.peer), ntohs(msg.peer_port),
		bitnamesof(sa_goal_bit_names, msg.goal));
	    DBG_log("Proxying: %s/%s<--->%s/%s", ocn, ocnm, pcn, pcnm));

	/* If we're doing proxy, turn tunnel goal on automatically */
	if (msg.explicit_local_client || msg.explicit_peer_client)
	{
	    DBG(DBG_CONTROL,
		if ((msg.goal & GOAL_TUNNEL) == 0)
		    DBG_log("adding GOAL_TUNNEL since <local subnet> is not 0.0.0.0"));
	    msg.goal |= GOAL_TUNNEL;
	}

	mksin(peer, msg.peer.s_addr, ntohs(msg.peer_port));
	mksin(local_client_net, msg.local_client_net.s_addr, 0);
	mksin(local_client_mask, msg.local_client_mask.s_addr, 0);
	mksin(peer_client_net, msg.peer_client_net.s_addr, 0);
	mksin(peer_client_mask, msg.peer_client_mask.s_addr, 0);

	ipsecdoi_initiate(sock, peer,
	    local_client_net, local_client_mask,
	    peer_client_net, peer_client_mask,
	    msg.goal, KEY_IKE);
    }

    if (msg.whack_shutdown)
    {
	/* ??? not yet implemented */
	DBG(DBG_CONTROL, DBG_log("shutting down"));
	exit_pluto(0);	/* delete lock and leave, with 0 status */
    }
}
