/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include "defs.h"
#include "constants.h"
#include "log.h"

bool log_to_stderr = FALSE;

/*
 * This file defines error logging functions
 */

void
init_log(void)
{
    if (log_to_stderr)
	setbuf(stderr, NULL);
    else
	openlog("Pluto", LOG_CONS | LOG_NDELAY | LOG_PID, LOG_AUTHPRIV);
}

static void
close_log(void)
{
    if (!log_to_stderr)
	closelog();
}

/*
 * If the first argument is true, then this is an error, otherwise
 * it's a warning. It takes additional arguments to be passed
 * syslog/printf verbatim.
 */

#include <stdarg.h>

void
log(char *message, ...)
{
    va_list args;
    char m[1024];

    va_start(args, message);
    vsnprintf(m, sizeof(m), message, args);
    va_end(args);

    if (log_to_stderr)
	fprintf(stderr, "%s\n", m);
    else
	syslog(LOG_WARNING, "%s", m);
}

void
log_errno(char *message, ...)
{
    va_list args;
    int e = errno;  /* capture before we change it (may be too late) */
    char m[1024];

    va_start(args, message);
    vsnprintf(m, sizeof(m), message, args);
    va_end(args);

    if (log_to_stderr)
	fprintf(stderr, "ERROR: %s. Errno %d: %s\n", m, e, strerror(e));
    else
	syslog(LOG_ERR, "ERROR: %s. Errno %d: %s", m, e, strerror(e));
}

void
exit_log(char *message, ...)
{
    va_list args;
    char m[1024];

    va_start(args, message);
    vsnprintf(m, sizeof(m), message, args);
    va_end(args);

    if (log_to_stderr)
	fprintf(stderr, "FATAL ERROR: %s\n", m);
    else
	syslog(LOG_ERR, "FATAL ERROR: %s", m);
    close_log();
    exit_pluto(-1);
}

void
exit_log_errno(char *message, ...)
{
    va_list args;
    int e = errno;  /* capture before we change it (may be too late) */
    char m[1024];

    va_start(args, message);
    vsnprintf(m, sizeof(m), message, args);
    va_end(args);

    if (log_to_stderr)
	fprintf(stderr, "FATAL ERROR: %s. Errno %d: %s\n", m, e, strerror(e));
    else
	syslog(LOG_ERR, "FATAL ERROR: %s. Errno %d: %s", m, e, strerror(e));
    close_log();
    exit_pluto(-1);
}


/*
 * This routine returns a user readable form of an address contained in a
 * sockaddr structure. The return value CAN be a statically allocated
 * object (as is the case with inet_ntoa()).
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

char *
get_address_in(struct sockaddr_in sin)
{
    static char mess[SOCKADDR_STRING_SIZE];

    switch (sin.sin_family)
    {
	case AF_INET:
	    return inet_ntoa(sin.sin_addr);

	default:
	    snprintf(mess, sizeof(mess), "(UNEXPECTED address family %d)",
		sin.sin_family);
	    return mess;
    }
}

char *
get_address(struct sockaddr sa)
{
    static char mess[SOCKADDR_STRING_SIZE];
    struct sockaddr_in sin;

    switch (sa.sa_family)
    {
	case AF_INET:
	    memcpy(&sin, &sa, sizeof(sa));
	    return get_address_in(sin);

	default:
	    snprintf(mess, sizeof(mess), "(unknown address family %d)",
		sin.sin_family);
	    return mess;
    }
}

/*
 * Return a port (if applicable), from a struct sockaddr. If not applicable,
 * return -1.
 */

int
get_port_in(struct sockaddr_in sin)
{
    switch (sin.sin_family)
    {
	case AF_INET:
	    return ntohs(sin.sin_port);

	default:
	    return -1;
    }
}

int
get_port(struct sockaddr sa)
{
    struct sockaddr_in sin;

    switch (sa.sa_family)
    {
	case AF_INET:
	    memcpy(&sin, &sa, sizeof(sa));
	    return get_port_in(sin);

	default:
	    return -1;
    }
}

/* format address and port together in a static buffer */

const char *
show_sa(const struct sockaddr *sa)
{
    static char buf[SOCKADDR_STRING_SIZE];

    switch (sa->sa_family)
    {
	case AF_INET:
	    snprintf(buf, sizeof(buf), "%s, port %d",
		get_address(*sa), get_port(*sa));
	    break;

	default:
	    snprintf(buf, sizeof(buf), "(unexpected address family %d)",
		sa->sa_family);
	    break;
    }
    return buf;
}

const char *
show_sa_in(const struct sockaddr_in *sin)
{
    return show_sa((struct sockaddr *)sin);
}


/* Debugging message support */

#ifdef DEBUG

unsigned int debugging = DBG_NONE;	/* default to reporting nothing */

/* log a debugging message (prefixed by "| ") */

void
DBG_log(char *message, ...)
{
    va_list args;
    char m[1024];

    va_start(args, message);
    vsnprintf(m, sizeof(m), message, args);
    va_end(args);

    if (log_to_stderr)
	fprintf(stderr, "| %s\n", m);
    else
	syslog(LOG_DEBUG, "%s", m);
}

/* dump raw bytes in hex to stderr (for lack of any better destination) */

void
DBG_dump(const char *label, const void *p, size_t len)
{
    char buf[100];
    char *bp;
    const unsigned char *cp = p;
    
    bp = buf;

    if (label != NULL && label[0] != '\0')
    {
	size_t len = strlen(label);

	assert(len + 1 <= sizeof(buf));
	strcpy(buf, label);
	if (buf[len-1] == '\n')
	{
	    buf[len-1] = '\0';	/* get rid of newline */
	    DBG_log("%s", buf);
	}
	else
	{
	    assert(len + 4 * (1 + 4 * 3) + 1 <= sizeof(buf));
	    bp = buf + len;
	}
    }

    do {
	int i, j;

	for (i = 0; len!=0 && i!=4; i++) {
	    *bp++ = ' ';
	    for (j = 0; len!=0 && j!=4; len--, j++)
	    {
		static const char hexdig[] = "0123456789abcdef";

		*bp++ = ' ';
		*bp++ = hexdig[(*cp >> 4) & 0xF];
		*bp++ = hexdig[*cp & 0xF];
		cp++;
	    }
	}
	*bp = '\0';
	DBG_log("%s", buf);
	bp = buf;
    } while (len != 0);
}

#endif /* DEBUG */
