/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include <assert.h>

#include "constants.h"
#include "defs.h"
#include "server.h"
#include "kernel.h"
#include "log.h"
#include "preshared.h"
#include "rnd.h"
#include "state.h"

#include "sha1.h"
#include "md5.h"
#include "crypto.h"	/* requires sha1.h and md5.h */
#include "version.c"

static void
usage(const char *mess)
{
    if (mess != NULL && *mess != '\0')
	fprintf(stderr, "%s\n", mess);
    fprintf(stderr,
	"Usage: pluto"
	    " [--help]"
	    " [--version]"
	    " [--nofork]"
	    " [--stderrlog]"
	    " [--noklips]"
	    "\n\t"
	    "[--port <port-number>]"
	    "\n\t"
	    "[--lockdir <lock-directory>]"
	    "\n\t"
	    "[--secretsfile <secrets-file>]"
	    "\n"
#ifdef DEBUG
	"\t[--debug-none]"
	    " [--debug-all]"
	    " [--debug-raw]"
	    " [--debug-crypt]"
	    "\n\t"
	    "[--debug-parsing]"
	    " [--debug-emitting]"
	    " [--debug-control]"
	    " [--debug-klips]"
	    "\n"
#endif
	"FreeS/WAN %s\n",
	freeswan_version);
    exit_pluto(mess == NULL? 0 : 1);
}

/* lock file support
 * - provides convenient way for scripts to find Pluto's pid
 * - prevents multiple Plutos competing for the same port
 * NOTE: will not take account of sharing LOCK_DIR with other systems.
 */

#ifndef LOCK_DIR
# define LOCK_DIR "/var/run"
#endif

static const char *pluto_lock_dir = LOCK_DIR;
static const char *pluto_lock = NULL;

static void
create_lock()
{
    char *name;

    /* build complete path name */
    {
	char buf[30];	/* holds "/pluto.$UID.pid" */
	size_t dl = strlen(pluto_lock_dir);

	snprintf(buf, sizeof(buf),
	    pluto_port == IKE_UDP_PORT? "/pluto.pid" : "/pluto.%u.pid",
	    (unsigned int)pluto_port);
	name = alloc_bytes(dl + strlen(buf), "Pluto lock name");
	strcpy(name, pluto_lock_dir);
	strcpy(name + dl, buf + (dl != 0 && name[dl-1] == '/'));
    }

    /* create lockfile, or die in the attempt */
    {
	pid_t pid = getpid();
	int fd = open(name, O_WRONLY | O_CREAT | O_EXCL | O_TRUNC,
	    S_IRUSR | S_IRGRP | S_IROTH);
	char buf[30];	/* holds "<pid>\n" */

	if (fd < 0)
	{
	    if (errno == EEXIST)
		exit_log("Pluto lock file already exists %s", name);
	    else
		exit_log_errno("unable to create Pluto lock file %s", name);
	}
	pluto_lock = name;
	snprintf(buf, sizeof(buf), "%u\n", (unsigned int) pid);
	write(fd, buf, strlen(buf));	/* is noting failure useful? */
	close(fd);	/* is noting failure useful? */
    }
}

static void
delete_lock()
{
    if (pluto_lock != NULL)
	unlink(pluto_lock);	/* is noting failure useful? */
}

int
main(int argc, char **argv)
{
    bool fork_desired = TRUE;

    /* handle arguments */
    for (;;)
    {
	static const struct option long_opts[] = {
	    /* name, has_arg, flag, val */
	    { "help", FALSE, NULL, 'h' },
	    { "version", FALSE, NULL, 'v' },
	    { "nofork", FALSE, NULL, 'd' },
	    { "stderrlog", FALSE, NULL, 'e' },
	    { "noklips", FALSE, NULL, 'n' },
	    { "port", TRUE, NULL, 'p' },
	    { "lockdir", TRUE, NULL, 'l' },
	    { "secretsfile", TRUE, NULL, 's' },
#ifdef DEBUG
	    { "debug-none", FALSE, NULL, 'N' },
	    { "debug-all]", FALSE, NULL, 'A' },
	    { "debug-raw", FALSE, NULL, 'R' },
	    { "debug-crypt", FALSE, NULL, 'X' },
	    { "debug-parsing", FALSE, NULL, 'P' },
	    { "debug-emitting", FALSE, NULL, 'E' },
	    { "debug-control", FALSE, NULL, 'C' },
	    { "debug-klips", FALSE, NULL, 'K' },
#endif
	    { 0,0,0,0 }
	    };
	/* Note: we don't like the way short options get parsed
	 * by getopt_long, so we simply pass an empty string as
	 * the list.  It could be "hvdenp:l:s:" "NARXPECK".
	 */
	int c = getopt_long(argc, argv, "", long_opts, NULL);

	switch (c)
	{
	case EOF:	/* end of flags */
	    break;

	case 0: /* long option already handled */
	    continue;

	case '?':	/* diagnostic already printed by getopt_long */
	    usage("");
	    break;   /* not actually reached */

	case 'h':	/* --help */
	    usage(NULL);
	    break;	/* not actually reached */

	case 'v':	/* --version */
	    fprintf(stderr, "FreeS/WAN %s\n", freeswan_version);
	    exit_pluto(0);
	    break;	/* not actually reached */

	case 'd':	/* --nofork*/
	    fork_desired = FALSE;
	    continue;

	case 'e':	/* --stderrlog */
	    log_to_stderr = TRUE;
	    continue;

	case 'n':	/* --noklips */
	    no_klips = TRUE;
	    continue;

	case 'p':	/* --port <portnumber> */
	    if (optarg == NULL || !isdigit(optarg[0]))
		usage("missing port number");
	    {
		char *endptr;
		long port = strtol(optarg, &endptr, 0);

		if (*endptr != '\0' || endptr == optarg
		|| port <= 0 || port > 0x10000)
		    usage("<port-number> must be a number between 1 and 65535");
		pluto_port = port;
	    }
	    continue;

	case 'l':	/* --lockfile <lock-directory> */
	    pluto_lock_dir = optarg;
	    continue;

	case 's':	/* --secretsfile <secrets-file> */
	    shared_secrets_file = optarg;
	    continue;

#ifdef DEBUG
	case 'N':	/* --debug-none */
	    debugging = DBG_NONE;
	    continue;

	case 'A':	/* --debug-all */
	    debugging = DBG_ALL;
	    continue;

	case 'R':	/* --debug-raw */
	    debugging |= DBG_RAW;
	    continue;

	case 'X':	/* --debug-crypt */
	    debugging |= DBG_CRYPT;
	    continue;

	case 'P':	/* --debug-parsing */
	    debugging |= DBG_PARSING;
	    continue;

	case 'E':	/* --debug-emitting */
	    debugging |= DBG_EMITTING;
	    continue;

	case 'C':	/* --debug-control */
	    debugging |= DBG_CONTROL;
	    continue;

	case 'K':	/* --debug-klips */
	    debugging |= DBG_KLIPS;
	    continue;
#endif
	default:
	    assert(FALSE);
	}
	break;
    }
    if (optind != argc)
	usage("unexpected argument");

    /* If not suppressed, do daemon fork */

    if (fork_desired)
    {
# if defined(__OpenBSD__) || defined(__NetBSD__)
	/* ??? how do we preserve stdout is we wish to? */
	if (daemon(0, 0))  /* Conviniently available */
	{
	    int e = errno;

	    fprintf(stderr, "daemon() failed in main(). Errno %d: %s\n",
		errno, strerror(e));
	    exit_pluto(1);
	}
# else
	{
	    int i = fork();

	    if (i < 0)
	    {
		int e = errno;

		fprintf(stderr, "fork() failed in main(). Errno %d: %s\n",
		    errno, strerror(e));
		exit_pluto(1);
	    }
	    if (i != 0)
		exit_pluto(0);  /* Parent dies */
	}

	if (setsid() < 0)
	{
	    int e = errno;

	    fprintf(stderr, "setsid() failed in main(). Errno %d: %s\n",
		errno, strerror(e));
	    exit_pluto(1);
	}

	{
	    int i;

	    for (i = getdtablesize() - 1; i >= 0; i--)  /* Bad hack */
		if (!(log_to_stderr && i == 2))
		    close(i);

	    /* make sure that stdin, stdout, stderr are reserved */
	    if (open("/dev/null", O_RDONLY) != 0)
		abort();
	    if (dup2(0, 1) != 1)
		abort();
	    if (!log_to_stderr && dup2(0, 2) != 2)
		abort();
	}
# endif
    }

    init_log();
    log("Starting Pluto (FreeS/WAN Version %s)", freeswan_version);
    create_lock();
    init_rnd_pool();
    init_secret();
    init_states();
    init_crypto();
    call_server();
    return -1;        /* Shouldn't ever reach this */
}

void
exit_pluto(int status)
{
    delete_every_sa();
    delete_lock();
    exit(status);
}
