/* This program is an implementation of the ISAKMP Internet Standard.
 * This file is:
 * Copyright (C) 1998 D. Hugh Redelmeier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This file was written while the author was in Canada.
 *
 * You may contact the author by e-mail: hugh@mimosa.com
 */

/* Get preshared key for remote IP from the SHARED_SECRETS_FILE.
 *
 * Leading and trailing blanks on a line are ignored.
 * Empty lines and those starting with # are treated as comments.
 * Each entry has an IP address and a secret, normally separated by whitespace.
 * The secret can have quotes at either end to indicate that the enclosed
 * Whitespace is significant.
 * If nothing follows the IP address, the whole next line is taken as the
 * secret, including any whitespace and quotes.
 *
 * A sample secret file might be:
 *
 * # sssh! secrets!  Make this file only readable by owner.
 * 19.98.01.24 no way to know
 * 19.98.01.25 " leading and following space "
 * 19.98.01.26
 *	leading and following space
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "constants.h"
#include "defs.h"
#include "state.h"
#include "preshared.h"
#include "log.h"

const char *shared_secrets_file = SHARED_SECRETS_FILE;

u_char *
get_preshared_key(struct state *st, int *length)
{
    u_char buffer[513];
    u_char *p = (u_char *) NULL;
    FILE *fp;

    fp = fopen(shared_secrets_file, "r");
    if (fp == (FILE *) NULL)
    {
	log_errno("could not open %s", shared_secrets_file);
	return NULL;
    }

    while (fgets(buffer, sizeof(buffer), fp) != (char *) NULL)
    {
	struct in_addr ia;
	u_char *s, *b, *e;

#if 0
	DBG_log("looking for preshared key: %s", buffer);
#endif

	/* strip leading and trailing whitespace, including \n */
	for (e = buffer+strlen(buffer); e!=buffer && isspace(e[-1]); e--)
	    ;
	*e = '\0';

	for (s = buffer; isspace(*s); s++)
	    ;
	/* treat empty line or one starting with # as a comment
	 * Note that the secret may start with a # or be empty.
	 */
	if (s[0] == '\0' || s[0] == '#')
	    continue;

	/* find end of first field, the IP address */
	for (b = s; *b != '\0' && !isspace(*b); b++)
	    ;
	if (*b != '\0') {
	    *b = '\0';  /* terminate field */
	    do b++; while (*b != '\0' && isspace(*b));  /* move to next field */
	}

#ifdef SOLARIS
	if ((ia.s_addr = inet_addr(s)) == -1)
#else
	if (inet_aton(s, &ia) == 0)
#endif
	{
	    log("Expecting IP address in %s, but found \"%s\"",
		shared_secrets_file, buffer);
	    continue;	/* ignore this line */
	}

	if (*b == '\0') {
	    /* if no secret field is found on this line, use next line, raw */
	    if (fgets(buffer, sizeof(buffer), fp) == (char *) NULL)
	    {
		log("last entry in %s is incomplete", shared_secrets_file);
		break;
	    }
	    b = buffer;
	    e = b + strlen(b);
	    if (e != b && e[-1] == '\n')
		*--e = '\0';
	} else {
	    /* strip any quote from either end of secret */
	    if (*b == '"')
		b++;
	    if (e != b && e[-1] == '"')
		*--e = '\0';
	}

	if (st->st_peer.sa_family == AF_INET
	&& memcmp(&ia, &(*(struct sockaddr_in *)&st->st_peer).sin_addr, sizeof(ia)) == 0)
	{
	    p = clone_bytes(b, e - b, "preshared key");
	    *length = e - b;
	    DBG(DBG_CRYPT, DBG_log("secret used is [%s], length = %d", p, e - b));
	    break;
	}
    }
    fclose(fp);
    return p;
}
