/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef SOLARIS
# include <sys/sockio.h>	/* for Solaris 2.6: defines SIOCGIFCONF */
#endif
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <netdb.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/ioctl.h>

#include "constants.h"
#include "state.h"
#include "defs.h"
#include "log.h"
#include "server.h"
#include "timer.h"
#include "demux.h"
#include "kernel_comm.h"

/*
 *  Server main loop and socket initialization routines.
 */

/*
 * Initialize the kernel socket.
 * XXX For now, just get a port.
 */
static int
init_kernelfd(void)
{
    struct sockaddr_in sin;
    int s;

    s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (s == -1)
	exit_log_errno("socket() in init_kernelfd()");

    mksin(sin, htonl(INADDR_ANY), pluto_port + 1);
    DBG(DBG_KLIPS, DBG_log("init_kernelfd(): listening to port %d", pluto_port + 1));
    if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
	exit_log_errno("bind() in init_kernelfd()");

    return s;
}

/*
 * Initialize the sockets.
 */
static int
init_socket(int *socks, int maxsocks)
{
    struct sockaddr_in sin, *sin2;
    int i, j, k;
    struct ifconf ifconf;
    char buf[1024];

    /* initialize mask32 */
    mksin(mask32, inet_addr("255.255.255.255"), 0);

    /* Get a UDP socket */
    socks[0] = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

    if (socks[0] == -1)
	exit_log_errno("socket() failed in init_socket()");

    mksin(sin, htonl(INADDR_ANY), pluto_port);
    DBG(DBG_RAW | DBG_CONTROL, DBG_log("init_socket(): listening to port %d", pluto_port));
    k = 1;
    if (setsockopt(socks[0], SOL_SOCKET, SO_REUSEADDR, (void *)&k,
		   sizeof(k)) < 0)
	exit_log_errno("setsockopt() in init_socket()");

    /* bind the socket */
    if (bind(socks[0], (struct sockaddr *)&sin, sizeof(sin)) < 0)
	exit_log_errno("bind() failed in init_socket()");

    /* Get local interfaces */
    ifconf.ifc_len = sizeof(buf);
    ifconf.ifc_buf = buf;
    memset(buf, '\0', sizeof(buf));

    if (ioctl(socks[0], SIOCGIFCONF, &ifconf) == -1)
	exit_log_errno("ioctl() in init_socket()");

    /* ??? I'm not sure that socks[0] is ever used again. */

    /* This loop by Niels Provos */
    for (i = 1, j = 0; j < ifconf.ifc_len && i < maxsocks; j += IFNAMSIZ +
#if defined(__NetBSD__) || defined(__OpenBSD__)
	     buf[IFNAMSIZ + j]
#else
	     sizeof(struct sockaddr)
#endif
	     , i++)
    {
	sin2 = (struct sockaddr_in *) &buf[IFNAMSIZ + j];

	if (sin2->sin_family != AF_INET)
	{
	    i--;
	    continue;
	}

	/* ??? earlier, we seemed to think proto could be NULL */
	if ((socks[i] = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	    exit_log_errno("socket() in init_socket()");

	if (setsockopt(socks[i], SOL_SOCKET, SO_REUSEADDR, (void *)&k,
		       sizeof(k)) < 0)
	    exit_log_errno("setsockopt() in init_socket()");

	mksin(sin, sin2->sin_addr.s_addr, pluto_port);

	if (bind(socks[i], (struct sockaddr *)&sin, sizeof(sin)) < 0)
	    exit_log_errno("bind() in init_socket()");
	DBG(DBG_RAW | DBG_CONTROL, DBG_log("listening at %s", inet_ntoa(sin2->sin_addr)));
    }
    return i;
}

/*
 * This routine listens for incoming ISAKMP packets, kernel PF_KEY requests,
 * handles packet retransmitts and other events.
 */
void
call_server(void)
{
    int kernelfd;
    int sock[256];
    int numsocks;
    fd_set readfds;
    struct timeval tm;
    int i, maxfd, k;

    numsocks = init_socket(sock, (int) (sizeof(sock) / sizeof(*sock)));
    kernelfd = init_kernelfd();

    DBG(DBG_RAW | DBG_CONTROL,
	for (k = 0; k < numsocks; k++)
	    DBG_log("socket %d: %d", k, sock[k]);
	DBG_log("kernel socket: %d", kernelfd));

    maxfd = kernelfd;
    for (i = 1; i < numsocks; i++)
	if (sock[i] > maxfd)
	    maxfd = sock[i];
    maxfd++;	/* roof, not ceiling */

    while (1)
    {
	FD_ZERO(&readfds);
	for (k = 1; k < numsocks; k++)
	    FD_SET(sock[k], &readfds);
	FD_SET(kernelfd, &readfds);

	i = next_event();   /* Is there any pending event ? */
	if (i == -1)
	    i = select(maxfd, &readfds, NULL, NULL, NULL);
	else
	{
	    tm.tv_sec = i;
	    tm.tv_usec = 0;
	    i = select(maxfd, &readfds, NULL, NULL, &tm);
	}

	if (i == -1)
	    exit_log_errno("select() failed in call_server()");

	for (k = 1; k < numsocks; k++)
	{
	    if (FD_ISSET(sock[k], &readfds))
	    {
		DBG(DBG_CONTROL,
		    DBG_log("");
		    DBG_log("received packet"));
		comm_handle(kernelfd, sock[k]);
	    }
	}

	if (FD_ISSET(kernelfd, &readfds))
	{
	    DBG(DBG_CONTROL,
		DBG_log("");
		DBG_log("received kernel message"));
	    kernel_handle(kernelfd, sock[0]);
	}

	if (i == 0)   /* Event */
	{
	    DBG(DBG_CONTROL,
		DBG_log("");
		DBG_log("time to handle event"));
	    event_handle(kernelfd, sock[0]);
	}
    }
}
