/* This program is an implementation of the ISAKMP Internet Standard.
 * This file is:
 * Copyright (C) 1998 D. Hugh Redelmeier
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This file was written while the author was in Canada.
 *
 * You may contact the author by e-mail: hugh@mimosa.com
 */

/* Security Policy Data Base (such as it is) */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "constants.h"
#include "state.h"
#include "packet.h"
#include "preshared.h"
#include "defs.h"
#include "kernel.h"
#include "log.h"
#include "spdb.h"

#include "sha1.h"
#include "md5.h"
#include "crypto.h" /* requires sha1.h and md5.h */

#define AD(x) x, elemsof(x)	/* Array Description */

/**************** Oakely (main mode) SA database ****************/

/* arrays of attributes for transforms */

#if 0 /* NOT CURRENTLY USED */
static struct db_attr ot768desmd5[] = {
	{ OAKLEY_ENCRYPTION_ALGORITHM, OAKLEY_DES_CBC },
	{ OAKLEY_HASH_ALGORITHM, OAKLEY_MD5 },
	{ OAKLEY_AUTHENTICATION_METHOD, OAKLEY_PRESHARED_KEY },
	{ OAKLEY_GROUP_DESCRIPTION, OAKLEY_GROUP_MODP768 },
	{ OAKLEY_LIFE_TYPE, OAKLEY_LIFE_SECONDS },
	{ OAKLEY_LIFE_DURATION, OAKLEY_ISAKMP_SA_LIFETIME },
	};
#endif

static struct db_attr ot768des3md5[] = {
	{ OAKLEY_ENCRYPTION_ALGORITHM, OAKLEY_3DES_CBC },
	{ OAKLEY_HASH_ALGORITHM, OAKLEY_MD5 },
	{ OAKLEY_AUTHENTICATION_METHOD, OAKLEY_PRESHARED_KEY },
	{ OAKLEY_GROUP_DESCRIPTION, OAKLEY_GROUP_MODP768 },
	{ OAKLEY_LIFE_TYPE, OAKLEY_LIFE_SECONDS },
	{ OAKLEY_LIFE_DURATION, OAKLEY_ISAKMP_SA_LIFETIME },
	};

#if 0 /* NOT CURRENTLY USED */
static struct db_attr ot1024desmd5[] = {
	{ OAKLEY_ENCRYPTION_ALGORITHM, OAKLEY_DES_CBC },
	{ OAKLEY_HASH_ALGORITHM, OAKLEY_MD5 },
	{ OAKLEY_AUTHENTICATION_METHOD, OAKLEY_PRESHARED_KEY },
	{ OAKLEY_GROUP_DESCRIPTION, OAKLEY_GROUP_MODP1024 },
	{ OAKLEY_LIFE_TYPE, OAKLEY_LIFE_SECONDS },
	{ OAKLEY_LIFE_DURATION, OAKLEY_ISAKMP_SA_LIFETIME },
	};
#endif

static struct db_attr ot1024des3md5[] = {
	{ OAKLEY_ENCRYPTION_ALGORITHM, OAKLEY_3DES_CBC },
	{ OAKLEY_HASH_ALGORITHM, OAKLEY_MD5 },
	{ OAKLEY_AUTHENTICATION_METHOD, OAKLEY_PRESHARED_KEY },
	{ OAKLEY_GROUP_DESCRIPTION, OAKLEY_GROUP_MODP1024 },
	{ OAKLEY_LIFE_TYPE, OAKLEY_LIFE_SECONDS },
	{ OAKLEY_LIFE_DURATION, OAKLEY_ISAKMP_SA_LIFETIME },
	};

#if 0 /* NOT CURRENTLY USED */
static struct db_attr ot1024dessha[] = {
	{ OAKLEY_ENCRYPTION_ALGORITHM, OAKLEY_DES_CBC },
	{ OAKLEY_HASH_ALGORITHM, OAKLEY_SHA },
	{ OAKLEY_AUTHENTICATION_METHOD, OAKLEY_PRESHARED_KEY },
	{ OAKLEY_GROUP_DESCRIPTION, OAKLEY_GROUP_MODP1024 },
	{ OAKLEY_LIFE_TYPE, OAKLEY_LIFE_SECONDS },
	{ OAKLEY_LIFE_DURATION, OAKLEY_ISAKMP_SA_LIFETIME },
	};
#endif

static struct db_attr ot1024des3sha[] = {
	{ OAKLEY_ENCRYPTION_ALGORITHM, OAKLEY_3DES_CBC },
	{ OAKLEY_HASH_ALGORITHM, OAKLEY_SHA },
	{ OAKLEY_AUTHENTICATION_METHOD, OAKLEY_PRESHARED_KEY },
	{ OAKLEY_GROUP_DESCRIPTION, OAKLEY_GROUP_MODP1024 },
	{ OAKLEY_LIFE_TYPE, OAKLEY_LIFE_SECONDS },
	{ OAKLEY_LIFE_DURATION, OAKLEY_ISAKMP_SA_LIFETIME },
	};

/* We won't accept this, but by proposing it, we get to test
 * our rejection.  We better not propose it to an IKE daemon
 * that will accept it!
 */
#ifdef TEST_INDECENT_PROPOSAL

static struct db_attr ot1024destiger[] = {
	{ OAKLEY_ENCRYPTION_ALGORITHM, OAKLEY_DES_CBC },
	{ OAKLEY_HASH_ALGORITHM, OAKLEY_TIGER },
	{ OAKLEY_AUTHENTICATION_METHOD, OAKLEY_PRESHARED_KEY },
	{ OAKLEY_GROUP_DESCRIPTION, OAKLEY_GROUP_MODP1024 },
	{ OAKLEY_LIFE_TYPE, OAKLEY_LIFE_SECONDS },
	{ OAKLEY_LIFE_DURATION, OAKLEY_ISAKMP_SA_LIFETIME },
	};

static struct db_attr ot1024des3tiger[] = {
	{ OAKLEY_ENCRYPTION_ALGORITHM, OAKLEY_3DES_CBC },
	{ OAKLEY_HASH_ALGORITHM, OAKLEY_TIGER },
	{ OAKLEY_AUTHENTICATION_METHOD, OAKLEY_PRESHARED_KEY },
	{ OAKLEY_GROUP_DESCRIPTION, OAKLEY_GROUP_MODP1024 },
	{ OAKLEY_LIFE_TYPE, OAKLEY_LIFE_SECONDS },
	{ OAKLEY_LIFE_DURATION, OAKLEY_ISAKMP_SA_LIFETIME },
	};

#endif

/* table of transforms, in preference order */
static struct db_trans oakley_trans[] = {
#ifdef TEST_INDECENT_PROPOSAL
	{ KEY_IKE, AD(ot1024destiger) },
#endif
	{ KEY_IKE, AD(ot1024des3sha) },
	{ KEY_IKE, AD(ot1024des3sha) },
	{ KEY_IKE, AD(ot1024des3md5) },
	{ KEY_IKE, AD(ot768des3md5) },
    };

/* array of proposals to be conjoined (can only be one for Oakley) */

static struct db_prop oakley_pc[] =
    { { PROTO_ISAKMP, AD(oakley_trans) } };

/* array of proposal conjuncts (can only be one) */

static struct db_prop_conj oakley_props[] = { { AD(oakley_pc) } };

/* the sadb entry */
struct db_sa oakley_sadb = { AD(oakley_props) };

/**************** IPsec (quick mode) SA database ****************/

/* arrays of attributes for transforms
 * "espa" means ESP with auth.
 * The db_ attr arrays are carefully designed so that skipping the
 * first db_attr will skip the auth.
 */

static struct db_attr espa_attr_tunnel[] = {
    { AUTH_ALGORITHM, AUTH_ALGORITHM_HMAC_MD5 },	/* skip if AH too */
    { SA_LIFE_TYPE, SA_LIFE_TYPE_SECONDS },
    { SA_LIFE_DURATION, SA_LIFE_DURATION_DEFAULT },
    { ENCAPSULATION_MODE, ENCAPSULATION_MODE_TUNNEL },
    };

static struct db_attr espa_attr_transport[] = {
    { AUTH_ALGORITHM, AUTH_ALGORITHM_HMAC_MD5 },	/* skip if AH too */
    { SA_LIFE_TYPE, SA_LIFE_TYPE_SECONDS },
    { SA_LIFE_DURATION, SA_LIFE_DURATION_DEFAULT },
    { ENCAPSULATION_MODE, ENCAPSULATION_MODE_TRANSPORT },
    };

static struct db_attr ah_attr_tunnel[] = {
    { SA_LIFE_TYPE, SA_LIFE_TYPE_SECONDS },
    { SA_LIFE_DURATION, SA_LIFE_DURATION_DEFAULT },
    { ENCAPSULATION_MODE, ENCAPSULATION_MODE_TUNNEL },
    };

static struct db_attr ah_attr_transport[] = {
    { SA_LIFE_TYPE, SA_LIFE_TYPE_SECONDS },
    { SA_LIFE_DURATION, SA_LIFE_DURATION_DEFAULT },
    { ENCAPSULATION_MODE, ENCAPSULATION_MODE_TRANSPORT },
    };

/* arrays of transforms, each in in preference order */

static struct db_trans espa_trans_tunnel[] = {
    { ESP_3DES, AD(espa_attr_tunnel) },
    };

static struct db_trans espa_trans_transport[] = {
    { ESP_3DES, AD(espa_attr_transport) },
    };

static struct db_trans esp_trans_tunnel[] = {
    { ESP_3DES, espa_attr_tunnel+1, elemsof(espa_attr_tunnel)-1 },	/* no auth! */
    };

static struct db_trans esp_trans_transport[] = {
    { ESP_3DES, espa_attr_transport+1, elemsof(espa_attr_transport)-1 },	/* no auth! */
    };

static struct db_trans ah_trans_tunnel[] = {
    { AH_MD5, AD(ah_attr_tunnel) },
    };

static struct db_trans ah_trans_transport[] = {
    { AH_MD5, AD(ah_attr_transport) },
    };

/* arrays of proposals to be conjoined */

static struct db_prop ah_pc_tunnel[] = {
    { PROTO_IPSEC_AH, AD(ah_trans_tunnel) },
    };

static struct db_prop ah_pc_transport[] = {
    { PROTO_IPSEC_AH, AD(ah_trans_transport) },
    };

static struct db_prop esp_pc_tunnel[] = {
    { PROTO_IPSEC_ESP, AD(espa_trans_tunnel) },
    };

static struct db_prop esp_pc_transport[] = {
    { PROTO_IPSEC_ESP, AD(espa_trans_transport) },
    };

static struct db_prop ah_esp_pc_tunnel[] = {
    { PROTO_IPSEC_AH, AD(ah_trans_tunnel) },
    { PROTO_IPSEC_ESP, AD(esp_trans_tunnel) },
    };

static struct db_prop ah_esp_pc_transport[] = {
    { PROTO_IPSEC_AH, AD(ah_trans_transport) },
    { PROTO_IPSEC_ESP, AD(esp_trans_transport) },
    };

/* arrays of proposal alternatives (each element is a conjunction) */

static struct db_prop_conj ah_tunnel_props[] =
    { { AD(ah_pc_tunnel) } };

static struct db_prop_conj ah_transport_props[] =
    { { AD(ah_pc_transport) } };

static struct db_prop_conj esp_tunnel_props[] =
    { { AD(esp_pc_tunnel) } };

static struct db_prop_conj esp_transport_props[] =
    { { AD(esp_pc_transport) } };

static struct db_prop_conj ah_esp_tunnel_props[] =
    { { AD(ah_esp_pc_tunnel) } };

static struct db_prop_conj ah_esp_transport_props[] =
    { { AD(ah_esp_pc_transport) } };

/* The IPsec sadb is subscripted by a bitset (subset of goal)
 * with members from { GOAL_ENCRYPT, GOAL_AUTHENTICATE, GOAL_TUNNEL }
 */
struct db_sa ipsec_sadb[1 << 3] = {
    { NULL, 0 },	/* none */
    { AD(esp_transport_props) },	/* GOAL_ENCRYPT */
    { AD(ah_transport_props) },	/* GOAL_AUTHENTICATE */
    { AD(ah_esp_transport_props) },	/* GOAL_ENCRYPT+GOAL_AUTHENTICATE */
    { NULL, 0 },	/* GOAL_TUNNEL */
    { AD(esp_tunnel_props) },	/* GOAL_ENCRYPT+GOAL_TUNNEL */
    { AD(ah_tunnel_props) },	/* GOAL_AUTHENTICATE+GOAL_TUNNEL */
    { AD(ah_esp_tunnel_props) },	/* GOAL_ENCRYPT+GOAL_AUTHENTICATE+GOAL_TUNNEL */
    };

#undef AD

/* Output an SA, as described by a db_sa.
 * This has the side-effect of allocating SPIs for us.
 */
bool
out_sa(
    pb_stream *outs,
    struct db_sa *sadb,
    struct state *st,
    bool oakley_mode,
    u_int8_t np)
{
    pb_stream sa_pbs;
    int pcn;
    bool ah_spi_generated = FALSE;
    bool esp_spi_generated = FALSE;

    /* SA header out */
    {
	struct isakmp_sa sa;

	sa.isasa_np = np;
	st->st_doi = sa.isasa_doi = ISAKMP_DOI_IPSEC;	/* all we know */
	if (!out_struct(&sa, &isakmp_sa_desc, outs, &sa_pbs))
	    return FALSE;
    }

    /* within SA: situation out */
    st->st_situation = SIT_IDENTITY_ONLY;
    if (!out_struct(&st->st_situation, &ipsec_sit_desc, &sa_pbs, NULL))
	return FALSE;

    /* within SA: Proposal Payloads
     *
     * Multiple Proposals with the same number are simultaneous
     * (conjuncts) and must deal with different protocols (AH or ESP).
     * Proposals with different numbers are alternatives (disjuncts),
     * in preference order.
     * Proposal numbers must be monotonic.
     * See draft-ietf-ipsec-isakmp-09.txt 4.2
     */

    for (pcn = 0; pcn != sadb->prop_conj_cnt; pcn++)
    {
	struct db_prop_conj *pc = &sadb->prop_conjs[pcn];
	int pn;

	for (pn = 0; pn != pc->prop_cnt; pn++)
	{
	    struct db_prop *p = &pc->props[pn];
	    pb_stream proposal_pbs;
	    struct isakmp_proposal proposal;
	    struct_desc *trans_desc;
	    struct_desc *attr_desc;
	    enum_desc **attr_val_descs;
	    int tn;

	    /* Proposal header */
	    proposal.isap_np = pcn == sadb->prop_conj_cnt-1 && pn == pc->prop_cnt-1
		? ISAKMP_NEXT_NONE : ISAKMP_NEXT_P;
	    proposal.isap_proposal = pcn;
	    proposal.isap_protoid = p->protoid;
	    proposal.isap_spisize = oakley_mode? 0 : IPSEC_DOI_SPI_SIZE;
	    proposal.isap_notrans = p->trans_cnt;
	    if (!out_struct(&proposal, &isakmp_proposal_desc, &sa_pbs, &proposal_pbs))
		return FALSE;

	    /* Per-protocols stuff:
	     * Set trans_desc.
	     * Set attr_desc.
	     * Set attr_val_descs.
	     * If not oakley_mode, emit SPI.
	     * We allocate SPIs on demand.
	     * All ESPs in an SA will share a single SPI.
	     * All AHs in an SAwill share a single SPI.
	     * AHs' SPI will be distinct from ESPs'.
	     * This latter is needed because KLIPS doesn't
	     * use the protocol when looking up a (dest, protocol, spi).
	     * ??? If multiple ESPs are composed, how should their SPIs
	     * be allocated?
	     */
	    {
		u_char *spi_ptr = NULL;
		bool *spi_generated;

		switch (p->protoid)
		{
		case PROTO_ISAKMP:
		    trans_desc = &isakmp_isakmp_transform_desc;
		    attr_desc = &isakmp_oakley_attribute_desc;
		    attr_val_descs = oakley_attr_val_descs;
		    /* no SPI needed */
		    break;
		case PROTO_IPSEC_AH:
		    trans_desc = &isakmp_ah_transform_desc;
		    attr_desc = &isakmp_ipsec_attribute_desc;
		    attr_val_descs = ipsec_attr_val_descs;
		    spi_ptr = st->st_ah.our_spi;
		    spi_generated = &ah_spi_generated;
		    break;
		case PROTO_IPSEC_ESP:
		    trans_desc = &isakmp_esp_transform_desc;
		    attr_desc = &isakmp_ipsec_attribute_desc;
		    attr_val_descs = ipsec_attr_val_descs;
		    spi_ptr = st->st_esp.our_spi;
		    spi_generated = &esp_spi_generated;
		    break;
		default:
		    assert(FALSE);
		}
		if (spi_ptr != NULL)
		{
		    if (!*spi_generated)
		    {
			get_ipsec_spi(spi_ptr);
			*spi_generated = TRUE;
		    }
		    if (!out_raw(spi_ptr, IPSEC_DOI_SPI_SIZE
		    , &proposal_pbs, "SPI"))
			return FALSE;
		}
	    }

	    /* within proposal: Transform Payloads */
	    for (tn = 0; tn != p->trans_cnt; tn++)
	    {
		struct db_trans *t = &p->trans[tn];
		pb_stream trans_pbs;
		struct isakmp_transform trans;
		int an;
		int pfs_group = 0;	/* default: NONE */

		trans.isat_np = (tn == p->trans_cnt - 1)
		    ? ISAKMP_NEXT_NONE : ISAKMP_NEXT_T;
		trans.isat_transnum = tn;
		trans.isat_transid = t->transid;
		if (!out_struct(&trans, trans_desc, &proposal_pbs, &trans_pbs))
		    return FALSE;

		/* within tranform: Attributes */
		for (an = 0; an != t->attr_cnt; an++)
		{
		    struct db_attr *a = &t->attrs[an];
		    struct isakmp_attribute attr;

		    attr.isaat_af_type = a->type | ISAKMP_ATTR_AF_TV;
		    attr.isaat_lv = a->val;
		    if (!out_struct(&attr, attr_desc, &trans_pbs, NULL))
			return FALSE;

		    if (attr_desc == &isakmp_ipsec_attribute_desc
		    && a->type == GROUP_DESCRIPTION)
			pfs_group = a->val;
		    DBG(DBG_PARSING,
			enum_desc *d = attr_val_descs[a->type];

			if (d != NULL)
			    DBG_log("  [%u is %s]",
				a->val, enum_show(d, a->val)));
		}
		if (st->st_pfs_group == -1)
		    st->st_pfs_group = pfs_group;
		if (st->st_pfs_group != pfs_group)
		{
		    /* can't happen? */
		    log("inconsistent GROUP_DESCRIPTION in our SA");
		    return FALSE;
		}
		close_output_pbs(&trans_pbs);
	    }
	    close_output_pbs(&proposal_pbs);
	}
	/* end of a conjunction of proposals */
    }
    close_output_pbs(&sa_pbs);
    return TRUE;
}

/* Parse the body of an ISAKMP SA Payload (i.e. Phase 1 / Main Mode).
 * Various shortcuts are taken.  In particular, the policy, such as
 * it is, is hardwired.
 *
 * If r_sa is non-NULL, the body of an SA representing the selected
 * proposal is emitted.
 *
 * If "selection" is true, the SA is supposed to represent the
 * single tranform that the peer has accepted.
 * ??? We only check that it is acceptable, not that it is one that we offered!
 *
 * Only IPsec DOI is accepted (what is the ISAKMP DOI?).
 * Error response is rudimentary.
 *
 * This routine is used by main_inI1_outR1() and main_inR1_outI2().
 */
bool
parse_isakmp_sa_body(
    pb_stream *sa_pbs,	/* body of input SA Payload */
    const struct isakmp_sa *sa,	/* header of input SA Payload */
    pb_stream *r_sa_pbs,	/* if non-NULL, where to emit winning SA */
    bool selection,	/* if this SA is a selection, only one tranform can appear */
    struct state *st,	/* current state object */
    const u_char my_cookie[COOKIE_SIZE],
    const u_char his_cookie[COOKIE_SIZE],
    const struct sockaddr *sap)	/* network address */
{
    u_int32_t ipsecdoisit;
    pb_stream proposal_pbs;
    struct isakmp_proposal proposal;
    unsigned no_trans_left;
    int last_transnum;

    /* DOI */
    if (sa->isasa_doi != ISAKMP_DOI_IPSEC)
    {
	log("Unknown/unsupported DOI %d from %s",
	    enum_show(&doi_desc, sa->isasa_doi),
	    show_sa(sap));
	/* XXX Could send notification back */
	return FALSE;
    }

    /* Situation */
    if (!in_struct(&ipsecdoisit, &ipsec_sit_desc, sa_pbs, NULL))
    {
	return FALSE;
    }

    if (ipsecdoisit != SIT_IDENTITY_ONLY)
    {
	log("unsupported IPsec DOI situation (%s) received from %s",
	    bitnamesof(sit_bit_names, ipsecdoisit),
	    show_sa(sap));
	/* XXX Could send notification back */
	return FALSE;
    }

    /* The rules for ISAKMP SAs are scattered.
     * draft-ietf-ipsec-isakmp-oakley-07.txt section 5 says that there
     * can only be SA, and it can have only one proposal in it.
     * There may well be multiple transforms.
     */
    if (!in_struct(&proposal, &isakmp_proposal_desc, sa_pbs, &proposal_pbs))
	return FALSE;

    if (proposal.isap_np != ISAKMP_NEXT_NONE)
    {
	log("Proposal Payload must be alone in Oakely SA; found %s in proposal from %s",
	    enum_show(&payload_desc, proposal.isap_np),
	    show_sa(sap));
    }

    if (proposal.isap_protoid != PROTO_ISAKMP)
    {
	log("unexpected protoid (%s) in Oakley proposal from %s",
	    enum_show(&protocol_desc, proposal.isap_protoid),
	    show_sa(sap));
	return FALSE;
    }

    if (proposal.isap_spisize == 0)
    {
	/* empty (0) SPI -- fine */
    }
    else if (proposal.isap_spisize == COOKIE_SIZE)
    {
	u_char cookie[COOKIE_SIZE];

	if (!in_raw(cookie, COOKIE_SIZE, &proposal_pbs, "Oakley SPI"))
	    return FALSE;

	if (memcmp(cookie, his_cookie, COOKIE_SIZE) != 0)
	{
	    log("invalid SPI in Oakley proposal from %s",
		show_sa(sap));
	    return FALSE;
	}
    }
    else
    {
	log("invalid SPI size (%u) in Oakley proposal from %s",
	    (unsigned)proposal.isap_spisize, show_sa(sap));
	return FALSE;
    }

    if (selection && proposal.isap_notrans != 1)
    {
	log("a single transform is required in a selecting Oakley Proposal; %u from %s",
	    (unsigned)proposal.isap_notrans, show_sa(sap));
	return FALSE;
    }

    /* for each transform payload... */

    last_transnum = -1;
    no_trans_left = proposal.isap_notrans;
    for (;; )
    {
	pb_stream trans_pbs;
	u_char *attr_start;
	size_t attr_len;
	struct isakmp_transform trans;
	unsigned long seen_attrs = 0;
	u_int16_t life_type;
	struct oakley_trans_attrs ta;
	char ugh[100];	/* mailbox for diagnostics */

	ugh[0] = '\0';	/* empty mailbox */

	/* initialize only optional field in ta */
	ta.expire =  SA_LIFE_DURATION_DEFAULT;	/* When this SA expires (seconds) */

	if (no_trans_left == 0)
	{
	    log("number of Transform Payloads disagrees with Oakley Proposal Payload from %s",
		show_sa(sap));
	    return FALSE;
	}

	if (!in_struct(&trans, &isakmp_isakmp_transform_desc, &proposal_pbs, &trans_pbs))
	    return FALSE;

	if (trans.isat_transnum <= last_transnum)
	{
	    /* picky, picky, picky */
	    log("transform numbers were not monotonically increasing in Oakley Proposal from %s",
		show_sa(sap));
	    return FALSE;
	}
	last_transnum = trans.isat_transnum;

	if (trans.isat_transid != KEY_IKE)
	{
	    log("expected KEY_IKE but found %s in Oakley Transform from %s",
		enum_show(&isakmp_transformid_desc, trans.isat_transid),
		show_sa(sap));
	    return FALSE;
	}

	attr_start = trans_pbs.cur;
	attr_len = pbs_left(&trans_pbs);

	/* process all the attributes that make up the transform */

	while (pbs_left(&trans_pbs) != 0)
	{
	    struct isakmp_attribute a;
	    pb_stream attr_pbs;
	    enum_desc *vdesc;

	    if (!in_struct(&a, &isakmp_oakley_attribute_desc, &trans_pbs, &attr_pbs))
		return FALSE;

	    DBG(DBG_PARSING,
		enum_desc *d = oakley_attr_val_descs[a.isaat_af_type & ISAKMP_ATTR_RTYPE_MASK];

		if (d != NULL)
		    DBG_log("  [%u is %s]",
			a.isaat_lv, enum_show(d, a.isaat_lv)));

	    assert((a.isaat_af_type & ISAKMP_ATTR_RTYPE_MASK) < 32);

	    if (seen_attrs & 1ul << (a.isaat_af_type & ISAKMP_ATTR_RTYPE_MASK))
	    {
		log("repeated %s attribute in Oakley Transform %u from %s",
		    enum_show(&oakley_attr_desc, a.isaat_af_type),
		    trans.isat_transnum, show_sa(sap));
		return FALSE;
	    }

	    seen_attrs |= 1ul << (a.isaat_af_type & ISAKMP_ATTR_RTYPE_MASK);

	    vdesc  = oakley_attr_val_descs[a.isaat_af_type & ISAKMP_ATTR_RTYPE_MASK];
	    if (vdesc != NULL)
	    {
		if (enum_name(vdesc, a.isaat_lv) == NULL)
		{
		    log("invalid value %u for attribute %s in Oakley Transform from %s",
			a.isaat_lv, enum_show(&oakley_attr_desc, a.isaat_af_type),
			show_sa(sap));
		    return FALSE;
		}
	    }

	    switch (a.isaat_af_type)
	    {
		case OAKLEY_ENCRYPTION_ALGORITHM | ISAKMP_ATTR_AF_TV:
		    switch (a.isaat_lv)
		    {
		    case OAKLEY_DES_CBC:
		    case OAKLEY_3DES_CBC:
			ta.encrypt = a.isaat_lv;
			ta.encrypter = &oakley_encrypter[a.isaat_lv];
			break;
		    default:
			snprintf(ugh, sizeof(ugh), "%s is not supported",
			enum_show(&oakley_enc_desc, a.isaat_lv));
		    }
		    break;

		case OAKLEY_HASH_ALGORITHM | ISAKMP_ATTR_AF_TV:
		    switch (a.isaat_lv)
		    {
		    case OAKLEY_MD5:
		    case OAKLEY_SHA:
			ta.hash = a.isaat_lv;
			ta.hasher = &oakley_hasher[a.isaat_lv];
			break;
		    default:
			snprintf(ugh, sizeof(ugh), "%s is not supported",
			    enum_show(&oakley_hash_desc, a.isaat_lv));
		    }
		    break;

		case OAKLEY_AUTHENTICATION_METHOD | ISAKMP_ATTR_AF_TV:
		    if (a.isaat_lv != OAKLEY_PRESHARED_KEY)
		    {
			snprintf(ugh, sizeof(ugh), "only OAKLEY_PRESHARED_KEY supported");
			break;
		    }
		    else
		    {
			int key_len;
			u_char *key = get_preshared_key(st, &key_len);

			if (key == (u_char *) NULL)
			    snprintf(ugh, sizeof(ugh), "Can't authenticate: no preshared key");
			else
			    free(key);
			ta.auth = a.isaat_lv;
		    }
		    break;

		case OAKLEY_GROUP_DESCRIPTION | ISAKMP_ATTR_AF_TV:
		    switch (a.isaat_lv)
		    {
			case OAKLEY_GROUP_MODP768:
			    ta.modp_modulus = &modp768_modulus;
			    break;
			case OAKLEY_GROUP_MODP1024:
			    ta.modp_modulus = &modp1024_modulus;
			    break;
			default:
			    snprintf(ugh, sizeof(ugh), "only OAKLEY_GROUP_MODP768 and OAKLEY_GROUP_MODP1024 supported");
			    break;
		    }
		    ta.group = a.isaat_lv;
		    break;

		case OAKLEY_LIFE_TYPE | ISAKMP_ATTR_AF_TV:
		    life_type = a.isaat_lv;
		    break;

		case OAKLEY_LIFE_DURATION | ISAKMP_ATTR_AF_TV:
		case OAKLEY_LIFE_DURATION | ISAKMP_ATTR_AF_TLV:
		    {
		    u_int32_t val;

		    if ((seen_attrs & 1ul << OAKLEY_LIFE_DURATION) == 0)
		    {
			snprintf(ugh, sizeof(ugh),
			    "OAKLEY_LIFE_DURATION attribute not preceded by OAKLEY_LIFE_TYPE attribute");
			break;
		    }
		    seen_attrs &=  ~(1ul << OAKLEY_LIFE_DURATION | 1ul << OAKLEY_LIFE_TYPE);

		    if ((a.isaat_af_type & ISAKMP_ATTR_AF_MASK) == ISAKMP_ATTR_AF_TLV)
		    {
			/* handle long form of attribute, but limited to 32 bits */

			/* ignore leading zeros */
			while (pbs_left(&attr_pbs) != 0 && *attr_pbs.cur == '\0')
			    attr_pbs.cur++;

			if (pbs_left(&attr_pbs) > sizeof(val))
			{
			    snprintf(ugh, sizeof(ugh), "OAKLEY_LIFE_DURATION attribute too large (more than 32 bits)");
			    break;
			}

			/* decode number */
			val = 0;
			while (pbs_left(&attr_pbs) != 0)
			    val = (val << BITS_PER_BYTE) | *attr_pbs.cur++;
		    }
		    else
		    {
			val = a.isaat_lv;
		    }

		    switch (life_type)
		    {
			case OAKLEY_LIFE_SECONDS:
			    if (val <= OAKLEY_ISAKMP_SA_LIFETIME_MINIMUM
			    || val >= OAKLEY_ISAKMP_SA_LIFETIME_MAXIMUM)
				snprintf(ugh, sizeof(ugh),
				    "OAKLEY_LIFE_TYPE_SECONDS out of range [%d, %d]",
				     OAKLEY_ISAKMP_SA_LIFETIME_MINIMUM,
				     OAKLEY_ISAKMP_SA_LIFETIME_MAXIMUM
				     );
			    ta.expire = val;
			    break;
			case OAKLEY_LIFE_KILOBYTES:
			    ta.exhaust = val;
			    break;
			default:
			    assert(FALSE);
		    }
		    }
		    break;

#if 0 /* not yet supported */
		case OAKLEY_GROUP_TYPE | ISAKMP_ATTR_AF_TV:
		case OAKLEY_PRF | ISAKMP_ATTR_AF_TV:
		case OAKLEY_KEY_LENGTH | ISAKMP_ATTR_AF_TV:
		case OAKLEY_FIELD_SIZE | ISAKMP_ATTR_AF_TV:

		case OAKLEY_GROUP_PRIME | ISAKMP_ATTR_AF_TV:
		case OAKLEY_GROUP_PRIME | ISAKMP_ATTR_AF_TLV:
		case OAKLEY_GROUP_GENERATOR_ONE | ISAKMP_ATTR_AF_TV:
		case OAKLEY_GROUP_GENERATOR_ONE | ISAKMP_ATTR_AF_TLV:
		case OAKLEY_GROUP_GENERATOR_TWO | ISAKMP_ATTR_AF_TV:
		case OAKLEY_GROUP_GENERATOR_TWO | ISAKMP_ATTR_AF_TLV:
		case OAKLEY_GROUP_CURVE_A | ISAKMP_ATTR_AF_TV:
		case OAKLEY_GROUP_CURVE_A | ISAKMP_ATTR_AF_TLV:
		case OAKLEY_GROUP_CURVE_B | ISAKMP_ATTR_AF_TV:
		case OAKLEY_GROUP_CURVE_B | ISAKMP_ATTR_AF_TLV:
		case OAKLEY_GROUP_ORDER | ISAKMP_ATTR_AF_TV:
		case OAKLEY_GROUP_ORDER | ISAKMP_ATTR_AF_TLV:
#endif
		default:
		    snprintf(ugh, sizeof(ugh), "unsupported OAKLEY attribute");
		    break;
	    }

	    if (ugh[0] != '\0')
	    {
		log("%s.  Attribute %s in Message from %s",
		    ugh,
		    enum_show(&oakley_attr_desc, a.isaat_af_type),
		    show_sa(sap));
		break;
	    }
	}

	if (ugh[0] == '\0')
	{
	    /* a little more checking is in order */
	    {
		unsigned long need
		    = 1ul << OAKLEY_ENCRYPTION_ALGORITHM
		    | 1ul << OAKLEY_HASH_ALGORITHM
		    | 1ul << OAKLEY_AUTHENTICATION_METHOD
		    | 1ul << OAKLEY_GROUP_DESCRIPTION;

		if ((seen_attrs & need) != need)
		{
		    log("missing mandatory attributes in Oakley Transform %u from %s",
			trans.isat_transnum, show_sa(sap));
		    return FALSE;
		}
	    }
	    /* We must have liked this transform.
	     * Lets finish early and leave.
	     */

	    DBG(DBG_PARSING | DBG_CRYPT,
		DBG_log("Oakley Transform %u accepted", trans.isat_transnum));

	    if (r_sa_pbs != NULL)
	    {
		struct isakmp_proposal r_proposal = proposal;
		pb_stream r_proposal_pbs;
		struct isakmp_transform r_trans = trans;
		pb_stream r_trans_pbs;

		/* Situation */
		if (!out_struct(&ipsecdoisit, &ipsec_sit_desc, r_sa_pbs, NULL))
		    assert(FALSE);

		/* Proposal */
#ifdef EMIT_ISAKMP_SPI
		r_proposal.isap_spisize = COOKIE_SIZE;
#else
		r_proposal.isap_spisize = 0;
#endif
		r_proposal.isap_notrans = 1;
		if (!out_struct(&r_proposal, &isakmp_proposal_desc, r_sa_pbs, &r_proposal_pbs))
		    assert(FALSE);

		/* SPI */
#ifdef EMIT_ISAKMP_SPI
		if (!out_raw(my_cookie, COOKIE_SIZE, &r_proposal_pbs, "SPI"))
		    assert(FALSE);
		r_proposal.isap_spisize = COOKIE_SIZE;
#else
		/* none (0) */
#endif

		/* Transform */
		r_trans.isat_np = ISAKMP_NEXT_NONE;
		if (!out_struct(&r_trans, &isakmp_isakmp_transform_desc, &r_proposal_pbs, &r_trans_pbs))
		    assert(FALSE);

		if (!out_raw(attr_start, attr_len, &r_trans_pbs, "attributes"))
		    assert(FALSE);
		close_output_pbs(&r_trans_pbs);
		close_output_pbs(&r_proposal_pbs);
		close_output_pbs(r_sa_pbs);
	    }

	    /* if selection, save proposal in state */
	    if (selection)
	    {
		st->st_proposal = clone_bytes(proposal_pbs.start, pbs_room(&proposal_pbs),
		    "proposal in parse_isakmp_sa_body");
		DBG(DBG_CONTROL | DBG_CRYPT,
		    DBG_log("copying %d bytes of proposal into state object",
			pbs_room(&proposal_pbs)));
	    }
	    /* copy over the results */
	    st->st_oakley = ta;
	    return TRUE;
	}

	/* on to next transform */
	no_trans_left--;

	if (trans.isat_np == ISAKMP_NEXT_NONE)
	{
	    if (no_trans_left != 0)
	    {
		log("number of Transform Payloads disagrees with Oakley Proposal Payload from %s",
		    show_sa(sap));
		return FALSE;
	    }
	    break;
	}
	if (trans.isat_np != ISAKMP_NEXT_T)
	{
	    log("unexpected %s payload in Oakley Proposal from %s",
		enum_show(&payload_desc, proposal.isap_np), show_sa(sap));
	    return FALSE;
	}
    }
    log("no acceptable Oakley Transform from %s", show_sa(sap));
    return FALSE;
}

/* Parse the body of an IPsec SA Payload (i.e. Phase 2 / Quick Mode).
 *
 * The main routine is parse_ipsec_sa_body; other functions defined
 * between here and there are just helpers.
 *
 * Various shortcuts are taken.  In particular, the policy, such as
 * it is, is hardwired.
 *
 * If r_sa is non-NULL, the body of an SA representing the selected
 * proposal is emitted into it.
 *
 * If "selection" is true, the SA is supposed to represent the
 * single tranform that the peer has accepted.
 * ??? We only check that it is acceptable, not that it is one that we offered!
 *
 * Only IPsec DOI is accepted (what is the ISAKMP DOI?).
 * Error response is rudimentary.
 *
 * Since all ISAKMP groups in all SA Payloads must match, st->st_pfs_group
 * holds this across multiple payloads.
 * -1 signifies not yet "set"; 0 signifies NONE.
 *
 * This routine is used by quick_inI1_outR1() and quick_inR1_outI2().
 */

static struct ipsec_trans_attrs null_ipsec_trans_attrs = {
    0,					/* transid (NULL, for now) */
    { 0 },				/* spi */
    SA_LIFE_DURATION_DEFAULT,		/* life_seconds */
    SA_LIFE_DURATION_K_DEFAULT,		/* life_kilobytes */
    ENCAPSULATION_MODE_UNSPECIFIED,	/* encapsulation */
    AUTH_ALGORITHM_NONE,		/* auth */
    0,					/* key_len */
    0,					/* key_rounds */
};

static bool
parse_ipsec_transform(
    struct isakmp_transform *trans,
    struct ipsec_trans_attrs *attrs,
    pb_stream *prop_pbs,
    pb_stream *trans_pbs,
    struct_desc *trans_desc,
    int previous_transnum,	/* or -1 if none */
    bool is_last,
    struct state *st,	/* current state object */
    const struct sockaddr *sap
    )
{
    unsigned long seen_attrs = 0;
    u_int16_t life_type;
    int pfs_group = 0;

    if (!in_struct(trans, trans_desc, prop_pbs, trans_pbs))
	return FALSE;

    if (trans->isat_transnum <= previous_transnum)
    {
	log("transform numbers in proposal from %s are not monotonically increasing",
	    show_sa(sap));
	return FALSE;
    }

    switch (trans->isat_np)
    {
	case ISAKMP_NEXT_T:
	    if (is_last)
	    {
		log("Proposal Payload from %s has more transforms than specified",
		    show_sa(sap));
		return FALSE;
	    }
	    break;
	case ISAKMP_NEXT_NONE:
	    if (!is_last)
	    {
		log("Proposal Payload from %s has fewer transforms than specified",
		    show_sa(sap));
		return FALSE;
	    }
	    break;
	default:
	    log("expecting Transform Payload, but found %s in Proposal from %s",
		enum_show(&payload_desc, trans->isat_np), show_sa(sap));
	    return FALSE;
    }
    
    *attrs = null_ipsec_trans_attrs;
    attrs->transid = trans->isat_transid;

    while (pbs_left(trans_pbs) != 0)
    {
	struct isakmp_attribute a;
	pb_stream attr_pbs;
	enum_desc *vdesc;

	if (!in_struct(&a, &isakmp_ipsec_attribute_desc, trans_pbs, &attr_pbs))
	    return FALSE;

	assert((a.isaat_af_type & ISAKMP_ATTR_RTYPE_MASK) < 32);

	if (seen_attrs & 1ul << (a.isaat_af_type & ISAKMP_ATTR_RTYPE_MASK))
	{
	    log("repeated %s attribute in IPsec Transform %u from %s",
		enum_show(&ipsec_attr_desc, a.isaat_af_type),
		trans->isat_transnum, show_sa(sap));
	    return FALSE;
	}

	seen_attrs |= 1ul << (a.isaat_af_type & ISAKMP_ATTR_RTYPE_MASK);

	vdesc  = ipsec_attr_val_descs[a.isaat_af_type & ISAKMP_ATTR_RTYPE_MASK];
	if (vdesc != NULL)
	{
	    if (enum_name(vdesc, a.isaat_lv) == NULL)
	    {
		log("invalid value %u for attribute %s in IPsec Transform from %s",
		    a.isaat_lv, enum_show(&ipsec_attr_desc, a.isaat_af_type),
		    show_sa(sap));
		return FALSE;
	    }
	    DBG(DBG_PARSING,
		DBG_log("  [%u is %s]",
		    a.isaat_lv, enum_show(vdesc, a.isaat_lv)));
	}

	switch (a.isaat_af_type)
	{
	    case SA_LIFE_TYPE | ISAKMP_ATTR_AF_TV:
		life_type = a.isaat_lv;
		break;
	    case SA_LIFE_DURATION | ISAKMP_ATTR_AF_TV:
	    case SA_LIFE_DURATION | ISAKMP_ATTR_AF_TLV:
		{
		u_int32_t val;

		if ((seen_attrs & 1ul << SA_LIFE_DURATION) == 0)
		{
		    log("SA_LIFE_DURATION IPsec attribute not preceded by SA_LIFE_TYPE attribute in message from %s",
			show_sa(sap));
		    return FALSE;
		}
		seen_attrs &=  ~(1ul << SA_LIFE_DURATION | 1ul << SA_LIFE_TYPE);

		if ((a.isaat_af_type & ISAKMP_ATTR_AF_MASK) == ISAKMP_ATTR_AF_TLV)
		{
		    /* handle long form of attribute, but limited to 32 bits */

		    /* ignore leading zeros */
		    while (pbs_left(&attr_pbs) != 0 && *attr_pbs.cur == '\0')
			attr_pbs.cur++;

		    if (pbs_left(&attr_pbs) > sizeof(val))
		    {
			/* "clamp" too large value to max representable value */
			val = ~(u_int32_t) 0;
		    }
		    else
		    {
			/* decode number */
			val = 0;
			while (pbs_left(&attr_pbs) != 0)
			    val = (val << BITS_PER_BYTE) | *attr_pbs.cur++;
		    }
		}
		else
		{
		    val = a.isaat_lv;
		}

		switch (life_type)
		{
		    case SA_LIFE_TYPE_SECONDS:
			attrs->life_seconds = val;
			break;
		    case SA_LIFE_TYPE_KBYTES:
			attrs->life_kilobytes = val;
			break;
		    default:
			assert(FALSE);
		}
		}
		break;
	    case GROUP_DESCRIPTION | ISAKMP_ATTR_AF_TV:
		pfs_group = a.isaat_lv;
		break;
	    case ENCAPSULATION_MODE | ISAKMP_ATTR_AF_TV:
		attrs->encapsulation = a.isaat_lv;
		break;
	    case AUTH_ALGORITHM | ISAKMP_ATTR_AF_TV:
		attrs->auth = a.isaat_lv;
		break;
	    case KEY_LENGTH | ISAKMP_ATTR_AF_TV:
		attrs->key_len = a.isaat_lv;
		break;
	    case KEY_ROUNDS | ISAKMP_ATTR_AF_TV:
		attrs->key_rounds = a.isaat_lv;
		break;
#if 0 /* not yet implemented */
	    case COMPRESS_DICT_SIZE | ISAKMP_ATTR_AF_TV:
		break;
	    case COMPRESS_PRIVATE_ALG | ISAKMP_ATTR_AF_TV:
		break;

	    case SA_LIFE_DURATION | ISAKMP_ATTR_AF_TLV:
		break;
	    case COMPRESS_PRIVATE_ALG | ISAKMP_ATTR_AF_TLV:
		break;
#endif
	    default:
		log("unsupported IPsec attribute %s from %s",
		    enum_show(&ipsec_attr_desc, a.isaat_af_type),
		    show_sa(sap));
		return FALSE;
	}
    }
    if (st->st_pfs_group == -1)
	st->st_pfs_group = pfs_group;
    if (st->st_pfs_group != pfs_group)
    {
	log("inconsistent GROUP_DESCRIPTION in IPsec SA from %s",
	    show_sa(sap));
	return FALSE;
    }

    if ((seen_attrs & 1ul << SA_LIFE_DURATION) != 0)
    {
	log("SA_LIFE_TYPE IPsec attribute not followed by SA_LIFE_DURATION attribute in message from %s",
	    show_sa(sap));
	return FALSE;
    }

    if ((seen_attrs & (1ul << ENCAPSULATION_MODE | 1ul << AUTH_ALGORITHM)) == 0)
    {
	log("IPsec transform must specify ENCAPSULATION_MODE or AUTH_ALGORITHM (%s)",
	    show_sa(sap));
	return FALSE;
    }

    /* ??? should check for key_len and/or key_rounds if required */

    return TRUE;
}

static void
echo_proposal(
    struct isakmp_proposal r_proposal,	/* proposal to emit */
    struct isakmp_transform r_trans,	/* winning transformation within it */
    u_int8_t np,			/* Next Payload for proposal */
    pb_stream *r_sa_pbs,		/* SA PBS into which to emit */
    struct ipsec_proto_info *pi,	/* info about this protocol instance */
    struct_desc *trans_desc,		/* descriptor for this transformation */
    pb_stream *trans_pbs)		/* PBS for incoming transform */
{
    pb_stream r_proposal_pbs;
    pb_stream r_trans_pbs;

    /* Proposal */
    r_proposal.isap_np = np;
    r_proposal.isap_notrans = 1;
    if (!out_struct(&r_proposal, &isakmp_proposal_desc, r_sa_pbs, &r_proposal_pbs))
	assert(FALSE);

    /* allocate and emit our SPI */
    get_ipsec_spi(pi->our_spi);
    out_raw(pi->our_spi, IPSEC_DOI_SPI_SIZE, &r_proposal_pbs, "SPI");

    /* Transform */
    r_trans.isat_np = ISAKMP_NEXT_NONE;
    if (!out_struct(&r_trans, trans_desc, &r_proposal_pbs, &r_trans_pbs))
	assert(FALSE);

    /* Transform Attributes: pure echo */
    trans_pbs->cur = trans_pbs->start + sizeof(struct isakmp_transform);
    if (!out_raw(trans_pbs->cur, pbs_left(trans_pbs)
    , &r_trans_pbs, "attributes"))
	assert(FALSE);

    close_output_pbs(&r_trans_pbs);
    close_output_pbs(&r_proposal_pbs);
}

bool
parse_ipsec_sa_body(
    pb_stream *sa_pbs,		/* body of input SA Payload */
    const struct isakmp_sa *sa,	/* header of input SA Payload */
    pb_stream *r_sa_pbs,	/* if non-NULL, where to emit body of winning SA */
    bool selection,		/* if this SA is a selection, only one tranform may appear */
    struct state *st,		/* current state object */
    const struct sockaddr *sap)	/* network address */
{
    u_int32_t ipsecdoisit;
    pb_stream next_proposal_pbs;

    struct isakmp_proposal next_proposal;
    u_char next_spi[IPSEC_DOI_SPI_SIZE];

    bool next_full = TRUE;

    /* DOI */
    if (sa->isasa_doi != ISAKMP_DOI_IPSEC)
    {
	log("Unknown/unsupported DOI %d from %s",
	    enum_show(&doi_desc, sa->isasa_doi),
	    show_sa(sap));
	/* XXX Could send notification back */
	return FALSE;
    }

    /* Situation */
    if (!in_struct(&ipsecdoisit, &ipsec_sit_desc, sa_pbs, NULL))
	return FALSE;

    if (ipsecdoisit != SIT_IDENTITY_ONLY)
    {
	log("unsupported IPsec DOI situation (%s) received from %s",
	    bitnamesof(sit_bit_names, ipsecdoisit),
	    show_sa(sap));
	/* XXX Could send notification back */
	return FALSE;
    }

    /* The rules for IPsec SAs are scattered.
     * draft-ietf-ipsec-isakmp-09.txt section 4.2 gives some info.
     * There may be multiple proposals.  Those with identical proposal
     * numbers must be considered as conjuncts.  Those with different
     * numbers are disjuncts.
     * Each proposal may have several transforms, each considered
     * an alternative.
     * Each transform may have several attributes, all applying.
     *
     * To handle the way proposals are combined, we need to do a
     * look-ahead.
     */

    if (!in_struct(&next_proposal, &isakmp_proposal_desc, sa_pbs, &next_proposal_pbs))
	return FALSE;

    /* for each conjunction of proposals... */
    while (next_full)
    {
	int propno = next_proposal.isap_proposal;
	pb_stream ah_prop_pbs, esp_prop_pbs;
	struct isakmp_proposal ah_proposal, esp_proposal;
	u_char ah_spi[IPSEC_DOI_SPI_SIZE], esp_spi[IPSEC_DOI_SPI_SIZE];
	bool ah_seen = FALSE, esp_seen = FALSE;

	pb_stream ah_trans_pbs, esp_trans_pbs;
	struct isakmp_transform ah_trans, esp_trans;
	struct ipsec_trans_attrs ah_attrs, esp_attrs;

	/* for each proposal in the conjunction */
	do {

	    if (next_proposal.isap_spisize != IPSEC_DOI_SPI_SIZE)
	    {
		log("IPsec proposal from %s with improper SPI size (%u)",
		    show_sa(sap), next_proposal.isap_spisize);
		return FALSE;
	    }

	    if (!in_raw(next_spi, IPSEC_DOI_SPI_SIZE, &next_proposal_pbs, "SPI"))
		return FALSE;

	    /* SPI 0 is reserved.
	     * (So too are 1-256, but we don't check)
	     * draft-ietf-ipsec-auth-header-06.txt 2.4
	     * draft-ietf-ipsec-esp-v2-05.txt 2.1
	     */
	    {
		int i;

		for (i = 0; next_spi[i] == 0; i++)
		{
		    if (i == IPSEC_DOI_SPI_SIZE-1)
		    {
			log("IPsec proposal from %s contains SPI 0",
			    show_sa(sap));
			return FALSE;
		    }
		}
	    }

	    if (next_proposal.isap_notrans == 0)
	    {
		log("IPsec proposal from %s contains no transforms",
		    show_sa(sap));
		return FALSE;
	    }

	    switch (next_proposal.isap_protoid)
	    {
	    case PROTO_IPSEC_AH:
		if (ah_seen)
		{
		    log("IPsec SA from %s contains two simultaneous AH proposals",
			show_sa(sap));
		    return FALSE;
		}
		ah_seen = TRUE;
		ah_prop_pbs = next_proposal_pbs;
		ah_proposal = next_proposal;
		memcpy(ah_spi, next_spi, IPSEC_DOI_SPI_SIZE);
		break;
	    case PROTO_IPSEC_ESP:
		if (esp_seen)
		{
		    log("IPsec SA from %s contains two simultaneous ESP proposals",
			show_sa(sap));
		    return FALSE;
		}
		esp_seen = TRUE;
		esp_prop_pbs = next_proposal_pbs;
		esp_proposal = next_proposal;
		memcpy(esp_spi, next_spi, IPSEC_DOI_SPI_SIZE);
		break;
	    default:
		log("unexpected protoid (%s) in IPsec proposal from %s",
		    enum_show(&protocol_desc, next_proposal.isap_protoid),
		    show_sa(sap));
		return FALSE;
	    }

	    /* refill next_proposal */
	    if (next_proposal.isap_np == ISAKMP_NEXT_NONE)
	    {
		next_full = FALSE;
		break;
	    }
	    else if (next_proposal.isap_np != ISAKMP_NEXT_P)
	    {
		log("unexpected in Proposal from %s: %s",
		    show_sa(sap),
		    enum_show(&payload_desc, next_proposal.isap_np));
		return FALSE;
	    }

	    if (!in_struct(&next_proposal, &isakmp_proposal_desc, sa_pbs, &next_proposal_pbs))
		return FALSE;
	} while (next_proposal.isap_proposal == propno);

	/* Now that we have all conjuncts, we should try
	 * the Cartesian product of eachs tranforms!
	 * At the moment, we take short-cuts on account of
	 * our rudimentary hard-wired policy.
	 * For now, we find an acceptable AH (if any)
	 * and then an acceptable ESP.  The only interaction
	 * is that the ESP acceptance can know whether there
	 * was an acceptable AH and hence not require an AUTH.
	 */

	if (ah_seen)
	{
	    int previous_transnum = -1;
	    int tn;

	    for (tn = 0; tn != ah_proposal.isap_notrans; tn++)
	    {
		if (!parse_ipsec_transform(&ah_trans
		, &ah_attrs
		, &ah_prop_pbs
		, &ah_trans_pbs
		, &isakmp_ah_transform_desc
		, previous_transnum
		, tn == ah_proposal.isap_notrans - 1
		, st
		, sap))
		    return FALSE;

		previous_transnum = ah_trans.isat_transnum;

		/* ??? what does ah_attrs.auth mean? */

		switch (ah_attrs.transid)
		{
		    case AH_MD5:
		    case AH_SHA:
			break;
		    default:
			DBG(DBG_CONTROL | DBG_CRYPT,
			    DBG_log("unsupported AH transform %s from %s",
				enum_show(&ah_transformid_desc, ah_attrs.transid),
				show_sa(sap)));
			continue;   /* try another */
		}
		break;	/* we seem to be happy */
	    }
	    if (tn == ah_proposal.isap_notrans)
		continue;	/* we didn't find a nice one */
	    memcpy(ah_attrs.spi, ah_spi, IPSEC_DOI_SPI_SIZE);
	}

	if (esp_seen)
	{
	    int previous_transnum = -1;
	    int tn;

	    for (tn = 0; tn != esp_proposal.isap_notrans; tn++)
	    {
		if (!parse_ipsec_transform(&esp_trans
		, &esp_attrs
		, &esp_prop_pbs
		, &esp_trans_pbs
		, &isakmp_esp_transform_desc
		, previous_transnum
		, tn == esp_proposal.isap_notrans - 1
		, st
		, sap))
		    return FALSE;

		previous_transnum = esp_trans.isat_transnum;

		switch (esp_attrs.transid)
		{
		    case ESP_3DES:
		    case ESP_DES:
			break;

		    case ESP_NULL:
			if (esp_attrs.auth == AUTH_ALGORITHM_NONE)
			{
			    log("ESP_NULL requires auth algorithm.  SA from %s",
				show_sa(sap));
			    return FALSE;
			}
			break;

		    default:
			DBG(DBG_CONTROL | DBG_CRYPT,
			    DBG_log("unsupported ESP transform %s from %s",
				enum_show(&esp_transformid_desc, esp_attrs.transid),
				show_sa(sap)));
			continue;   /* try another */
		}

		switch (esp_attrs.auth)
		{
		    case AUTH_ALGORITHM_NONE:
		    case AUTH_ALGORITHM_HMAC_MD5:
		    case AUTH_ALGORITHM_HMAC_SHA1:
			break;
		    default:
			DBG(DBG_CONTROL | DBG_CRYPT,
			    DBG_log("unsupported ESP auth alf %s from %s",
				enum_show(&auth_alg_desc, esp_attrs.auth),
				show_sa(sap)));
			continue;   /* try another */
		}

		break;	/* we seem to be happy */
	    }
	    if (tn == esp_proposal.isap_notrans)
		continue;	/* we didn't find a nice one */
	    memcpy(esp_attrs.spi, esp_spi, IPSEC_DOI_SPI_SIZE);
	}

	/* Eureka: we liked what we saw -- accept it. */

	if (r_sa_pbs != NULL)
	{
	    /* emit what we've accepted */

	    /* Situation */
	    if (!out_struct(&ipsecdoisit, &ipsec_sit_desc, r_sa_pbs, NULL))
		assert(FALSE);

	    /* AH proposal */
	    if (ah_seen)
		echo_proposal(ah_proposal,
		    ah_trans,
		    esp_seen? ISAKMP_NEXT_P : ISAKMP_NEXT_NONE,
		    r_sa_pbs,
		    &st->st_ah,
		    &isakmp_ah_transform_desc,
		    &ah_trans_pbs);

	    /* ESP proposal */
	    if (esp_seen)
		echo_proposal(esp_proposal,
		    esp_trans,
		    ISAKMP_NEXT_NONE,
		    r_sa_pbs,
		    &st->st_esp,
		    &isakmp_esp_transform_desc,
		    &esp_trans_pbs);

	    close_output_pbs(r_sa_pbs);
	}

	/* save decoded version of winning SA in state */

	st->st_ah.present = ah_seen;
	if (ah_seen)
	    st->st_ah.attrs = ah_attrs;

	st->st_esp.present = esp_seen;
	if (esp_seen)
	    st->st_esp.attrs = esp_attrs;

	return TRUE;
    }

    log("no acceptable proposal from %s", show_sa(sap));
    return FALSE;
}
