/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <assert.h>

#include "constants.h"
#include "state.h"
#include "defs.h"
#include "kernel.h"
#include "log.h"
#include "rnd.h"
#include "timer.h"

/*
 * This file has the functions that handle the
 * state hash table and the messageid list.
 */

#define STATE_TABLE_SIZE 32

static struct state *statetable[STATE_TABLE_SIZE];
static struct messageid *msgidlist;

#define SA_EQUAL(x, y)   ((x).sa_family == (y).sa_family && \
    memcmp(&((x).sa_data), &((y).sa_data), FULL_INET_ADDRESS_SIZE) == 0)


/*
 * Global variables: had to go somewhere, might as well be this file.
 */

u_int16_t pluto_port = PORT;	/* Pluto's port */
struct sockaddr_in mask32;	/* 255.255.255.255 */

/*
 * Insert a new message-id-keeping-track structure.
 */
struct messageid *
insert_messageid(struct sockaddr sa)
{
    struct messageid *mid;

    mid = alloc_thing(struct messageid, "messageid in insert_messageid()");
    DBG(DBG_CONTROL, DBG_log("inserting messageid structure for %s",
	show_sa(&sa)));
    mid->md_peer = sa;
    mid->md_msgid = 1;	/* network order; ??? ought to be random? */
    mid->md_next = msgidlist;
    msgidlist = mid;
    return mid;
}

/*
 * Update a messageid structure. If one doesn't exist, create it.
 */
void
update_messageid(struct sockaddr *sa, u_int32_t msgid)
{
    struct messageid *mid;

    for (mid = msgidlist; mid != (struct messageid *) NULL; mid = mid->md_next)
    {
	if (SA_EQUAL(mid->md_peer, *sa))
	{
	    if (mid->md_msgid <= msgid)
	    {
		DBG(DBG_CONTROL, DBG_log("updating MSG-ID for %s from %d to %d",
		    get_address(*sa), mid->md_msgid, msgid));
		mid->md_msgid = msgid;
	    }
	    return;
	}
    }
    /* Not found, create one. */
    mid = insert_messageid(*sa);
    mid->md_msgid = msgid;
}

/*
 * Find a message-id-keeping-track structure, and return the
 * next msgid (network order!). If non, return 0.
 */
u_int32_t
find_messageid(struct sockaddr sa)
{
    struct messageid *mid;

    for (mid = msgidlist; mid != (struct messageid *) NULL; mid = mid->md_next)
	if (SA_EQUAL(mid->md_peer, sa))
	    return ++(mid->md_msgid);   /* network order; ??? ought to be random? */
    DBG(DBG_CONTROL, DBG_log("find_messageid(): search failed, no structure for %s",
	show_sa(&sa)));
    return 0;
}

static struct state **
state_hash(const u_char *icookie, const u_char *rcookie, struct sockaddr sa)
{
    u_int i, j;

    DBG(DBG_RAW | DBG_CONTROL,
	DBG_dump("ICOOKIE:", icookie, COOKIE_SIZE);
	DBG_dump("RCOOKIE:", rcookie, COOKIE_SIZE);
	DBG_dump("sockaddr:", sa.sa_data, 4));

    for (i = 0, j = 0; j < COOKIE_SIZE; j++)
	i += icookie[j] + rcookie[j];

    i += sa.sa_family;

    for (j = 0; j < FULL_INET_ADDRESS_SIZE; j++)
	i += sa.sa_data[j];

    DBG(DBG_CONTROL, DBG_log("state hash entry %d", i % STATE_TABLE_SIZE));

    return &statetable[i % STATE_TABLE_SIZE];
}

/*
 * Get a state object.
 */
struct state *
new_state(void)
{
    struct state *st;

    st = alloc_thing(struct state, "struct state in new_state()");

    /* Note: pointers are not guaranteed to be set to NULL
     * by zeroing their bytes
     */
    st->st_oakley.hasher = NULL;
    st->st_packet = NULL;
    st->st_proposal = NULL;
    st->st_p1isa = NULL;
    st->st_ni = NULL;
    st->st_nr = NULL;
    st->st_skeyid = NULL;
    st->st_skeyid_d = NULL;
    st->st_skeyid_a = NULL;
    st->st_skeyid_e = NULL;
    st->st_iv = NULL;
    st->st_enc_key = NULL;
    st->st_myidentity = NULL;
    st->st_peeridentity = NULL;
    st->st_lastblock = NULL;
    st->st_ah.our_keymat = NULL;
    st->st_ah.peer_keymat = NULL;
    st->st_esp.our_keymat = NULL;
    st->st_esp.peer_keymat = NULL;
    st->st_event = NULL;
    st->st_phase1 = NULL;
    st->st_next = NULL;
    st->st_prev = NULL;

    return st;
}

/*
 * Initialize the state table.
 */
void
init_states(void)
{
    int i;

    for (i = 0; i < STATE_TABLE_SIZE; i++)
	statetable[i] = (struct state *) NULL;

    msgidlist = (struct messageid *) NULL;
}

/*
 * Insert a state object in the hash table. The object is inserted
 * at the begining of list.
 */
void
insert_state(struct state *st)
{
    struct state **p = state_hash(st->st_icookie, st->st_rcookie, st->st_peer);

    assert(st->st_prev == NULL && st->st_next == NULL);
    st->st_prev = (struct state *) NULL;	/* redundant, we hope */
    st->st_next = *p;
    *p = st;
    if (st->st_next != (struct state *) NULL)
	st->st_next->st_prev = st;
}

/*
 * Delete a state object from the hash table, but don't free it
 */
void
delete_state(struct state *st)
{
    /* unlink from forward chain */
    struct state **p = st->st_prev == NULL
	? state_hash(st->st_icookie, st->st_rcookie, st->st_peer)
	: &st->st_prev->st_next;

    assert(*p == st);
    *p = st->st_next;

    /* unlink from backward chain */
    if (st->st_next != NULL)
    {
	assert(st->st_next->st_prev == st);
	st->st_next->st_prev = st->st_prev;
    }

    st->st_next = st->st_prev = NULL;
}

/*
 * Free a state object
 */
void
free_state(struct state *st)
{
    assert(st->st_prev == NULL && st->st_next == NULL);

    if (st->st_gi_in_use)
	mpz_clear(&(st->st_gi));

    if (st->st_gr_in_use)
	mpz_clear(&(st->st_gr));

    if (st->st_sec_in_use)
	mpz_clear(&(st->st_sec));

    if (st->st_shared_in_use)
	mpz_clear(&(st->st_shared));

#   define freeany(p) { if ((p) == NULL) free(p); }
    freeany(st->st_packet);
    freeany(st->st_proposal);
    freeany(st->st_p1isa);
    freeany(st->st_ni);
    freeany(st->st_nr);
    freeany(st->st_skeyid);
    freeany(st->st_skeyid_d);
    freeany(st->st_skeyid_a);
    freeany(st->st_skeyid_e);
    freeany(st->st_iv);
    freeany(st->st_enc_key);
    freeany(st->st_myidentity);
    freeany(st->st_peeridentity);
    freeany(st->st_lastblock);
    freeany(st->st_ah.our_keymat);
    freeany(st->st_ah.peer_keymat);
    freeany(st->st_esp.our_keymat);
    freeany(st->st_esp.peer_keymat);
#   undef freeany

    free(st);
}

/*
 * Duplicate a Phase 1 state object, to create a Phase 2 object.
 */
struct state *
duplicate_state(const struct state *st)
{
    struct state *nst;

    nst = new_state();

    nst->st_phase1 = st;

    memcpy(nst->st_icookie, st->st_icookie, COOKIE_SIZE);
    memcpy(nst->st_rcookie, st->st_rcookie, COOKIE_SIZE);
    nst->st_me = st->st_me;
    nst->st_peer = st->st_peer;

    nst->st_doi = st->st_doi;
    nst->st_situation = st->st_situation;

#   define clone_fld(fld, fld_len) { \
	nst->fld = clone_bytes(st->fld, st->fld_len, "field in duplicate_state()"); \
	nst->fld_len = st->fld_len; \
    }

    clone_fld(st_skeyid_d, st_skeyid_d_len);
    clone_fld(st_skeyid_e, st_skeyid_e_len);
    clone_fld(st_skeyid_a, st_skeyid_a_len);
    clone_fld(st_enc_key, st_enc_key_len);
    clone_fld(st_myidentity, st_myidentity_len);
    clone_fld(st_peeridentity, st_peeridentity_len);

    if (st->st_packet != NULL)
	clone_fld(st_packet, st_packet_len);    /* duplicate packet ??? */
#   undef clone_fld

    nst->st_myidentity_type = st->st_myidentity_type;
    nst->st_peeridentity_type = st->st_peeridentity_type;

    nst->st_goal = st->st_goal;

    nst->st_oakley = st->st_oakley;

    nst->st_pfs_group = -1;	/* unspecified */

    return nst;
}

/*
 * Find a state object.
 */
struct state *
find_full_state(const u_char *icookie, const u_char *rcookie,
		const struct sockaddr *sa, u_int32_t /*network order*/  msgid)
{
    struct state *st = *state_hash(icookie, rcookie, *sa);

    while (st != (struct state *) NULL)
	if (SA_EQUAL(*sa, st->st_peer)
	&& memcmp(icookie, st->st_icookie, COOKIE_SIZE) == 0
	&& memcmp(rcookie, st->st_rcookie, COOKIE_SIZE) == 0
	&& msgid == st->st_msgid)
	    break;
	else
	    st = st->st_next;

    DBG(DBG_CONTROL,
	if (st == NULL)
	    DBG_log("full state not found");
	else
	    DBG_log("full state found, state %s",
		enum_show(&state_desc, st->st_state)));

    return st;
}

/*
 * Find an ISAKMP SA state object.
 */
struct state *
find_phase1_state(const struct sockaddr *sa)
{
    struct state *st;
    int i;

    for (i = 0; i < STATE_TABLE_SIZE; i++)
	for (st = statetable[i]; st != (struct state *) NULL; st = st->st_next)
	    if (SA_EQUAL(*sa, st->st_peer)      /* Host we want */
	    && st->st_msgid == 0	        /* ISAKMP SA */
	    && (st->st_state == OAKLEY_MAIN_I_4 || st->st_state == OAKLEY_MAIN_R_3))
		return st;

    return (struct state *) NULL;
}

/*
 * Find a state object in the hash table, without using
 * the rcookie (to find state objects where we haven't yet
 * received a message from the responder).
 */
struct state *
find_half_state(const u_char *icookie, const struct sockaddr *sa)
{
    static const u_char rcookie[COOKIE_SIZE];	/* guaranteed 0 */
    struct state *st = *state_hash(icookie, rcookie, *sa);

    while (st != (struct state *) NULL)
	if (SA_EQUAL(*sa, st->st_peer)
	&& memcmp(icookie, st->st_icookie, COOKIE_SIZE) == 0)
	    break;
	else
	    st = st->st_next;

    DBG(DBG_CONTROL,
	if (st == NULL)
	    DBG_log("half state not found");
	else
	    DBG_log("half state found, state %s",
		enum_show(&state_desc, st->st_state)));
    return st;
}

void
delete_every_sa()
{
    /* ??? tell peer to delete IPsec SAs */

    /* delete IPsec SAs */
    {
	struct state *st;
	int i;

	for (i = 0; i < STATE_TABLE_SIZE; i++)
	{
	    for (st = statetable[i]; st != (struct state *) NULL; st = st->st_next)
	    {
		switch (st->st_state)
		{
		case OAKLEY_QUICK_I_2:
		    delete_ipsec_sa(st, TRUE);
		    break;
		case OAKLEY_QUICK_R_2:
		    delete_ipsec_sa(st, FALSE);
		    break;
		}
	    }
	}
    }

    /* ??? tell peer to delete IKE SAs */

    /* delete IKE SAs */
}
