/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

/*
 * This file defines the state and the event objects
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#include <gmp.h>    /* GNU MP library */

struct messageid
{
    struct sockaddr       md_peer;
    u_int32_t             md_msgid; /* network order */
    struct messageid     *md_next;
};

struct event
{
    time_t          ev_time;
    u_int32_t       ev_type;        /* Event type */
    void           *ev_ptr;         /* Pointer to event specific structure */
    u_int32_t       ev_ptr_len;     /* Length of ptr, where applicable */
    struct event   *ev_next;        /* Pointer to next event */
};

struct hash_desc;   /* declaration (not definition) of tag */


/* Oakley (Phase 1 / Main Mode) transform and attributes
 * This is a flattened/decoded version of what is represented
 * in the Transaction Payload.
 * Names are chosen to match corresponding names in state.
 */
struct oakley_trans_attrs {
    u_int16_t encrypt;		/* Encryption algorithm */
    const struct encrypt_desc *encrypter;	/* package of encryption routines */
    u_int16_t hash;		/* Hash algorithm */
    const struct hash_desc *hasher;	/* package of hashing routines */
    u_int16_t auth;		/* Authentication method */
    u_int16_t group;		/* Oakley group */
    MP_INT *modp_modulus;
    time_t expire;		/* When this SA expires (seconds) */
    u_int32_t exhaust;		/* When this SA is exhausted (kilobytes) */
#if 0 /* not yet */
    u_int16_t prf;		/* Pseudo Random Function */
#endif
};

/* IPsec (Phase 2 / Quick Mode) transform and attributes
 * This is a flattened/decoded version of what is represented
 * by a Transaction Payload.  There may be one for AH and one
 * for ESP.
 */
struct ipsec_trans_attrs {
    u_int8_t transid;	/* transform id */
    u_char spi[IPSEC_DOI_SPI_SIZE];
    time_t life_seconds;	/* When this SA expires */
    u_int32_t life_kilobytes;	/* When this SA expires */
    u_int16_t encapsulation;
    u_int16_t auth;
    u_int16_t key_len;
    u_int16_t key_rounds;
#if 0 /* not implemented yet */
    u_int16_t cmprs_dict_sz;
    u_int32_t cmprs_alg;
#endif
};

/* IPsec per protocol state information */
struct ipsec_proto_info {
    bool present;	/* was this transform specified? */
    struct ipsec_trans_attrs attrs;
    u_char our_spi[IPSEC_DOI_SPI_SIZE];
    u_int16_t keymat_len;	/* same for both */
    u_char *our_keymat;
    u_char *peer_keymat;
};

struct state
{
    struct oakley_trans_attrs st_oakley;

    struct ipsec_proto_info st_ah;
    struct ipsec_proto_info st_esp;
#ifdef KLIPS
    u_char             st_tunnel_spi[IPSEC_DOI_SPI_SIZE]; /* KLUDGE */
#endif

    int                st_pfs_group;           /* group for Phase 2 PFS (-1 for unset) */

    u_int32_t          st_doi;                 /* Domain of Interpretation */
    u_int32_t          st_situation;

    u_int32_t          st_goal;                /* What I am trying to accomplish */
    u_int32_t          st_msgid;	       /* MSG-ID from header.  Network Order!  */

    /* symmetric stuff */

    /* initiator stuff */
    u_int8_t           st_gi_in_use;	       /* bool: does st_gi hold a value */
    MP_INT             st_gi;		       /* Initiator public value */
    u_int8_t           st_icookie[COOKIE_SIZE];/* Initiator Cookie */
    u_char            *st_ni;                  /* Ni nonce */
    u_int16_t          st_ni_len;

    /* responder stuff */
    u_int8_t           st_gr_in_use;           /* bool: does st_gr hold a value */
    MP_INT             st_gr;		       /* Responder public value */
    u_int8_t           st_rcookie[COOKIE_SIZE];/* Responder Cookie */
    u_char            *st_nr;                  /* Nr nonce */
    u_int16_t          st_nr_len;

    /* my stuff */
    struct sockaddr    st_me;                  /* my address */

    /* phase 1 ID payload info: */
    u_char            *st_myidentity;          /* identity (network order) (struct in_addr) */
    u_int16_t          st_myidentity_len;
    u_int8_t           st_myidentity_type;     /* Type of identity */

    struct sockaddr_in st_myuser;              /* our client's net address */
    struct sockaddr_in st_myusernetmask;       /* our client's net mask */
    u_int8_t	       st_myuserprotoid;       /* IDcx.protoid */
    u_int16_t          st_myuserport;

    /* his stuff */
    struct sockaddr    st_peer;                /* Peer's address */

    /* phase 1 ID payload info: */
    u_char            *st_peeridentity;        /* identity (network order) (struct in_addr) */
    u_int16_t          st_peeridentity_len;
    u_int8_t           st_peeridentity_type;   /* Type of identity */

    struct sockaddr_in st_peeruser;            /* peer's client net address */
    struct sockaddr_in st_peerusernetmask;     /* peer's client's net mask */
    u_int8_t	       st_peeruserprotoid;     /* IDcx.protoid */
    u_int16_t          st_peeruserport;

    /* end of symmetric stuff */

    u_int8_t           st_sec_in_use;          /* bool: does st_sec hold a value */
    MP_INT             st_sec;		       /* Our local secret value */
    u_int8_t           st_shared_in_use;       /* bool: does st_shared hold a value */
    MP_INT             st_shared;	       /* Derived shared secret */

    u_int8_t	       st_peeridentity_protocol;
    u_int16_t	       st_peeridentity_port;

    u_int16_t          st_state;               /* State of exchange */
    u_char            *st_packet;              /* Transmitted packet */
    u_int32_t          st_packet_len;          /* Size of transmitted packet */
    u_int8_t           st_retransmit;          /* Number of retransmitts */
    u_char            *st_proposal;            /* Proposal accepted (Payload) */
    u_char            *st_p1isa;               /* Phase 1 initiator SA (Payload) for HASH */
    u_int16_t          st_p1isa_len;
    u_char            *st_skeyid;              /* Key material */
    u_int16_t          st_skeyid_len;
    u_char            *st_skeyid_d;            /* KM for non-ISAKMP key derivation */
    u_int16_t          st_skeyid_d_len;
    u_char            *st_skeyid_a;            /* KM for ISAKMP authentication */
    u_int16_t          st_skeyid_a_len;
    u_char            *st_skeyid_e;            /* KM for ISAKMP encryption */
    u_int16_t          st_skeyid_e_len;
    u_char            *st_iv;		       /* IV for encryption */
    u_int16_t          st_iv_len;
    u_char            *st_enc_key;	       /* Oakley Encryption key */
    u_int16_t          st_enc_key_len;
    u_char            *st_lastblock;	       /* Last block of Phase 1 */
    u_int16_t          st_lastblock_len;

    struct event      *st_event;               /* Pointer to pending event */
    const struct state *st_phase1;	       /* Pointer to Phase 1 object */
    struct state      *st_next;                /* Next in list */
    struct state      *st_prev;		       /* Previous in list */
};

/* global variables */

extern u_int16_t pluto_port;	/* Pluto's port */
extern struct sockaddr_in mask32;	/* 255.255.255.255 */

/* exported functions */

extern struct messageid *insert_messageid(struct sockaddr sa);
extern void update_messageid(struct sockaddr *sa, u_int32_t msgid);
extern u_int32_t find_messageid(struct sockaddr sa);

extern struct state *new_state(void);
extern void init_states(void);
extern void insert_state(struct state *st);
extern void delete_state(struct state *st);
extern void free_state(struct state *st);

extern struct state
    *duplicate_state(const struct state *st),
    *find_full_state(
	const u_char *icookie,
	const u_char *rcookie,
	const struct sockaddr *sa,
	u_int32_t msgid),
    *find_phase1_state(const struct sockaddr *sa),
    *find_half_state(const u_char *icookie, const struct sockaddr *sa);

extern void delete_every_sa(void);
