/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>

#include "constants.h"
#include "state.h"
#include "defs.h"
#include "server.h"
#include "log.h"
#include "rnd.h"
#include "timer.h"

/*
 * This file has the event handling routines. Events are
 * kept as a linked list of event structures. These structures
 * have information like event type, expiration time and a pointer
 * to event specific data (for example, to a state structure).
 */

static struct event *evlist = (struct event *) NULL;

/*
 * This routine places an event in the event list.
 */
void
event_schedule(int type, time_t tm, void *ptr, int ptrlen)
{
    struct event *ev, *evt;
    struct state *st;

    ev = alloc_thing(struct event, "struct event in event_schedule()");

    ev->ev_type = type;
    ev->ev_time = tm + time((time_t *) NULL);
    ev->ev_ptr = ptr;
    ev->ev_ptr_len = ptrlen;

    /*
     * If the event is a retransmission, put a backpointer to the event
     * in the state object, so we can delete the event if we receive
     * a reply.
     * Same if event is a cleanup.
     */
    if (type == EVENT_RETRANSMIT
    || type == EVENT_CLEANUP
    || type == EVENT_SA_EXPIRE)
    {
	st = (struct state *) ptr;
	st->st_event = ev;
    }

    DBG(DBG_CONTROL, DBG_log("inserting event %s, timeout in %u seconds",
	enum_show(&timer_event_desc, type), tm));

    if (evlist == (struct event *) NULL) /* No other events */
    {
	evlist = ev;
	return;
    }

    /* We're installing the most close-to-happen event */
    if (evlist->ev_time >= ev->ev_time)
    {
	ev->ev_next = evlist;
	evlist = ev;
	return;
    }

    for (evt = evlist; evt->ev_next != (struct event *) NULL;
	 evt = evt->ev_next)
	if (evt->ev_next->ev_time >= ev->ev_time)
	    break;

    DBG(DBG_CONTROL, DBG_log("event added after event %s (%p/%d)",
	enum_show(&timer_event_desc, evt->ev_type),
	evt->ev_ptr, evt->ev_ptr_len));

    ev->ev_next = evt->ev_next;
    evt->ev_next = ev;
}

/*
 * Handle the first event on the list.
 */
void
event_handle(int kernelfd, int sock)
{
    time_t tm;
    struct event *ev;
    struct state *st;
    struct sockaddr sa;

    ev = evlist;

    if (ev == (struct event *) NULL)    /* Just paranoid */
    {
	DBG(DBG_CONTROL, DBG_log("empty event list, yet we're called"));
	return;
    }

    tm = time((time_t *) NULL);

    if (tm < ev->ev_time)
    {
	DBG(DBG_CONTROL, DBG_log("called while no event expired (%u/%u, %s)", tm, ev->ev_time,
	    enum_show(&timer_event_desc, ev->ev_type)));
	/*
	 * This will happen if the most close-to-expire event was
	 * a retransmission or cleanup, and we received a packet
	 * at the same time as the event expired. Due to the processing
	 * order in call_server(), the packet processing will happen first,
	 * and the event will be removed.
	 */
	return;
    }

    evlist = evlist->ev_next;		/* Ok, we'll handle this event */

    DBG(DBG_CONTROL,
	if (evlist != (struct event *) NULL)
	    DBG_log("next event is %s (%p/%d)",
		enum_show(&timer_event_desc, evlist->ev_type),
		evlist->ev_ptr, evlist->ev_ptr_len);
	else
	    DBG_log("no event pending (yet)"));

    switch (ev->ev_type)
    {
	case EVENT_REINIT_SECRET:
	    DBG(DBG_CONTROL, DBG_log("event EVENT_REINIT_SECRET handled"));
	    init_secret();
	    break;

	case EVENT_RETRANSMIT:
	    st = (struct state *) ev->ev_ptr;
	    sa = st->st_peer;
	    DBG(DBG_CONTROL, DBG_log(
		"event EVENT_RETRANSMIT for %s, handled (%p)",
		show_sa(&sa), st));
	    if (st->st_retransmit >= MAXIMUM_RETRANSMISSIONS)
	    {
		log(
		    "max number of retransmissions(%d) reached for %s",
		    st->st_retransmit, show_sa(&sa));

		delete_state(st);
		free_state(st);

		/* XXX Send some message to the kernel */
		break;
	    }
	    else
		st->st_retransmit++;

	    DBG_cond_dump(DBG_RAW, "packet", st->st_packet, st->st_packet_len);
	    
	    if (sendto(sock, st->st_packet, st->st_packet_len, 0,
		       (struct sockaddr *) &(st->st_peer),
		       sizeof(st->st_peer)) != st->st_packet_len)
		log_errno("sendto() failed in event_handle, for %s",
		    show_sa(&sa));

	    event_schedule(EVENT_RETRANSMIT, EVENT_RETRANSMIT_DELAY, st, 0);
	    break;

	case EVENT_SA_EXPIRE:
	    /* XXX Do something with kernelfd */
	    st = (struct state *) ev->ev_ptr;
	    sa = st->st_peer;
	    delete_state(st);
	    free_state(st);
	    log("SA expired for %s", show_sa(&sa));
	    break;

	case EVENT_CLEANUP:
	    st = (struct state *) ev->ev_ptr;
	    sa = st->st_peer;
	    delete_state(st);		/* Remove from state table */
	    free_state(st);		/* Free the memory */
	    log("responder state expired for %s", show_sa(&sa));
	    break;

	default:
	    log("unknown event %d expired, ignoring", ev->ev_type);
    }

    free(ev);
}

/*
 * Return the time until the next event in the queue
 * expires, -1 if no jobs in queue.
 */
int32_t
next_event(void)
{
    time_t tm;

    if (evlist == (struct event *) NULL)
	return -1;

    DBG(DBG_CONTROL, DBG_log("next event %s in %d seconds (%p/%d)",
	enum_show(&timer_event_desc, evlist->ev_type),
	evlist->ev_time - time((time_t *) NULL),
	evlist->ev_ptr, evlist->ev_ptr_len));

    tm = time((time_t*) NULL);

    if (evlist->ev_time - tm <= 0)
	return 0;
    else
	return evlist->ev_time - tm;
}

/*
 * Delete an event.
 */
void
delete_event(struct state *st)
{
    struct event *ev;

    if (st->st_event == (struct event *) NULL)
	return;

    if (evlist == (struct event *) NULL)
	return;

    if (st->st_event == evlist)
    {
	if (evlist->ev_type == EVENT_RETRANSMIT)
	    st->st_retransmit = 0;
	evlist = evlist->ev_next;
	free(st->st_event);
	st->st_event = (struct event *) NULL;
	return;
    }

    for (ev = evlist; ev->ev_next != (struct event *) NULL
    ; ev = ev->ev_next)
	if (ev->ev_next == st->st_event)
	{
	    if (ev->ev_type == EVENT_RETRANSMIT)
		st->st_retransmit = 0;
	    ev->ev_next = ev->ev_next->ev_next;
	    free(st->st_event);
	    st->st_event = (struct event *) NULL;
	    return;
	}

    DBG(DBG_CONTROL, DBG_log("event %s to be deleted not found",
	enum_show(&timer_event_desc, st->st_event->ev_type)));
}
