/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <getopt.h>
#include <assert.h>

#include <freeswan.h>

#include "constants.h"
#include "version.c"
#include "whack.h"

static void
usage(const char *mess)
{
    if (mess != NULL && *mess != '\0')
	fprintf(stderr, "%s\n", mess);
    fprintf(stderr,
	"Usage:\n\n"
	"help: whack"
	    " [--help]"
	    " [--version]"
	    "\n\n"
	"initiate: whack"
	    " [--port <port-number>]"
	    " \\\n   "
	    " --peer <peer-ip-address>"
	    " [--peerport <port-number>]"
	    " \\\n   "
	    " [--client <subnet>]"
	    " [--peerclient <subnet>]"
	    " \\\n   "
	    " [--encrypt]"
	    " [--authenticate]"
	    " [--tunnel]"
	    "\n\n"
#ifdef DEBUG
	"debug: whack"
	    " [--port <port-number>]"
	    " [--debug-none]"
	    " [--debug-all]"
	    " \\\n   "
	    " [--debug-raw]"
	    " [--debug-crypt]"
	    " [--debug-parsing]"
	    " \\\n   "
	    " [--debug-emitting]"
	    " [--debug-control]"
	    " [--debug-klips]"
	    "\n\n"
#endif
	"shutdown: whack"
	    " [--port <port-number>]"
	    " --shutdown"
	    "\n\n"
	"FreeS/WAN %s\n",
	freeswan_version);
    exit(mess == NULL? 0 : 1);
}

/*
 * This is a hack for initiating ISAKMP exchanges.
 * All it does is send a message to a port where the daemon listens.
 * The message is composed of a sockaddr_in and a bitmask of
 * goals for the SA (see constants.h).
 */
int
main(int argc, char **argv)
{
    unsigned long pluto_port = IKE_UDP_PORT + 1;    /* host order */
    struct whack_message msg;
    const char *ugh = NULL;
    bool peer_specified = FALSE;

    memset(&msg, '\0', sizeof(msg));
    msg.magic = WHACK_MAGIC;
    msg.peer_port = htons(IKE_UDP_PORT);

    for (;;)
    {
	static const struct option long_opts[] = {
	    /* name, has_arg, flag, val */

	    { "help", FALSE, NULL, 'h' },
	    { "version", FALSE, NULL, 'v' },

	    { "port", TRUE, NULL, 'p' },
	    { "peer", TRUE, NULL, 'd' },
	    { "peerport", TRUE, NULL, 'q' },
	    { "client", TRUE, NULL, 'c' },
	    { "peerclient", TRUE, NULL, 'o' },
	    { "encrypt", FALSE, NULL, 'e' },
	    { "authenticate", FALSE, NULL, 'a' },
	    { "tunnel", FALSE, NULL, 't' },

	    { "shutdown", FALSE, NULL, 's' },

#ifdef DEBUG
	    { "debug-none", FALSE, NULL, 'N' },
	    { "debug-all]", FALSE, NULL, 'A' },
	    { "debug-raw", FALSE, NULL, 'R' },
	    { "debug-crypt", FALSE, NULL, 'X' },
	    { "debug-parsing", FALSE, NULL, 'P' },
	    { "debug-emitting", FALSE, NULL, 'E' },
	    { "debug-control", FALSE, NULL, 'C' },
	    { "debug-klips", FALSE, NULL, 'K' },
#endif
	    { 0,0,0,0 }
	};
	int c;

	if (ugh != NULL)
	    usage(ugh);

	/* Note: we don't like the way short options get parsed
	 * by getopt_long, so we simply pass an empty string as
	 * the list.  It could be "hp:d:c:o:eats" "NARXPECK".
	 */
	c = getopt_long(argc, argv, "", long_opts, NULL);

	switch (c)
	{
	case EOF:	/* end of flags */
	    break;

	case 0: /* long option already handled */
	    continue;

	case '?':	/* diagnostic already printed by getopt_long */
	    ugh = "";	/* print null diagnostic, but exit sadly */
	    break;   /* not actually reached */

	case 'h':	/* --help */
	    usage(NULL);
	    break;	/* not actually reached */

	case 'v':	/* --version */
	    fprintf(stderr, "FreeS/WAN %s\n", freeswan_version);
	    return 0;

	case 'p':	/* --port <port-number> */
	case 'q':	/* --peerport <port-number> */
	    if (optarg == NULL || !isdigit(optarg[0]))
		usage("missing port number");

	    {
		char *endptr;
		long p = strtol(optarg, &endptr, 0);

		if (*endptr != '\0' || endptr == optarg
		|| p <= 0 || p > 0x10000)
		    usage("<port-number> must be a number between 1 and 65535");

		switch (c)
		{
		case 'p':
		    pluto_port = p;
		    break;
		case 'q':
		    msg.peer_port = htons(p);
		    msg.whack_initiate = TRUE;
		    break;
		}
	    }
	    continue;

	case 'd':	/* --peer <peer-ip-address> */
	    ugh = atoaddr(optarg, 0, &msg.peer);
	    msg.whack_initiate = TRUE;
	    peer_specified = TRUE;
	    continue;

	case 'c':	/* --client <subnet> */
	    ugh = atosubnet(optarg, 0, &msg.local_client_net, &msg.local_client_mask);
	    if (ugh == NULL && msg.local_client_net.s_addr == 0)
	    {
		fprintf(stderr, "local client network cannot be 0.0.0.0\n");
		exit(1);
	    }
	    msg.explicit_local_client = TRUE;
	    msg.whack_initiate = TRUE;
	    continue;

	case 'o':	/* --peerclient <subnet> */
	    ugh = atosubnet(optarg, 0, &msg.peer_client_net, &msg.peer_client_mask);
	    msg.explicit_peer_client = TRUE;
	    msg.whack_initiate = TRUE;
	    continue;

	case 'e':	/* --encrypt */
	    msg.goal |= GOAL_ENCRYPT;
	    msg.whack_initiate = TRUE;
	    continue;

	case 'a':	/* --authenticate */
	    msg.goal |= GOAL_AUTHENTICATE;
	    msg.whack_initiate = TRUE;
	    continue;

	case 't':	/* --tunnel */
	    msg.goal |= GOAL_TUNNEL;
	    msg.whack_initiate = TRUE;
	    continue;


	case 's':	/* --shutdown */
	    msg.whack_shutdown = TRUE;
	    continue;

#ifdef DEBUG
	case 'N':	/* --debug-none */
	    msg.debugging = DBG_NONE;
	    msg.whack_options = TRUE;
	    continue;

	case 'A':	/* --debug-all */
	    msg.debugging = DBG_ALL;
	    msg.whack_options = TRUE;
	    continue;

	case 'R':	/* --debug-raw */
	    msg.debugging |= DBG_RAW;
	    msg.whack_options = TRUE;
	    continue;

	case 'X':	/* --debug-crypt */
	    msg.debugging |= DBG_CRYPT;
	    msg.whack_options = TRUE;
	    continue;

	case 'P':	/* --debug-parsing */
	    msg.debugging |= DBG_PARSING;
	    msg.whack_options = TRUE;
	    continue;

	case 'E':	/* --debug-emitting */
	    msg.debugging |= DBG_EMITTING;
	    msg.whack_options = TRUE;
	    continue;

	case 'C':	/* --debug-control */
	    msg.debugging |= DBG_CONTROL;
	    msg.whack_options = TRUE;
	    continue;

	case 'K':	/* --debug-klips */
	    msg.debugging |= DBG_KLIPS;
	    msg.whack_options = TRUE;
	    continue;
#endif
	default:
	    assert(FALSE);	/* unknown return value */
	}
	break;
    }

    if (optind != argc)
	usage("unexpected argument");

    if (!(msg.whack_options || msg.whack_initiate || msg.whack_shutdown))
	usage(NULL);	/* nothing specified -- treat as --help */

    if (msg.whack_initiate)
    {
	if (!peer_specified)
	    usage("must specify a peer when initiating a negotiation");
    }

#ifdef DEBUG
    if (pluto_port != IKE_UDP_PORT + 1)
	fprintf(stderr, "Pluto Daemon port %ld\n", pluto_port);

    if (msg.whack_options)
	fprintf(stderr, "sending debugging options 0%o\n", msg.debugging);

    if (msg.whack_initiate)
    {
	char buf[SUBNETTOA_BUF];

	fprintf(stderr, "initiating SA negotiating with %s port %d goal 0%o\n",
	    inet_ntoa(msg.peer), ntohs(msg.peer_port), msg.goal);
	if (msg.explicit_local_client)
	{
	    subnettoa(msg.local_client_net, msg.local_client_mask, 0, buf, sizeof(buf));
	    fprintf(stderr, "  our client is %s", buf);
	}
	if (msg.explicit_peer_client)
	{
	    subnettoa(msg.peer_client_net, msg.peer_client_mask, 0, buf, sizeof(buf));
	    fprintf(stderr, "  peer's client is %s", buf);
	}
	if (msg.explicit_local_client || msg.explicit_peer_client)
	    fprintf(stderr, "\n");
    }

    if (msg.whack_shutdown)
	fprintf(stderr, "initiating shutdown\n");
#endif

    /* send message to Pluto */
    {
	int sock = socket(PF_INET, SOCK_DGRAM, 0);
	struct sockaddr_in sin;

	if (sock == -1)
	{
	    perror("socket() failed");
	    exit(1);
	}

	memset(&sin, '\0', sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	sin.sin_port = htons(pluto_port);

	if (sendto(sock, &msg, sizeof(msg),
		   0, (struct sockaddr *)&sin, sizeof(sin)) == -1)
	{
	    perror("sendto() failed");
	    exit(1);
	}
    }

    return 0;
}
